
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      configuration.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#include "configuration.h"

/******************************************************************************
**
**    Function: osi_config_new
**
**    Purpose:  create a new config module with the specified
**              file name as the name of the config file.  here
**              we create the global block and add it to the
**              block list.
**
******************************************************************************/

configuration * osi_config_new( const char * file_path )
{
    configuration *new_config;
    block *global_block;

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "creating a new config module." );

    new_config   = osi_malloc( sizeof( configuration ) );
    global_block = new_block();

    new_config->file_data     = NULL;
    new_config->blocks        = list_new();
    new_config->block_names   = string_list_new();

    new_config->warning_count = 0;
    new_config->error_count   = 0;

    memset( new_config->file_name, 0, sizeof( new_config->file_name ) );

    if( file_path != NULL )
    {
        safe_strlcpy( new_config->file_name, file_path,
		              sizeof( new_config->file_name ) - 1 );
    }

    /* we create a default global block here. */

    list_add( new_config->blocks, global_block );

    return new_config;
}


/******************************************************************************
**
**    Function: osi_config_destroy
**
**    Purpose:  destroy the config file data, and destroy the
**              list of blocks.
**
******************************************************************************/

void osi_config_destroy( configuration *cfg )
{
    OSI_DEBUG( __POS__, DEBUG_CONFIG, "destroying a config module." );

    if( cfg != NULL )
    {
        /* destroy config file data. */

        string_list_destroy( cfg->file_data );

        /* get first block, then go through and */
        /* destroy all blocks, then destroy the */
        /* block list itself.                   */

        destroy_block_list( cfg->blocks );
    }
}

/*****************************************************************************
**
**    Function: osi_config_verify
**
**    Purpose:  open the config file, read in the data, then
**              verify both the syntax and the semantis.
**
*****************************************************************************/

bool osi_config_verify( configuration *cfg )
{
    OSI_DEBUG( __POS__, DEBUG_CONFIG, "verifying a config module." );

    if( cfg != NULL )
    {
        if( osi_config_open_and_read( cfg ) == FALSE )
        {
            fprintf( stderr, "unable to open file (%s)\n", cfg->file_name );
            cfg->error_count++;

            return FALSE;
        }

        if( osi_config_verify_syntax( cfg ) == FALSE )
        {
            return FALSE;
        }

        if( osi_config_verify_semantics( cfg ) == FALSE )
        {
            return FALSE;
        }

        return TRUE;
    }

    return FALSE;
}


/****************************************************************************
**
**    Function: osi_config_open_and_read
**
**    NOTE: this function cannot be called multiple times! we need to fix this.
**
**    Purpose:  open the file and read in line by line
**              into the file_data structure.
**              the end result is a file_data structure placed
**              in cfg, each array element being a line that
**              is stripped of any leading/trailing whitespace.
**
****************************************************************************/

bool osi_config_open_and_read( configuration *cfg )
{
    FILE *file = NULL;
    char *line = NULL;
    char temp_line[MAX_LINE_LENGTH];

    string_list *lines = NULL;

    OSI_DEBUG( __POS__, DEBUG_CONFIG,
              "opening and reading in config file data." );

    if( cfg == NULL || cfg->file_name == NULL )
    {
        return FALSE;
    }

    cfg->file_data = string_list_new();
    file = osi_fopen( cfg->file_name, "r" );

    if( file == NULL )
    {
        OSI_DEBUG( __POS__, DEBUG_CONFIG,
            "file: \"%s\" could not be opened, aborting.", cfg->file_name );

        return FALSE;
    }

    lines = string_list_new();

    while(1)
    {
        if( ( line = fgets( temp_line, MAX_LINE_LENGTH, file ) ) == NULL)
        {
            break;
        }

        line = trim_white_space( line );
        string_list_add_item( lines, line );
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "closing file: \"%s\"", cfg->file_name );

    cfg->file_data = lines;
    fclose( file );

    return TRUE;
}

/****************************************************************************
**
**    Function: osi_config_get_global_block
**
**    Purpose:  peer into the block list structure and return the
**              first block in the list, which is always the global one.
**
****************************************************************************/

block * osi_config_get_global_block( configuration *cfg )
{
    if( cfg != NULL && cfg->blocks != NULL )
    {
        return ( cfg->blocks->next->data );
    }

    return NULL;
}

/****************************************************************************
**
**    Function: new_block_from_block
**
**    Purpose:  allocate for a new block, but inherit the attributes
**              from the parent, everything except the action list.
**
*****************************************************************************/

block *new_block_from_block( block *parent_block )
{
    /* create a blank config block. */

    block *config_block = new_block();

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "creating new block from parent." );

    /* now copy all the options from the parent block. */

    config_block->verbose      = parent_block->verbose;
    config_block->prompt       = parent_block->prompt;
    config_block->show_errors  = parent_block->show_errors;
    config_block->recursive    = parent_block->recursive;
    config_block->follow_links = parent_block->follow_links;
    config_block->hash         = parent_block->hash;

    /* copy the database value as well. */

    safe_strlcpy( config_block->database, parent_block->database,
                  sizeof( config_block->database ) );

    /* we don't copy the actions....*/

    return config_block;
}

/******************************************************************************
**
**    Function: new_block
**
**    Purpose:  default allocation for a block, initialized to
**              nothing.
**
******************************************************************************/

block *new_block()
{
    struct block *config_block = osi_malloc( sizeof( struct block ) );

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "creating new block." );

    memset( config_block->directory, 0, sizeof( config_block->directory ) );
    memset( config_block->database, 0, sizeof( config_block->database ) );

    /* new block options according to our specified defaults. */

    config_block->verbose       = FALSE;
    config_block->prompt        = FALSE;
    config_block->show_errors   = FALSE;
    config_block->recursive     = FALSE;
    config_block->follow_links  = FALSE;

    config_block->hash          = HASH_DEFAULT;

    config_block->actions       = list_new();
    config_block->no_entry_list = string_list_new();
    config_block->db            = NULL;

    return config_block;
}


/******************************************************************************
**
**    Function: new_action
**
**    Purpose:  default allocation for an action, initialize to nothing.
**
******************************************************************************/

action *new_action()
{
    struct action *config_action = osi_malloc( sizeof( struct action ) );

    OSI_DEBUG( __POS__, DEBUG_CONFIG,
              "creating new action: %d.", config_action );

    memset( config_action->action_filter.argument, 0,
            sizeof( config_action->action_filter.argument ) );

    config_action->type        = 0;
    config_action->attributes  = 0;

    return config_action;
}


/******************************************************************************
**
**    Function: destroy_block
**
**    Purpose:  free the specified block, destroy the dynamic data
**              found in a block include the noentry and action lists.
**
******************************************************************************/

void destroy_block( block *config_block )
{
    if( config_block != NULL )
    {
        OSI_DEBUG( __POS__,  DEBUG_CONFIG, "destroying block: (%s)",
                   config_block->directory );

        destroy_action_list( config_block->actions );
        string_list_destroy( config_block->no_entry_list );
        osi_free( config_block );
    }
}


/******************************************************************************
**
**    Function: destroy_action
**
**    Purpose:  free the specified action.
**
******************************************************************************/

void destroy_action( action *config_action )
{
    if( config_action != NULL )
    {
        OSI_DEBUG( __POS__, DEBUG_CONFIG,
                   "destroying action: %d.", config_action );

        osi_free( config_action );
    }
}


/******************************************************************************
**
**    Function: osi_config_open_and_read
**
**    Purpose:  frees all of the blocks in the list, then frees
**              the list itself.
**
******************************************************************************/

void destroy_block_list( list block_list )
{
    block *temp_block;
    node *temp_node;

    if( block_list == NULL )
    {
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "destroying list of blocks." );

    /* go to first data node. */

    temp_node = block_list->next;

    /* while there are data nodes, get each */
    /* block and destroy it.                */

    while( temp_node )
    {
        temp_block = temp_node->data;
        destroy_block( temp_block );
        temp_node = temp_node->next;
    }

    list_destroy( block_list );
}


/******************************************************************************
**
**    Function: destroy_action_list
**
**    Purpose:  frees all of the actions in the list, then frees
**              the list itself.
**
******************************************************************************/

void destroy_action_list( list action_list )
{
    action *temp_action;
    node *temp_node;

    if( action_list == NULL )
    {
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "destroying list of actions." );

    /* go to first data node. */

    temp_node = action_list->next;

    /* while there are data nodes, get each */
    /* block and destroy it.                */

    while( temp_node )
    {
        temp_action = temp_node->data;
        destroy_action( temp_action );
        temp_node = temp_node->next;
    }

    list_destroy( action_list );
}

/******************************************************************************
**
**    Function: get_keyword_type
**
**    Purpose:  determine the type of keyword this is based upon
**              our lists.
**
******************************************************************************/

int get_keyword_type( char *string )
{
    char word[MAX_LINE_LENGTH] = "";
    int index;

    if( string == NULL )
    {
        return CONFIG_KEYWORD_NULL;
    }

    get_token( string, word );
    lowercase_string( word );

    for( index = 0; config_keywords[index].word != NULL; index++ )
    {
        if( strncmp( word, config_keywords[index].word,
                     strlen( config_keywords[index].word ) ) == 0 )
        {
            return ( config_keywords[index].type );
        }
    }

    return CONFIG_KEYWORD_UNKNOWN;
}


/******************************************************************************
**
**    Function: verify_multiple_values
**
**    Purpose:  determine if there are any invalid items in the
**              list based upon the list values given.
**
******************************************************************************/

char * verify_multiple_values( char *values, char *valid_values[] )
{
    char *token = NULL;

    int index   = 0;
    int valid   = 0;

    char temp_values[MAX_LINE_LENGTH] = "";

    if( valid_values == NULL || values == NULL )
    {
        return NULL;
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "verifying value list: \"%s\"", values );

    /* strtok will ruin the string we were given */
    /* so we copy the values data.               */

    safe_strlcpy( temp_values, values, ( strlen( values ) + 1 ) );

    /* only one value, don't seperate. */

    token = strtok( temp_values, "," );
    token = trim_white_space( token );

    while( token )
    {
        valid = 0;
        lowercase_string( token );

        for( index = 0; valid_values[index] != NULL; index++ )
        {
            if( strcmp( token, valid_values[index] ) == 0 )
            {
                valid = 1;
            }
        }

        if( !valid )
        {
            return token;
        }

        token = strtok( NULL, "," );
        token = trim_white_space( token );
    }

    return NULL;
}


/******************************************************************************
**
**    Function: verify_single_value
**
**    Purpose:  determine if the value is invalid based upon the
**              valid values list given.
**
******************************************************************************/

char * verify_single_value( char *value, char *valid_values[] )
{
    int index;

    if( valid_values == NULL || value == NULL )
    {
        return NULL;
    }

    lowercase_string( value );

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "verifying value: \"%s\"", value );

    for( index = 0; valid_values[index] != NULL; index++ )
    {
        if( strcmp( value, valid_values[index] ) == 0 )
        {
            return NULL;
        }
    }

    return value;
}


/******************************************************************************
**
**    Function: print_config_error
**
**    Purpose:  display pretty error information includeing
**              line number nad error message.
**
******************************************************************************/

void print_config_error( int line_number, char *line, char *error, ... )
{
    va_list ap;

    fprintf( stderr, "\nline %d: ", line_number );
    fprintf( stderr, "==> %s\n", line );

    va_start( ap, error );
    vfprintf( stderr, error, ap );
    va_end( ap );

    fprintf( stderr, "\n\n" );
}


/******************************************************************************
**
**    Function: get_boolean_value
**
**    Purpose:  examine keyword based upon our list of valid booleans
**              and return TRUE or FALSE appropriatly.
**
******************************************************************************/

int get_boolean_value( char *keyword, bool *value )
{
    int index;

    if( keyword == NULL || value == NULL )
    {
        return 0;
    }

    lowercase_string( keyword );

    for( index = 0; boolean_keywords[index].word != NULL; index++ )
    {
        if( strcmp( keyword, boolean_keywords[index].word ) == 0 )
        {
            *value = boolean_keywords[index].type;
            return 1;
        }
    }

    return 0;
}


/******************************************************************************
**
**    Function: get_keyword_type
**
**    Purpose:  determine the hash type of the string given, e.g. md5
**              and return the constant value associated with that algorithm.
**
******************************************************************************/

int get_hash_value( char *keyword, unsigned int *value )
{
    int index;

    if( keyword == NULL || value == NULL )
    {
        return 0;
    }

    lowercase_string( keyword );

    for( index = 0; hash_keywords[index].word != NULL; index++ )
    {
        if( strcmp( keyword, hash_keywords[index].word ) == 0 )
        {
            *value = hash_keywords[index].type;
            return 1;
        }
    }

    return 0;
}


/******************************************************************************
**
**    Function: get_attributes_value
**
**    Purpose:  get the attributes string from the config file
**              input line given.
**
******************************************************************************/

int get_attributes_value( char *line, unsigned int *value )
{
    char *token = NULL;
    int index   = 0;

    char temp_values[MAX_LINE_LENGTH] = "";

    if( line == NULL || value == NULL )
    {
        return 0;
    }

    /* strtok will ruin the string we were given */
    /* so we copy the values data.               */

    safe_strlcpy( temp_values, line, ( strlen( line ) + 1 ) );

    /* only one value, don't seperate. */

    token = strtok( temp_values, "," );
    token = trim_white_space( token );

    while( token )
    {
        lowercase_string( token );

        for( index = 0; attribute_keywords[index].word != NULL; index++ )
        {
            if( strcmp( token, attribute_keywords[index].word ) == 0 )
            {
                *value = ( (*value) | attribute_keywords[index].type );
                 break;
            }
        }

        token = strtok( NULL, "," );
        token = trim_white_space( token );
    }

    return 1;
}

/******************************************************************************
**
**    Function: get_block_with_name
**
**    Purpose:  get the block in the configuration with the specified
**              name.
**
******************************************************************************/

block * get_block_with_name( configuration *cfg, char *name )
{
    list head;
    
    if( cfg != NULL && name != NULL )
    {
        head = list_get_first_node( cfg->blocks );
        
        while( head )
        {
            block *b = (block *)head->data;
            
            if( b && ( strcmp( b->directory, name ) == 0 ) )
            {
                return ( (block *)head->data );
            }
                
            head = head->next;
        }
    }
    
    return NULL;
}

/******************************************************************************
**
**    Function: get_action_from_block_with_filter
**
**    Purpose:  get the action in the block with the specified
**              filter type.
**
******************************************************************************/

action * get_action_from_block_with_filter( block *config_block,
                                            int filter_type,
                                            char *filter_argument )
{
    list head;
    
    if( config_block != NULL )
    {
        head = list_get_first_node( config_block->actions );
                
        while( head )
        {
            action *a = (action *)head->data;
                        
            if( a && ( a->action_filter.type == filter_type ) )
            {
                /* filter argument is null, so there shouldn't be  */
                /* multiple, this is it.                           */

                if( filter_argument == NULL )
                {
                    return a;
                }

                /* otherwise, there is an argument and we must make   */
                /* sure we find the correct action inside this block. */

                else if( strcmp( filter_argument,
                                 a->action_filter.argument ) == 0 )
                {
                    return a;
                }
            }
                
            head = head->next;
        }
    }
    
    return NULL;
}




/* verify basic semantics.  check for the following: */
/*                                                   */
/*   - unseen directives                             */
/*                                                   */
/*     ExcludeAll                                    */
/*     Include user( "bob" )                         */
/*                                                   */
/*   - duplicate directives                          */
/*                                                   */
/*     Options Recursive,FollowLinks                 */
/*     Options Recursive                             */
/*                                                   */
/*   - multiple blocks for the same directory.       */
/*                                                   */
/*   - unecessary directives                         */
/*                                                   */
/*     IncludeAll                                    */
/*     ...                                           */
/*     <directory /usr>                              */
/*         Include file( "some_file" )               */
/*     </directory>                                  */
/*                                                   */
/* since these are not fatal errors, we just up the  */
/* warning count to give any app the chance to warn  */
/* user before continuing.                           */

bool osi_config_verify_semantics( configuration *cfg )
{
    OSI_DEBUG( __POS__, DEBUG_CONFIG, "verifying semantics." );
    return TRUE;
}

/******************************************************************************
**
**    Function: osi_config_verify_syntax
**
**    Purpose:  examine all of the config file data to make
**              certain there are no syntactical errors in
**              the file, return FALSE upon failure and print
**              any errors with detailed messages
**
******************************************************************************/

bool osi_config_verify_syntax( configuration *cfg )
{
    int index;
    int line_type;
    int line_number;
    int filter_type;

    char *line;
    char *complete_line;

    char *bad_token;
    char *attributes;

    char keyword[MAX_LINE_LENGTH] = "";
    bool inside_directory_block   = FALSE;

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "verifying file syntax." );

    for( index = 0; index < cfg->file_data->size; index++ )
    {
        line          = cfg->file_data->list[index];
        complete_line = line;

        line_type   = get_keyword_type( line );
        line        = get_token( line, keyword );
        line_number = ( index + 1 );

        if( cfg->error_count > CONFIG_MAX_ERRORS )
        {
            return FALSE;
        }

        switch( line_type )
        {
            /* skip over empty lines and comments. */

            case CONFIG_KEYWORD_NULL:
            case CONFIG_KEYWORD_COMMENT:

                break;

            case CONFIG_KEYWORD_DIRECTORY_START:

                if( inside_directory_block )
                {
                    print_config_error( line_number, complete_line,
                                        "misplaced <Directory> statement." );
                    cfg->error_count++;

                    break;
                }

                inside_directory_block = TRUE;

                if( line == NULL || strcmp( line, ">" ) == 0 )
                {
                    print_config_error( line_number, complete_line,
                                        "no directory specified." );
                    cfg->error_count++;

                    break;
                }

                break;

            case CONFIG_KEYWORD_DIRECTORY_END:

                if( !inside_directory_block )
                {
                    print_config_error( line_number, complete_line,
                                        "misplaced </Directory> statement." );
                    cfg->error_count++;

                    break;
                }

                inside_directory_block = FALSE;
                break;

            case CONFIG_KEYWORD_DATABASE:

                if( line == NULL )
                {
                    print_config_error( line_number, complete_line,
                                        "no database file specified." );
                    cfg->error_count++;

                    break;
                }

                break;

            case CONFIG_KEYWORD_VERBOSE:
            case CONFIG_KEYWORD_PROMPT:
            case CONFIG_KEYWORD_SHOWERRORS:
            case CONFIG_KEYWORD_RECURSIVE:
            case CONFIG_KEYWORD_FOLLOWLINKS:

                if( line == NULL )
                {
                    print_config_error( line_number, complete_line,
                                        "no boolean value specified." );
                    cfg->error_count++;

                    break;
                }

                if( verify_single_value( line, valid_boolean_values ) != NULL )
                {
                    print_config_error( line_number, complete_line,
                                    "unrecognized boolean value: %s", line );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_HASH:

                if( line == NULL )
                {
                    print_config_error( line_number, complete_line,
                                        "no hash algorithm specified." );
                    cfg->error_count++;

                    break;
                }

                if( verify_single_value( line, valid_hash_values ) != NULL )
                {
                    print_config_error( line_number, complete_line,
                                    "unrecognized hash algorithm: %s", line );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_INCLUDEALL:

                if( line == NULL )
                {
                    print_config_error( line_number, complete_line,
                                        "no attribute list specified." );
                    cfg->error_count++;

                    break;
                }

                bad_token = verify_multiple_values( line,
                                                    valid_attributes_values );

                if( bad_token != NULL )
                {
                    print_config_error( line_number, complete_line,
                                        "invalid attribute: %s", bad_token );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_EXCLUDEALL:
                break;

            case CONFIG_KEYWORD_INCLUDE:

                if( inside_directory_block )
                {
                    if( line == NULL )
                    {
                        print_config_error( line_number, complete_line,
                                            "no filter specified." );
                        cfg->error_count++;

                        break;
                    }

                    /* verify filter name. */

                    filter_type = get_filter_type( line );

                    if( filter_type == FILTER_TYPE_UNKNOWN )
                    {
                        print_config_error( line_number, complete_line,
                                            "invalid filter: %s", line );
                        cfg->error_count++;

                        break;
                    }

                    /* verify filter value, and any attribute list. */

                    if( filter_is_dynamic( filter_type ) )
                    {
                        attributes = get_filter_value( line, keyword );

                        if( attributes == NULL )
                        {
                            print_config_error( line_number, complete_line, 
                                                "invalid filter: %s", line );
                            cfg->error_count++;

                            break;
                        }

                        bad_token = verify_multiple_values( attributes,
                                                    valid_attributes_values );

                        if( bad_token != NULL )
                        {
                            print_config_error( line_number, line,
                                        "invalid attribute: %s", bad_token );
                            cfg->error_count++;
                        }

                        break;
                    }

                    /* verify attribute list, if any, from static filters. */

                    bad_token = verify_multiple_values( get_token( line,
                                          keyword ), valid_attributes_values );

                    if( bad_token != NULL )
                    {
                        print_config_error( line_number, line,
                                          "invalid attribute: %s", bad_token );
                        cfg->error_count++;
                    }
                }

                else
                {
                    print_config_error( line_number, complete_line,
                           "'Include' directive not allowed in global block." );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_EXCLUDE:

                if( inside_directory_block )
                {
                    if( line == NULL )
                    {
                        print_config_error( line_number, complete_line,
                                            "no filter specified." );
                        cfg->error_count++;

                        break;
                    }

                    /* verify filter name. */

                    filter_type = get_filter_type( line );

                    if( filter_type  == FILTER_TYPE_UNKNOWN )
                    {
                        print_config_error( line_number, line,
                                            "unrecognized filter." );
                        cfg->error_count++;

                        break;
                    }

                    /* if dynamic, verify it's value/syntax. */

                    if( filter_is_dynamic( filter_type ) &&
                        get_filter_value( line, keyword ) == NULL )
                    {
                        print_config_error( line_number, complete_line,
                                            "invalid filter: %s", line );
                        cfg->error_count++;

                        break;
                    }
                }

                else
                {
                    print_config_error( line_number, complete_line,
                        "'Exclude' directive not allowed in global block." );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_NOENTRY:

                if( !inside_directory_block )
                {
                    print_config_error( line_number, complete_line,
                        "'NoEntry' directive not allowed in global block." );
                    cfg->error_count++;
                }

                break;

            case CONFIG_KEYWORD_UNKNOWN:
            default:

                print_config_error( line_number, complete_line,
                                    "unrecognized directive: %s", keyword );
                cfg->error_count++;

                break;

        } /* end switch */

    } /* end for loop */

    return ( cfg->error_count == 0 ) ? TRUE : FALSE;
}

/******************************************************************************
**
**    Function: osi_config_parse
**
**    Purpose:  parse the global data first, then we parse the block
**              data.  we do this because the blocks are to inherit
**              from the global block.
**
******************************************************************************/

bool osi_config_parse( configuration *cfg )
{
    OSI_DEBUG( __POS__, DEBUG_CONFIG, "parsing config data." );

    if( osi_config_parse_globals( cfg ) == FALSE )
    {
        return FALSE;
    }

    if( osi_config_parse_non_globals( cfg ) == FALSE )
    {
        return FALSE;
    }

    return TRUE;
}


/******************************************************************************
**
**    Function: osi_config_parse_globals
**
**    Purpose:  parse the global block data into the specified config module.
**
******************************************************************************/

bool osi_config_parse_globals( configuration *cfg )
{
    int index;
    int line_type;

    char *line;
    char *complete_line;

    char keyword[MAX_LINE_LENGTH] = "";
    action *temp_action;

    block *global_block;
    bool inside_block = FALSE;

    if( osi_config_verify( cfg ) == FALSE )
    {
        return FALSE;
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "parsing global config data." );

    /* create the global block, and make it the */
    /* current block.  there should always be   */
    /* a global block.                          */

    global_block  = cfg->blocks->next->data;

    for( index = 0; index < cfg->file_data->size; index++ )
    {
        line          = cfg->file_data->list[index];
        complete_line = line;

        line_type     = get_keyword_type( line );
        line          = get_token( line, keyword );

        switch( line_type )
        {
            /* skip over empty lines and comments. */

            case CONFIG_KEYWORD_NULL:
            case CONFIG_KEYWORD_COMMENT:

                break;

            case CONFIG_KEYWORD_DIRECTORY_START:

                inside_block = TRUE;

                break;

            case CONFIG_KEYWORD_DIRECTORY_END:

                inside_block = FALSE;

                break;

            case CONFIG_KEYWORD_DATABASE:

                if( inside_block )
                {
                    break;
                }

                get_token( line, global_block->database );

                break;

            case CONFIG_KEYWORD_VERBOSE:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &global_block->verbose );

                break;

            case CONFIG_KEYWORD_PROMPT:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &global_block->prompt );

                break;

            case CONFIG_KEYWORD_SHOWERRORS:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &global_block->show_errors );

                break;

            case CONFIG_KEYWORD_RECURSIVE:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &global_block->recursive );

                break;

            case CONFIG_KEYWORD_FOLLOWLINKS:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &global_block->follow_links );

                break;

            case CONFIG_KEYWORD_HASH:

                if( inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_hash_value( keyword, &global_block->hash );

                break;

            case CONFIG_KEYWORD_INCLUDEALL:

                if( inside_block )
                {
                    break;
                }

                temp_action = new_action();

                temp_action->type = ACTION_TYPE_INCLUDE_ALL;
                get_attributes_value( line, &temp_action->attributes );

                /* add action to block's list of actions. */

                list_add( global_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_EXCLUDEALL:

                if( inside_block )
                {
                    break;
                }

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_EXCLUDE_ALL;

                list_add( global_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_INCLUDE:

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_INCLUDE;

                /* set the filter type. */

                temp_action->action_filter.type = get_filter_type( line );

                /* get argument if there is one, setup filter, then */
                /* get the attributes.                              */

                if( filter_is_dynamic( temp_action->action_filter.type ) )
                {
                   line = get_filter_value( line,
                                        temp_action->action_filter.argument );
                }

                setup_filter_function( &temp_action->action_filter,
                                       temp_action->action_filter.type );

                get_attributes_value( line, &temp_action->attributes );

                /* add action to block's list of actions. */

                list_add( global_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_EXCLUDE:

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_EXCLUDE;

                /* set the filter type. */

                temp_action->action_filter.type = get_filter_type( line );

                /* get argument if there is one, then setup filter function. */

                if( filter_is_dynamic( temp_action->action_filter.type ) )
                {
                    line = get_filter_value( line,
                                    temp_action->action_filter.argument );
                }

                setup_filter_function( &temp_action->action_filter,
                                       temp_action->action_filter.type );

                /* add action to block's list of actions. */

                list_add( global_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_NOENTRY:

                get_token( line, keyword );
                string_list_add_item( global_block->no_entry_list, keyword );

                break;

            case CONFIG_KEYWORD_UNKNOWN:
            default:

                fprintf( stderr, "error: fatal error parsing config file!\n" );
                exit(1);

                /* this should NEVER happen if config data was verified. */

                break;

        } /* end switch */

    } /* end for loop */

    return TRUE;
}

/******************************************************************************
**
**    Function: osi_config_parse_non_globals
**
**    Purpose:  parse all of the blocks, inheriting any missing portions
**              from the global block.  store blocks into the specified
**              config module.
**
******************************************************************************/

bool osi_config_parse_non_globals( configuration *cfg )
{
    int index;
    int line_type;
    int default_attributes = 0;

    bool inside_block = FALSE;
    list action_list;

    char *line;
    char *complete_line;

    char keyword[MAX_LINE_LENGTH] = "";
    action *temp_action;

    block *current_block;
    block *global_block;

    if( osi_config_verify( cfg ) == FALSE )
    {
        return FALSE;
    }

    OSI_DEBUG( __POS__, DEBUG_CONFIG, "parsing all config blocks." );

    /* get the global block, and make it the    */
    /* current block.  there should always be   */
    /* a global block.                          */

    global_block  = cfg->blocks->next->data;
    current_block = global_block;

    /* note default attribute list for the   */
    /* actions with none specified.  we have */
    /* to be careful here in case there is   */
    /* no global attributes list present.    */

    action_list = global_block->actions->next;

    if( action_list != NULL && action_list->data != NULL )
    {
        temp_action = action_list->data;
        default_attributes = temp_action->attributes;
    }

    for( index = 0; index < cfg->file_data->size; index++ )
    {
        line          = cfg->file_data->list[index];
        complete_line = line;

        line_type     = get_keyword_type( line );
        line          = get_token( line, keyword );

        switch( line_type )
        {
            /* skip over empty lines and comments. */

            case CONFIG_KEYWORD_NULL:
            case CONFIG_KEYWORD_COMMENT:

                break;

            case CONFIG_KEYWORD_DIRECTORY_START:

                inside_block = TRUE;

                /* create a new block, add it to the list.  it */
                /* is now the current block.                   */

                current_block = new_block_from_block( global_block );
                list_add( cfg->blocks, current_block );

                /* acquire the directory path and remove the   */
                /* closing ">" if there is one.                */

                get_token( line, current_block->directory );
                line = strrchr( current_block->directory, '>' );

                if( line != NULL )
                {
                    *line = '\0';
                }

                /* add to the block directory names list. */

                string_list_add_item( cfg->block_names,
                                      current_block->directory );

                break;

            case CONFIG_KEYWORD_DIRECTORY_END:

                current_block = global_block;
                inside_block  = FALSE;

                break;

            case CONFIG_KEYWORD_DATABASE:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, current_block->database );

                break;

            case CONFIG_KEYWORD_VERBOSE:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &current_block->verbose );

                break;

            case CONFIG_KEYWORD_PROMPT:

                if( !inside_block )
                {
                   break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &current_block->prompt );

                break;

            case CONFIG_KEYWORD_SHOWERRORS:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &current_block->show_errors );

                break;

            case CONFIG_KEYWORD_RECURSIVE:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &current_block->recursive );

                break;

            case CONFIG_KEYWORD_FOLLOWLINKS:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_boolean_value( keyword, &current_block->follow_links );

                break;

            case CONFIG_KEYWORD_HASH:

                if( !inside_block )
                {
                    break;
                }

                get_token( line, keyword );
                get_hash_value( keyword, &current_block->hash );

                break;

            case CONFIG_KEYWORD_INCLUDEALL:

                if( !inside_block )
                {
                    break;
                }

                temp_action = new_action();

                temp_action->type = ACTION_TYPE_INCLUDE_ALL;
                get_attributes_value( line, &temp_action->attributes );

                /* add action to block's list of actions. */

                list_add( current_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_EXCLUDEALL:

                if( !inside_block )
                     break;

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_EXCLUDE_ALL;

                list_add( current_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_INCLUDE:

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_INCLUDE;

                /* set the filter type. */

                temp_action->action_filter.type = get_filter_type( line );

                /* get argument if there is one, setup filter, then */
                /* get the attributes.                              */

                if( filter_is_dynamic( temp_action->action_filter.type ) )
                {
                    line = get_filter_value( line,
                                temp_action->action_filter.argument );
                }

                else
                {
                    line = strstr( line, " " );
                }

                setup_filter_function( &temp_action->action_filter,
                                       temp_action->action_filter.type );

                get_attributes_value( line, &temp_action->attributes );

                /* assume global attributes if none found. */

                if( temp_action->attributes == 0 )
                {
                    temp_action->attributes = default_attributes;
                }

                /* add action to block's list of actions. */

                list_add( current_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_EXCLUDE:

                temp_action = new_action();
                temp_action->type = ACTION_TYPE_EXCLUDE;

                /* set the filter type. */

                temp_action->action_filter.type = get_filter_type( line );

                /* get argument if there is one, then setup filter function. */

                if( filter_is_dynamic( temp_action->action_filter.type ) )
                {
                    line = get_filter_value( line,
                                    temp_action->action_filter.argument );
                }

                setup_filter_function( &temp_action->action_filter,
                                       temp_action->action_filter.type );

                /* add action to block's list of actions. */

                list_add( current_block->actions, temp_action );

                break;

            case CONFIG_KEYWORD_NOENTRY:

                get_token( line, keyword );
                string_list_add_item( current_block->no_entry_list, keyword );

                break;

            case CONFIG_KEYWORD_UNKNOWN:
            default:

                fprintf( stderr, "error: fatal error parsing config file!\n" );
                exit(1);

                /* this should NEVER happen if config data was verified. */

                break;

        } /* end switch */

    } /* end for loop */

        return TRUE;
}


/******************************************************************************
**
**    Function: osi_config_dump
**
**    Purpose:  debug routine that is a very raw dump of the config data.
**              this is a bit old and unused since the implementation of
**              the debug macros.
**
******************************************************************************/

void osi_config_dump( configuration *cfg )
{
    node *temp_node;

    if( cfg == NULL )
    {
        return;
    }

    fprintf( stderr,
        "\n\n---------------- [ config dump ] ----------------\n\n" );

    fprintf( stderr, "FILE:  " );

    if( cfg->file_name != NULL )
    {
        fprintf( stderr, "%s\n", cfg->file_name );
    }

    else
    {
        fprintf( stderr, "<null>\n" );
    }

    fprintf( stderr, "LINES: " );

    if( cfg->file_data != NULL )
    {
        fprintf( stderr, "%d\n", cfg->file_data->size );
    }

    else
    {
        fprintf( stderr, "<unknown>\n" );
    }

    fprintf( stderr, "\nERRORS:   %d\nWARNINGS: %d\n\n",
             cfg->error_count, cfg->warning_count );

    /* dump all blocks in config. */

    if( cfg->blocks )
    {
        temp_node = cfg->blocks->next;
    }

    fprintf( stderr, "( GLOBAL BLOCK )\n" );
    osi_config_dump_block( temp_node->data );
    temp_node = temp_node->next;

    while( temp_node != NULL )
    {
        osi_config_dump_block( temp_node->data );
        temp_node = temp_node->next;
    }
}

/*****************************************************************************
**
**    Function: osi_config_dump_block
**
**    Purpose:  raw dump of config block for debug purposes.
**
******************************************************************************/

void osi_config_dump_block( block *config_block )
{
    node *temp_node;

    if( config_block == NULL )
    {
        return;
    }

    fprintf( stderr, "\n====================\n\n" );

    fprintf( stderr, "DIRECTORY:  %s\n", config_block->directory );
    fprintf( stderr, "DATABASE:   %s\n", config_block->database );
    fprintf( stderr, "VERBOSE:    %d\n", config_block->verbose );
    fprintf( stderr, "PROMPT:     %d\n", config_block->prompt );
    fprintf( stderr, "SHOWERRORS: %d\n", config_block->show_errors );
    fprintf( stderr, "RECURSIVE:  %d\n", config_block->recursive );
    fprintf( stderr, "FOLLOWLINKS:%d\n", config_block->follow_links );
    fprintf( stderr, "HASH:       %d\n", config_block->hash );

    fprintf( stderr, "NO ENTRY LIST: \n" );
    string_list_dump( config_block->no_entry_list );

    if( config_block->actions )
    {
        temp_node = config_block->actions->next;
    }

    while( temp_node != NULL )
    {
        osi_config_dump_action( temp_node->data );
        temp_node = temp_node->next;
    }

    fprintf( stderr, "\n" );
}

/******************************************************************************
**
**    Function: osi_config_dump_action
**
**    Purpose:  raw dump of action for debugging purposes.
**
******************************************************************************/

void osi_config_dump_action( action *config_action )
{
    fprintf( stderr,
   "[action %d] type(%d) attributes(%d) filter type(%d) filter argument(%s)\n",
    config_action, config_action->type, config_action->attributes,
    config_action->action_filter.type, config_action->action_filter.argument );
}
