
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      configuration.h
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#ifndef CONFIGURATION_H
#define CONFIGURATION_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "debug.h"

#include "list.h"
#include "filter.h"
#include "utilities.h"
#include "string_list.h"
#include "database_handler.h"


/* config file constants. */

#define CONFIG_MAX_ERRORS   5
#define MAX_LINE_LENGTH     4028

/* hash types. */

#define HASH_DEFAULT HASH_MD5
#define HASH_MD5     1
#define HASH_SHA     2
#define HASH_HAVAL   3
#define HASH_RIPEMD  4

/* config keyword types */

#define CONFIG_KEYWORD_UNKNOWN          -1
#define CONFIG_KEYWORD_NULL              0
#define CONFIG_KEYWORD_DATABASE          1
#define CONFIG_KEYWORD_VERBOSE           2
#define CONFIG_KEYWORD_PROMPT            3
#define CONFIG_KEYWORD_SHOWERRORS        4
#define CONFIG_KEYWORD_RECURSIVE         5
#define CONFIG_KEYWORD_FOLLOWLINKS       6
#define CONFIG_KEYWORD_HASH              7
#define CONFIG_KEYWORD_INCLUDEALL        8
#define CONFIG_KEYWORD_EXCLUDEALL        9
#define CONFIG_KEYWORD_COMMENT          10

#define CONFIG_KEYWORD_DIRECTORY_START  11
#define CONFIG_KEYWORD_DIRECTORY_END    12
#define CONFIG_KEYWORD_INCLUDE          13
#define CONFIG_KEYWORD_EXCLUDE          14
#define CONFIG_KEYWORD_NOENTRY          15


/* action types. */

#define ACTION_TYPE_UNKNOWN        0
#define ACTION_TYPE_INCLUDE        1
#define ACTION_TYPE_INCLUDE_ALL    2
#define ACTION_TYPE_EXCLUDE        3
#define ACTION_TYPE_EXCLUDE_ALL    4
#define ACTION_TYPE_NO_ENTRY       5


/* action attributes. */

#define ATTRIBUTE_PERM             0x1
#define ATTRIBUTE_MTIME            0x2
#define ATTRIBUTE_ATIME            0x4
#define ATTRIBUTE_CTIME            0x8
#define ATTRIBUTE_INODE            0x10
#define ATTRIBUTE_LINKS            0x20
#define ATTRIBUTE_UID              0x40
#define ATTRIBUTE_GID              0x80
#define ATTRIBUTE_BYTES            0x100
#define ATTRIBUTE_BLOCKS           0x200
#define ATTRIBUTE_FLAGS            0x400

#define ATTRIBUTE_ALL    ( ATTRIBUTE_PERM | ATTRIBUTE_MTIME | ATTRIBUTE_ATIME \
                         | ATTRIBUTE_CTIME | ATTRIBUTE_INODE | ATTRIBUTE_LINKS \
                         | ATTRIBUTE_UID | ATTRIBUTE_GID | ATTRIBUTE_BYTES \
                         | ATTRIBUTE_BLOCKS | ATTRIBUTE_FLAGS )



static struct keywords config_keywords[] =
{
        { "database", CONFIG_KEYWORD_DATABASE},
        { "verbose", CONFIG_KEYWORD_VERBOSE},
        { "prompt", CONFIG_KEYWORD_PROMPT},
        { "showerrors", CONFIG_KEYWORD_SHOWERRORS},
        { "recursive", CONFIG_KEYWORD_RECURSIVE},
        { "followlinks", CONFIG_KEYWORD_FOLLOWLINKS },
        { "hash", CONFIG_KEYWORD_HASH},
        { "includeall", CONFIG_KEYWORD_INCLUDEALL},
        { "excludeall", CONFIG_KEYWORD_EXCLUDEALL},
        { "#", CONFIG_KEYWORD_COMMENT},
        { "<directory", CONFIG_KEYWORD_DIRECTORY_START},
        { "</directory>", CONFIG_KEYWORD_DIRECTORY_END},
        { "include", CONFIG_KEYWORD_INCLUDE},
        { "exclude", CONFIG_KEYWORD_EXCLUDE},
        { "noentry", CONFIG_KEYWORD_NOENTRY},
        { NULL, 0 }
};

static struct keywords boolean_keywords[] =
{
        { "1", TRUE },
        { "y", TRUE },
        { "yes", TRUE },
        { "true", TRUE },
        { "yup", TRUE },
        { "0", FALSE },
        { "n", FALSE },
        { "no", FALSE },
        { "nope", FALSE },
        { "false", FALSE },
        { NULL, 0 }
};

static struct keywords hash_keywords[] =
{
        { "md5", HASH_MD5 },
        { "sha", HASH_SHA },
        { "haval", HASH_HAVAL },
        { "ripemd", HASH_RIPEMD },
        { NULL, 0 }
};

static struct keywords attribute_keywords[] =
{
        { "all", ATTRIBUTE_ALL },
        { "perm", ATTRIBUTE_PERM },
        { "mtime", ATTRIBUTE_MTIME },
        { "atime", ATTRIBUTE_ATIME },
        { "ctime", ATTRIBUTE_CTIME },
        { "inode", ATTRIBUTE_INODE },
        { "links", ATTRIBUTE_LINKS },
        { "uid", ATTRIBUTE_UID },
        { "gid", ATTRIBUTE_GID },
        { "bytes", ATTRIBUTE_BYTES },
        { "blocks", ATTRIBUTE_BLOCKS },
        { "flags", ATTRIBUTE_FLAGS },
        { NULL, 0 }
};


static char * valid_hash_values[] =
{
    "md5",
    "sha",
    "haval",
    "ripemd",
    NULL
};

static char * valid_attributes_values[] =
{
        "all",
        "perm",
        "mtime",
        "atime",
        "ctime",
        "inode",
        "links",
        "uid",
        "gid",
        "bytes",
        "blocks",
        "flags",
        NULL
};

static char * valid_boolean_values[] =
{
        "1",
        "y",
        "yes",
        "true",
        "yup",
        "0",
        "n",
        "no",
        "nope",
        "false",
        NULL
};

static char * valid_filter_values[] =
{
    "sticky",
    "suid",
    "suid",
    "executable",
    "perl",
    "python",
    "script",
    "gzip",
    "zip",
    "rpm",
    "tar",
    "pgp",
    "uid",
    "gid",
    "user",
    "group",
    "header",
    "file",
    "suffix",
    "md5",
    "sha",
    "haval",
    "ripemd",
    "permissions",
    NULL
};

/* config file structures. */

typedef struct action
{
    int type;
    unsigned int attributes;

    filter action_filter;

} action;

typedef struct block
{
    char directory[MAX_PATH_LEN];
    char database[MAX_PATH_LEN];

    database_handler *db;

    bool verbose;
    bool prompt;
    bool show_errors;
    bool recursive;
    bool follow_links;

    list actions;
    unsigned int hash;
    string_list *no_entry_list;

} block;


typedef struct configuration
{
    char file_name[MAX_PATH_LEN];

    string_list *file_data;
    string_list *block_names;

    unsigned int error_count;
    unsigned int warning_count;

    list blocks;

} configuration;



/* public methods relating to config structures. */

configuration * osi_config_new( const char *file_path );
void osi_config_destroy( configuration *cfg );

bool osi_config_parse( configuration *cfg );
bool osi_config_verify( configuration *cfg );

block * osi_config_get_global_block( configuration *cfg );
block * get_block_with_name( configuration *cfg, char *name );

action * get_action_from_block_with_filter( block *config_block,
                                            int filter_type,
                                            char *filter_argument );

/* semi-public */

bool osi_config_open_and_read( configuration *cfg );

bool osi_config_parse_globals( configuration *cfg );
bool osi_config_parse_non_globals( configuration *cfg );

bool osi_config_verify_semantics( configuration *cfg );
bool osi_config_verify_syntax( configuration *cfg );

void osi_config_dump( configuration *cfg );
void osi_config_dump_block( block *configuration_block );
void osi_config_dump_action( action *configuration_action );



/* internal config parsing and verification methods. */

int get_keyword_type( char *string );

int get_hash_value( char *keyword, unsigned int *value );
int get_boolean_value( char *keyword, bool *value );
int get_attributes_value( char *keyword, unsigned int *value );

char * verify_multiple_values( char *values, char *valid_values[] );
char * verify_single_value( char *value, char *valid_values[] );

void print_config_error( int line_number, char *line, char *error, ... );

/* internal block and action methods. */

block *new_block();
action *new_action();

block *new_block_from_block( block *parent_block );

void destroy_block( block *config_block );
void destroy_action( action *config_action );

void destroy_block_list( list block_list );
void destroy_action_list( list action_list );

#endif
