
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      crypto_utilities.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#include "crypto_utilities.h"

/******************************************************************************
**
**    Function: hash_file_haval
**
**    Purpose:  hash file contents with the haval algorithm.
**
******************************************************************************/

char * hash_file_haval( const char *file_path, char *buffer, int buffer_size )
{
    unsigned char fingerprint[FPTLEN >> 3];
    char file_hash[65] = "";

    int index;

    if( haval_file( (char *)file_path, fingerprint ) )
    {
        return NULL;
    }

    for( index = 0; index < FPTLEN >> 3; index++ )
    {
        safe_snprintf( &file_hash[index * 2], 3, "%02x", fingerprint[index] );
    }

    file_hash[64] = '\0';
    safe_strlcpy( buffer, file_hash, buffer_size );

    return buffer;
}

/******************************************************************************
**
**    Function: hash_file_md5
**
**    Purpose:  hash file contents with the md5 algorithm.
**
******************************************************************************/

char * hash_file_md5( const char *file_path, char *buffer, int buffer_size )
{
    FILE *file_to_hash  = NULL;
    MD5_CTX context;
    unsigned char digest[MD5_DIGEST_LENGTH];

    int burst;
    int index;

    static unsigned char hash_buffer[BUFFER_SIZE];
    char file_hash[33];

    if( ( file_to_hash = fopen( file_path, "r" ) ) == NULL )
    {
        return NULL;
    }

    /* initialize the md5 library and our hash string */
    /* which is 32 bytes long for md5.                */

    MD5_Init( &context );

    while(1)
    {
        burst = read( fileno( file_to_hash ), hash_buffer, BUFFER_SIZE );

        if( burst <= 0 )
        {
            break;
        }

        MD5_Update( &context, hash_buffer, (unsigned long)burst );
    }

    MD5_Final( &( digest[0] ), &context );

    for ( index = 0; index < MD5_DIGEST_LENGTH; index++)
    {
        safe_snprintf( &file_hash[index * 2], 3, "%02x", digest[index] );
    }

    file_hash[32] = '\0';
    fclose( file_to_hash );
    safe_strlcpy( buffer, file_hash, buffer_size );

    return buffer;
}

/******************************************************************************
**
**    Function: hash_file_sha
**
**    Purpose:  hash file contents with the sha1 algorithm.
**
******************************************************************************/

char * hash_file_sha( const char *file_path, char *buffer, int buffer_size )
{
    FILE *file_to_hash  = NULL;
    SHA_CTX context;
    unsigned char digest[SHA_DIGEST_LENGTH];

    int burst;
    int index;

    static unsigned char hash_buffer[BUFFER_SIZE];
    char file_hash[41];

    if( ( file_to_hash = fopen( file_path, "r" ) ) == NULL )
    {
        return NULL;
    }

    /* initialize the sha library and our hash string */
    /* which is 40 bytes long for sha.                */

    SHA1_Init( &context );

    while(1)
    {
        burst = read( fileno( file_to_hash ), hash_buffer, BUFFER_SIZE );

        if( burst <= 0 )
        {
            break;
        }

        SHA1_Update( &context, hash_buffer, (unsigned long)burst );
    }

    SHA1_Final( &( digest[0] ), &context );

    for ( index = 0; index < SHA_DIGEST_LENGTH; index++)
    {
        safe_snprintf( &file_hash[index * 2], 3, "%02x", digest[index] );
    }

    file_hash[40] = '\0';
    fclose( file_to_hash );
    safe_strlcpy( buffer, file_hash, buffer_size );

    return buffer;
}

/******************************************************************************
**
**    Function: hash_file_ripemd
**
**    Purpose:  hash file contents with the ripemd 160 bit algorithm.
**
******************************************************************************/

char * hash_file_ripemd( const char *file_path, char *buffer, int buffer_size )
{
    FILE *file_to_hash;                 /* pointer to file <fname>      */
    unsigned char data[1024];           /* contains current mess. block */
    unsigned long nbytes;               /* length of this block         */
    unsigned long MDbuf[RMDsize/32];    /* contains (A, B, C, D(, E))   */
    unsigned long X[16];                /* current 16-word chunk        */
    unsigned char hashcode[RMDsize/8];  /* temporary output buffer      */
    unsigned int i, j;                  /* counters                     */
    unsigned long length[2];            /* length in bytes of message   */
    unsigned long offset;               /* # of unprocessed bytes at    */
                                        /*          call of MDfinish    */

    /* try to open the file. */

    if( ( file_to_hash = fopen( file_path, "rb" ) ) == NULL )
    {
        return NULL;
    }

    if( buffer_size < ( RMDsize/8 ) )
    {
        return NULL;
    }

    MDinit( MDbuf );
    length[0] = 0;
    length[1] = 0;

    while( ( nbytes = fread( data, 1, 1024, file_to_hash ) ) != 0 )
    {
        /* process all complete blocks */

        for( i=0; i < ( nbytes >> 6 ); i++ )
        {
            for( j=0; j < 16; j++ )
            {
                X[j] = BYTES_TO_DWORD( data + 64 * i + 4 * j );
            }

            compress( MDbuf, X );
        }

        /* update length[] */

        if( length[0] + nbytes < length[0] )
        {
            length[1]++;                  /* overflow to msb of length */
        }

        length[0] += nbytes;
    }

    /* finish: */

    offset = length[0] & 0x3C0;   /* extract bytes 6 to 10 inclusive */
    MDfinish( MDbuf, ( data + offset ), length[0], length[1] );

    for( i = 0; ( i < ( RMDsize / 8 ) ); i += 4)
    {
        hashcode[i]   =  MDbuf[i>>2];
        hashcode[i+1] = ( MDbuf[i>>2] >>  8 );
        hashcode[i+2] = ( MDbuf[i>>2] >> 16 );
        hashcode[i+3] = ( MDbuf[i>>2] >> 24 );
    }

    /* turn into hex. */

    for( j = 0; j < ( RMDsize/8 ); j++ )
    {
        safe_snprintf( &buffer[j*2], 3, "%02x", hashcode[j] );
    }

    fclose( file_to_hash );
    return buffer;
}

