
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      filter.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
*****************************************************************************/

#include "filter.h"

/******************************************************************************
**
**    Function: setup_filter_function
**
**    Purpose:  based upon the value/type of the filter, we
**              set the filter process function pointer to
**              the correct one.
**
******************************************************************************/

int setup_filter_function( filter *action_filter, int type )
{
    if( action_filter == NULL )
    {
        return 0;
    }

    switch( type )
    {
        case FILTER_TYPE_STICKY:
            action_filter->perform = filter_sticky;
            break;

        case FILTER_TYPE_SUID:
            action_filter->perform = filter_suid;
            break;

        case FILTER_TYPE_SGID:
            action_filter->perform = filter_sgid;
            break;

        case FILTER_TYPE_EXECUTABLE:
            action_filter->perform = filter_executable;
            break;

        case FILTER_TYPE_PERL:
            action_filter->perform = filter_perl;
            break;

        case FILTER_TYPE_PYTHON:
            action_filter->perform = filter_python;
            break;

        case FILTER_TYPE_SCRIPT:
            action_filter->perform = filter_script;
            break;

        case FILTER_TYPE_GZIP:
            action_filter->perform = filter_gzip;
            break;

        case FILTER_TYPE_ZIP:
            action_filter->perform = filter_zip;
            break;

        case FILTER_TYPE_TAR:
            action_filter->perform = filter_tar;
            break;

        case FILTER_TYPE_PGP:
            action_filter->perform = filter_pgp;
            break;

        case FILTER_TYPE_RPM:
            action_filter->perform = filter_rpm;
            break;

        case FILTER_TYPE_UID:
            action_filter->perform = filter_uid;
            break;

        case FILTER_TYPE_GID:
            action_filter->perform = filter_gid;
            break;

        case FILTER_TYPE_USER:
            action_filter->perform = filter_user;
            break;

        case FILTER_TYPE_GROUP:
            action_filter->perform = filter_group;
            break;

        case FILTER_TYPE_HEADER:
            action_filter->perform = filter_header;
            break;

        case FILTER_TYPE_FILE:
            action_filter->perform = filter_file;
            break;

        case FILTER_TYPE_SUFFIX:
            action_filter->perform = filter_suffix;
            break;

        case FILTER_TYPE_MD5:
            action_filter->perform = filter_md5;
            break;

        case FILTER_TYPE_SHA:
            action_filter->perform = filter_sha;
            break;

        case FILTER_TYPE_HAVAL:
            action_filter->perform = filter_haval;
            break;

        case FILTER_TYPE_RIPEMD:
            action_filter->perform = filter_ripemd;
            break;

        case FILTER_TYPE_PERMISSIONS:
            action_filter->perform = filter_permissions;
            break;

        case FILTER_TYPE_UNKNOWN:
           default:

            return 0;
            break;
     }

    return 1;
}

/******************************************************************************
**
**    Function: filter_is_dynamic
**
**    Purpose:  determine if the filter takes an argument ( dyname ) or
**              if it's just a static filter.
**
******************************************************************************/

bool filter_is_dynamic( int filter_type )
{
    switch( filter_type )
    {
        case FILTER_TYPE_UID:
        case FILTER_TYPE_GID:
        case FILTER_TYPE_USER:
        case FILTER_TYPE_GROUP:
        case FILTER_TYPE_HEADER:
        case FILTER_TYPE_FILE:
        case FILTER_TYPE_SUFFIX:
        case FILTER_TYPE_MD5:
        case FILTER_TYPE_SHA:
        case FILTER_TYPE_HAVAL:
        case FILTER_TYPE_RIPEMD:
        case FILTER_TYPE_PERMISSIONS:

            return TRUE;
            break;

        default:

            return FALSE;
            break;
     }
}


/******************************************************************************
**
**    Function: get_filter_type
**
**    Purpose:  determines the type of the filter based upon the
**              filter names we have.
**
******************************************************************************/

int get_filter_type( char *line )
{
    int index = 0;
    int type  = FILTER_TYPE_UNKNOWN;

    if( line == NULL )
    {
        return type;
    }

    /* check for static filter first. */

    for( index = 0; filter_keywords[index].word != NULL; index++ )
    {
        int size   = strlen( filter_keywords[index].word );
        int type   = filter_keywords[index].type;

        char *word = filter_keywords[index].word;

        if( strncmp( line, word, size ) == 0 )
        {
            /* skip over the filter name, to next visible character. */

            line = line + size;

            if( !isgraph( *line ) || *line == '(' )
            {
                return type;
            }

            /* whoops, not quite right, restore line back to          */
            /* normal, we might still find this to be a valid filter. */

            line = line - size;
        }
    }

    return FILTER_TYPE_UNKNOWN;
}


/******************************************************************************
**
**    Function: get_filter_value
**
**    Purpose:  gets the value found in the first filter found.
**              the filter name need not be valid, but the
**              syntax of the filter must be.  if not, the
**              buffer contents is undefined and NULL is
**              returned.  otherwise, the buffer contains the
**              value ( minus any outside quotes ) and a valid
**              pointer to the data after the filter is returned.
**
******************************************************************************/


char * get_filter_value( char *line, char *value )
{
    char *end;
    char *start;
    char *attributes;

    if( line == NULL || value == NULL )
    {
        return NULL;
    }

    /* define the start and end of the filter. */

    start = strchr( line, '(' );

    if( start == NULL )
    {
        return NULL;
    }

    end = strrchr( start, ')' );

    if( end == NULL )
    {
        return NULL;
    }

    attributes = ( end + 1 );

    /*  remove the () from each end. */

    start++;
    end--;

    /* remove white space. */

    while( *start && !isgraph( *start ) )
    {
        start++;
    }

    while( *end && !isgraph( *end ) )
    {
        end--;
    }

    if( start > end )
    {
        return NULL;
    }

    /* remove any quotes, and basic consistency. */

    if( *start == '"' || *end == '"' )
    {
        if( *end != '"' || *start != '"' )
        {
            return NULL;
        }

        start++;
        end--;
    }

    /* copy value into the buffer. */

    while( start <= end )
    {
        *value = *start;

        start++;
        value++;
    }

    *value = '\0';
    return attributes;
}


/******************************************************************************
**
**    Function: get_filter_name_from_type
**
**    Purpose:  use our filter name->constant table to return
**              the string name of the specified filter type.
**
******************************************************************************/

char *get_filter_name_from_type( unsigned int filter_type )
{
    static char *name;
    int index;

    for( index = 0; filter_keywords[index].word != NULL; index++ )
    {
        if( filter_keywords[index].type == filter_type )
        {
            return filter_keywords[index].word;
        }
    }

    return NULL;
}


/* filter implementations. */


bool filter_sticky( struct stat *stats, const char *file_name,
                    const char *argument )
{
    if( stats->st_mode & S_ISVTX )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_suid( struct stat *stats, const char *file_name,
                  const char *argument )
{
    if( stats->st_mode & S_ISUID )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_sgid( struct stat *stats, const char *file_name,
                  const char *argument )
{
    if( stats->st_mode & S_ISGID )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_executable( struct stat *stats, const char *file_name,
                        const char *argument )
{
    int length              = 0;
    long lvalue             = 0;
    unsigned short svalue   = 0;

    unsigned char number[6] = "\0";
    unsigned char elf[5]    = "\177ELF";

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 6, file );
    fclose( file );

    if( length != 6 )
    {
        return FALSE;
    }

    if( memcmp( elf, number, 4 ) == 0 )
    {
        return TRUE;
    }

#if   defined( PLATFORM_DARWIN ) || defined( PLATFORM_OPENBSD )

    memcpy( &lvalue, number, 4 );

    if( lvalue == 000000407 )
    {
        return TRUE;
    }

    if( lvalue == 4277009102 )
    {
        return TRUE;
    }

#ifndef WORDS_BIGENDIAN

    if( ( lvalue & 0377777777 ) == 041400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400413 )
    {
        return TRUE;
    }

#endif

    if( ( lvalue & 0377777777 ) == 041600413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041600410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041600407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042000413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042000410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042200407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042200410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042200413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042400410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042400413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042600407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042600410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 042600413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043000407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043000410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043000413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043400410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043400413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043600407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043600410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 043600413 )
    {
        return TRUE;
    }

#elif defined( PLATFORM_FREEBSD )

    memcpy( &lvalue, number, 4 );

    if( ( lvalue & 0377777777 ) == 041400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400410 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400413 )
    {
        return TRUE;
    }

    if( ( lvalue & 0377777777 ) == 041400314 )
    {
        return TRUE;
    }

#elif defined( PLATFORM_BSD ) && !defined( WORDS_BIGENDIAN )

    if( number[0] == 204  || number[0] == 263 )
    {
        return TRUE;
    }

    if( number[0] == 264 || number[0] == 267 )
    {
        return TRUE;
    }

#elif defined( PLATFORM_LINUX )

    memcpy( &svalue, number, 4 );

    if( svalue == 0x00640107 )
    {
        return TRUE;
    }

    if( svalue == 0x00640108 )
    {
        return TRUE;
    }

    if( svalue == 0x0064010b )
    {
        return TRUE;
    }

    if( svalue == 0x006400cc )
    {
        return TRUE;
    }

    if( number[0] == '\01' && number[1] == '\03' )
    {
        if( number[2] == '\020' && number[3] == '\04' )
        {
            return TRUE;
        }

        if( number[2] == '\040' && number[3] == '\04' )
        {
            return TRUE;
        }
	}

#elif defined( PLATFORM_SOLARIS ) || defined( PLATFORM_SUNOS )

    memcpy( &lvalue, number, 4 );

    if( ( lvalue & 077777777 ) == 0600413 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0600410 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0600407 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0400413 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0400410 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0400407 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0200413 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0200410 )
    {
        return TRUE;
    }

    if( ( lvalue & 077777777 ) == 0200407 )
    {
        return TRUE;
    }

#endif

    return FALSE;
}

bool filter_uid( struct stat *stats, const char *file_name,
                 const char *argument )
{
    int uid;

    if( argument != NULL )
    {
        uid = atoi( argument );
    }

    if( uid == stats->st_uid )
    {
       return TRUE;
    }

    return FALSE;
}

bool filter_gid( struct stat *stats, const char *file_name,
                 const char *argument )
{
    int gid;

    if( argument != NULL )
    {
        gid = atoi( argument );
    }

    if( gid == stats->st_gid )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_user( struct stat *stats, const char *file_name,
                  const char *argument )
{
    struct passwd *file_owner = getpwuid( stats->st_uid );

    if( file_owner != NULL && argument != NULL )
    {
        if( strcmp( file_owner->pw_name, argument ) == 0 )
        {
            return TRUE;
        }
    }

    return FALSE;
}

bool filter_group( struct stat *stats, const char *file_name,
                   const char *argument )
{
    struct group *file_group = getgrgid( stats->st_gid );

    if( file_group != NULL && argument != NULL )
    {
        if( strcmp( file_group->gr_name, argument ) == 0 )
        {
            return TRUE;
        }
    }

    return FALSE;
}

bool filter_header( struct stat *stats, const char *file_name,
                    const char *argument )
{
    unsigned char number[MAX_FILTER_ARGUMENT_LENGTH] = "\0";
    unsigned char buffer[MAX_FILTER_ARGUMENT_LENGTH] = "\0";

    FILE *file = NULL;

    int index  = 0;
    int length = 0;

    memset( number, 0, sizeof( number ) );

    if( file_name == NULL || argument == NULL )
    {
        return FALSE;
    }

    /* lowercase the value so we can compare it */
    /* if there is a hex prefix, skip past it.  */

    lowercase_string( (char *)argument );

    if( strncmp( argument, "0x", 2 ) == 0 )
    {
        argument += 2;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, strlen( argument ), file );
    fclose( file );

    for( index = 0; index < length; index++ )
    {
        safe_snprintf( &buffer[index*2], 2, "%02x", number[index] );
    }

    buffer[length*2] = '\0';

    if( memcmp( argument, buffer, strlen(argument) ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filename_regex( const char *name, const char *pattern )
{
    regex_t preg;
    regmatch_t results;

    if( name != NULL && pattern != NULL )
    {
        memset( &preg, 0, sizeof( preg ) );
        memset( &results, 0, sizeof( results ) );

        if( regcomp( &preg, pattern, REG_EXTENDED ) == 0 )
        {
            if( regexec( &preg, name, 1, &results, 0 ) == 0 )
            {
                return TRUE;
            }
        }
    }
  
    return FALSE;
}

bool filter_file( struct stat *stats, const char *file_name,
                  const char *argument )
{
    char *file;

    if( file_name == NULL || argument == NULL )
    {
        return FALSE;
    }

    file = strrchr( file_name, '/' );

    if( file == NULL )
    {
        return FALSE;
    }

    file++;
    return ( filename_regex( file, argument ) );
}

bool filter_suffix( struct stat *stats, const char *file_name,
                    const char *argument )
{
    char *start;

    if( file_name == NULL || argument == NULL )
    {
        return FALSE;
    }

    /* find last occurance of "." in the filename. */

    start = strrchr( file_name, '.' );

    if( start != NULL )
    {
        start++;

        if( strcmp( start, argument ) == 0 )
        {
            return TRUE;
        }
    }

    return FALSE;
}

bool filter_md5( struct stat *stats, const char *file_name,
                 const char *argument )
{
    char hash[64];

    if( argument == NULL || file_name == NULL )
    {
        return FALSE;
    }

    /* we allow optional hex prefix here, skip */
    /* over it if there is one.                */

    if( strncmp( argument, "0x", 2 ) == 0 )
    {
        argument += 2;
    }

    hash_file_md5( file_name, hash, sizeof( hash ) );

    if( strcmp( hash, argument ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_sha( struct stat *stats, const char *file_name,
                 const char *argument )
{
    char hash[64];

    if( argument == NULL || file_name == NULL )
    {
        return FALSE;
    }

    /* we allow optional hex prefix here, skip */
    /* over it if there is one.                */

    if( strncmp( argument, "0x", 2 ) == 0 )
    {
        argument += 2;
    }

    hash_file_sha( file_name, hash, sizeof( hash ) );

    if( strcmp( hash, argument ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_haval( struct stat *stats, const char *file_name,
                   const char *argument )
{
    char hash[64];

    if( argument == NULL || file_name == NULL )
    {
        return FALSE;
    }

    /* we allow optional hex prefix here, skip */
    /* over it if there is one.                */

    if( strncmp( argument, "0x", 2 ) == 0 )
    {
        argument += 2;
    }

    hash_file_haval( file_name, hash, sizeof( hash ) );

    if( strcmp( hash, argument ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_ripemd( struct stat *stats, const char *file_name,
                    const char *argument )
{
    char hash[64];

    if( argument == NULL || file_name == NULL )
    {
        return FALSE;
    }

    /* we allow optional hex prefix here, skip */
    /* over it if there is one.                */

    if( strncmp( argument, "0x", 2 ) == 0 )
    {
        argument += 2;
    }

    hash_file_ripemd( file_name, hash, sizeof( hash ) );

    if( strcmp( hash, argument ) == 0 )
    {
        return TRUE;
    }

    return FALSE;}

bool filter_permissions( struct stat *stats, const char *file_name,
                         const char *argument )
{
    char permissions[11] = "";
    int index            = 0;
    int argument_length  = 0;

    if( argument == NULL || file_name == NULL || stats == NULL )
    {
        return FALSE;
    }

    argument_length = strlen( argument );

    get_file_attribute_string( permissions, sizeof( permissions ),
                               stats->st_mode );

    for( index = 0; ( index < 11 ) && ( index < argument_length ); index++ )
    {
	    /* bit is equal or wildcard, move on to the next one. */

        if( ( argument[index] == permissions[index] ) ||
              argument[index] == '*' )
        {
            continue;
        }

        /* otherwise, it's not equal, filter fails to catch this one. */

        return FALSE;
    }

    return TRUE;
}

bool filter_perl( struct stat *stats, const char *file_name,
                  const char *argument )
{
    char number[30] = "\0";
    int length = 0;

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 30, file );
    fclose( file );

    if( length != 30 )
    {
        return FALSE;
    }

    /* look for script headers. */

    if( strstr( number, "/bin/perl" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/perl" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/perl" ) != 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_python( struct stat *stats, const char *file_name,
                    const char *argument )
{
    int length     = 0;

    char number[30] = "\0";
    char magic[5]   = { 10, 13, 78, 153 };

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 30, file );
    fclose( file );

    if( length != 30 )
    {
        return FALSE;
    }

    /* look for script headers. */

    if( strstr( number, "/bin/python" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/python" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/python" ) != 0 )
    {
        return TRUE;
    }

    if( memcmp( number, "\"\"\"", 3 ) == 0 )
    {
        return TRUE;
    }

    if( memcmp( number, magic, 4 ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

/* here we look for shells,zsh,ash.ae,nawk,gawk, and env */

bool filter_script( struct stat *stats, const char *file_name,
                    const char *argument )
{
    char number[25] = "\0";
    int length = 0;

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 25, file );
    fclose( file );

    if( length != 25 )
    {
        return FALSE;
    }

    /* shell archive or script for antique kernel text */

    if( strncmp( number, ": ", 2 ) == 0 )
    {
        return TRUE;
    }

    /* awk script text */

    if( strncmp( number, "BEGIN", 5 ) == 0 )
    {
        return TRUE;
    }

    /* for non scripts, bail now. */

    if( strncmp( number, "#!", 2 ) != 0 )
    {
        return FALSE;
    }

    /* look for shell scripts. */

    if( strstr( number, "/bin/sh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/csh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/bash" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/bash" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/ksh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/tcsh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/tcsh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/tcsh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/zsh" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/ash" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/ae" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/nawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/nawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/nawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/gawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/gawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/local/bin/gawk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/awk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/awk" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/bin/rc" ) != 0 )
    {
        return TRUE;
    }

    if( strstr( number, "/usr/bin/env" ) != 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_pgp( struct stat *stats, const char *file_name,
                 const char *argument )
{
    unsigned char number[15] = "\0";
    int length      = 0;

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 15, file );
    fclose( file );

    if( length != 15 )
    {
        return FALSE;
    }

    if( number[0] == 153 && number[1] == 0 )
    {
        return TRUE;
    }

    if( number[0] == 153 && number[1] == 1 )
    {
        return TRUE;
    }

    if( number[0] == 149 && number[1] == 0 )
    {
        return TRUE;
    }

    if( number[0] == 149 && number[1] == 1 )
    {
        return TRUE;
    }

    if( number[0] == 166 && number[1] == 0 )
    {
        return TRUE;
    }

    if( memcmp( number, "-----BEGIN\040PGP", 14 ) == 0 )
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_gzip( struct stat *stats, const char *file_name,
                  const char *argument )
{
    char number[3] = "\0\0\0";
    int length = 0;

    FILE *file;

    if( file_name == NULL )
    {
        return FALSE;
    }

    if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
        return FALSE;
    }

    length = fread( number, 1, 2, file );
    fclose( file );

    if( length != 2 )
    {
        return FALSE;
    }

    if( number[0] == '\037' && number[1] == '\213')
    {
        return TRUE;
    }

    return FALSE;
}

bool filter_zip( struct stat *stats, const char *file_name,
                 const char *argument )
{
	char number[5] = "\0\0\0\0\0";
    char magic[5]  = "PK\003\004";

    int length = 0;
    FILE *file;

	if( file_name == NULL )
    {
	    return FALSE;
    }

	if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
	    return FALSE;
    }

	length = fread( number, 1, 4, file );
	fclose( file );

	if( length != 4 )
    {
	    return FALSE;
    }

    if( memcmp( number, magic, 4 ) == 0 )
    {
	    return TRUE;
    }

    return FALSE;
}

bool filter_tar( struct stat *stats, const char *file_name,
                 const char *argument )
{
	char number[5] = "\0";
	int length = 0;

	FILE *file;

	if( file_name == NULL || stats == NULL )
    {
	    return FALSE;
    }

	if( stats->st_size < 263 )
    {
	    return FALSE;
    }

	if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
	    return FALSE;
    }

    if( fseek( file, 257, SEEK_SET ) != 0 )
    {
        fclose( file );
        return FALSE;
	}

	length = fread( number, 1, 5, file );
	fclose( file );

	if( length != 5 )
    {
	    return FALSE;
    }

    if( memcmp( number, "ustar", 5 ) == 0 )
    {
	    return TRUE;
    }

    return FALSE;
}

bool filter_rpm( struct stat *stats, const char *file_name,
                 const char *argument )
{
	int length = 0;

	unsigned char number[5] = "\0\0\0\0";
    unsigned char magic[5]  = { 237, 171, 238, 219 };

	FILE *file;

	if( file_name == NULL )
    {
	    return FALSE;
    }

	if( ( file = osi_fopen( file_name, "r" ) ) == NULL )
    {
	    return FALSE;
    }

	length = fread( number, 1, 4, file );
	fclose( file );

	if( length != 4 )
    {
	    return FALSE;
    }

    if( memcmp( number, magic, 4 ) == 0 )
    {
	    return TRUE;
    }

    return FALSE;
}
