
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      filter.h
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#ifndef FILTER_H
#define FILTER_H

#include <sys/types.h>
#include <regex.h>

#include <pwd.h>
#include <grp.h>

#include "utilities.h"
#include "crypto_utilities.h"


#define MAX_FILTER_ARGUMENT_LENGTH       MAX_PATH_LEN

/* filter types */

/* urrently, this translates to: IncludeAll */

#define FILTER_TYPE_UNKNOWN              0 
#define FILTER_TYPE_STICKY               1
#define FILTER_TYPE_SUID                 2
#define FILTER_TYPE_SGID                 3
#define FILTER_TYPE_EXECUTABLE           4
#define FILTER_TYPE_PERL                 5
#define FILTER_TYPE_PYTHON               6
#define FILTER_TYPE_SCRIPT               7
#define FILTER_TYPE_GZIP                 8
#define FILTER_TYPE_ZIP                  9
#define FILTER_TYPE_TAR                 10
#define FILTER_TYPE_PGP                 11
#define FILTER_TYPE_RPM                 12

#define FILTER_TYPE_UID                 13
#define FILTER_TYPE_GID                 14
#define FILTER_TYPE_USER                15
#define FILTER_TYPE_GROUP               16
#define FILTER_TYPE_HEADER              17
#define FILTER_TYPE_FILE                18
#define FILTER_TYPE_SUFFIX              19
#define FILTER_TYPE_MD5                 20
#define FILTER_TYPE_SHA                 21
#define FILTER_TYPE_HAVAL               22
#define FILTER_TYPE_RIPEMD              23
#define FILTER_TYPE_PERMISSIONS         24



struct keywords
{
    char *word;
    int  type;
};

static struct keywords filter_keywords[] =
{
    { "sticky", FILTER_TYPE_STICKY },
    { "suid", FILTER_TYPE_SUID },
    { "sgid", FILTER_TYPE_SGID },
    { "executable", FILTER_TYPE_EXECUTABLE },
    { "perl", FILTER_TYPE_PERL },
    { "python", FILTER_TYPE_PYTHON },
    { "script", FILTER_TYPE_SCRIPT },
    { "gzip", FILTER_TYPE_GZIP },
    { "zip", FILTER_TYPE_ZIP },
    { "tar", FILTER_TYPE_TAR },
    { "pgp", FILTER_TYPE_PGP },
    { "rpm", FILTER_TYPE_RPM },
    { "uid", FILTER_TYPE_UID },
    { "gid", FILTER_TYPE_GID },
    { "user", FILTER_TYPE_USER },
    { "group", FILTER_TYPE_GROUP },
    { "header", FILTER_TYPE_HEADER },
    { "file", FILTER_TYPE_FILE },
    { "suffix", FILTER_TYPE_SUFFIX },
    { "md5", FILTER_TYPE_MD5 },
    { "sha", FILTER_TYPE_SHA },
    { "haval", FILTER_TYPE_HAVAL },
    { "ripemd", FILTER_TYPE_RIPEMD },
    { "permissions", FILTER_TYPE_PERMISSIONS },
    { NULL, 0 }
};


typedef struct filter
{
    int type;
    char argument[MAX_FILTER_ARGUMENT_LENGTH];

    bool (*perform)( struct stat *stats, const char *file_name,
                     const char *argument );

} filter;


int setup_filter_function( filter *action_filter, int type );
bool filter_is_dynamic( int filter_type );

int get_filter_type( char *line );
char * get_filter_value( char *line, char *value );

char *get_filter_name_from_type( unsigned int filter_type );


/* static filters. */

bool filter_sticky( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_suid( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_sgid( struct stat *stats, const char *file_name,
                  const char *argument );

bool filter_executable( struct stat *stats, const char *file_name,
                        const char *argument );

bool filter_perl( struct stat *stats, const char *file_name,
                  const char *argument );

bool filter_python( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_script( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_gzip( struct stat *stats, const char *file_name,
                  const char *argument );

bool filter_zip( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_tar( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_pgp( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_rpm( struct stat *stats, const char *file_name,
                 const char *argument );

/* dynamic filters with arguments. */

bool filter_uid( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_gid( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_user( struct stat *stats, const char *file_name,
                  const char *argument );

bool filter_group( struct stat *stats, const char *file_name,
                   const char *argument );

bool filter_header( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_file( struct stat *stats, const char *file_name,
                  const char *argument );

bool filter_suffix( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_md5( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_sha( struct stat *stats, const char *file_name,
                 const char *argument );

bool filter_haval( struct stat *stats, const char *file_name,
                   const char *argument );

bool filter_ripemd( struct stat *stats, const char *file_name,
                    const char *argument );

bool filter_permissions( struct stat *stats, const char *file_name,
                         const char *argument );


#endif
