
#include <gdbm.h>

#include "debug.h"
#include "database_handler.h"

#define MAX_BUFFER_SIZE 8192


bool module_gdbm_open( database_handler *db, int mode )
{
    GDBM_FILE gdbm_file = NULL;

    if( ( db == NULL ) || ( db->name == NULL ) )
    {
        return FALSE;
    }

    if( mode == DB_READ )
    {
        gdbm_file = gdbm_open( db->name, 0, GDBM_READER, 0, NULL );
    }

	else if( mode == DB_WRITE )
    {
        gdbm_file = gdbm_open( db->name, 0, GDBM_WRITER, 0600, NULL );
    }

    else if( mode == DB_NEW )
    {
        gdbm_file = gdbm_open( db->name, 0, GDBM_NEWDB, 0600, NULL );
    }

	else
    {
	    return FALSE;
    }

    if( gdbm_file == NULL )
    {
	    return FALSE;
    }

	db->state  = DB_STATE_OPEN;
	db->handle = gdbm_file;

    return TRUE;
}

bool module_gdbm_close( database_handler *db )
{
    if( db != NULL && db->state == DB_STATE_OPEN )
    {
	    gdbm_close( (GDBM_FILE)db->handle );
	    db->state = DB_STATE_CLOSED;

	    return TRUE;
	}

	return FALSE;
}


bool module_gdbm_store_record( database_handler *db, record *osi_record )
{
    datum store_key;
    datum store_data;

    int length        = 0;
    int string_length = 0;
    
    unsigned char buffer[MAX_BUFFER_SIZE] = "";
    unsigned char *index;

    int temp = 0;

    if( ( db == NULL ) || ( osi_record == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    store_key.dptr   = (char *)( osi_record->file_path );
    store_key.dsize  = strlen( store_key.dptr );

    /* sanity check. */

    if( store_key.dptr == NULL || store_key.dsize == 0 )
    {
        return FALSE;
    }

    /* first, we determine the length of the data we are to store. */

    length = strlen( osi_record->hash );
    length += strlen( osi_record->hash_algorithm );
    length += strlen( osi_record->filter_description );
    length += strlen( osi_record->block_name );

    length += 4;  // FOR NULLS.

    string_length = length;

    length += ( sizeof( osi_record->reason ) +
                sizeof( osi_record->attributes ) );

    length += sizeof( struct osi_stat );

    /* check buffer to see if it's big enough, then initialize it. */

    if( length >= sizeof( buffer ) )
    {
        return FALSE;
    }

    memset( buffer, 0, sizeof( buffer ) );

    /* before anything, we wrap the record. */

    wrap_record( osi_record );

    /* If a value is the empty string, it is represented by NULL. */
    /* first, copy in the hash string.                            */

    index = buffer;

    memcpy( index, osi_record->hash, strlen( osi_record->hash ) );
    index += ( strlen( osi_record->hash ) + 1 );

    /* hasl algorithm string. */

    memcpy( index, osi_record->hash_algorithm,
            strlen( osi_record->hash_algorithm ) );

    index += ( strlen( osi_record->hash_algorithm ) + 1 );

    /* filter description string. */

    memcpy( index, osi_record->filter_description,
            strlen( osi_record->filter_description) );

    index += ( strlen( osi_record->filter_description ) + 1 );

    /* name of the block this record belongs to. */

    memcpy( index, osi_record->block_name, strlen( osi_record->block_name ) );
    index += ( strlen( osi_record->block_name ) + 1 );

    /* now pack the binary data. */

    index = ( buffer + string_length );

    memcpy( index, &(osi_record->reason), sizeof( osi_record->reason ) );
    index += sizeof( osi_record->reason );

    memcpy( index, &(osi_record->attributes),
            sizeof( osi_record->attributes ) );

    index += sizeof( osi_record->attributes );

    memcpy( index, &(osi_record->stats), sizeof( struct osi_stat ) );

    /* now store the data into gdbm. */

    store_data.dptr  = buffer;
    store_data.dsize = length;

    if( gdbm_store( (GDBM_FILE)db->handle, store_key,
        store_data, GDBM_REPLACE ) == 0 )
    {
	    db->counter++;
	    return TRUE;
    }

    return FALSE;
}

bool module_gdbm_retrieve_record( database_handler *db, const char *path,
                                  record *osi_record )
{
    datum fetch_key;
    datum fetch_data;

    int length   = 0;
    osi_stat *temp_stats;

    char *buffer = NULL;
    char *temp   = NULL;

    if( ( db == NULL ) || ( osi_record == NULL ) || ( path == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    fetch_key.dptr  = (char *)path;
    fetch_key.dsize = strlen( path );

    fetch_data = gdbm_fetch( (GDBM_FILE)db->handle, fetch_key );

    if( fetch_data.dptr == NULL )
    {
        return FALSE;
    }

    buffer = (char *)fetch_data.dptr;

    /* place the file path into the new record, then we unpack */
    /* the string portion of the data.                         */

    safe_strlcpy( osi_record->file_path, path,
                  sizeof( osi_record->file_path ) );

    safe_strlcpy( osi_record->hash, buffer, sizeof( osi_record->hash ) );
    length = ( strlen( osi_record->hash ) + 1 );

    safe_strlcpy( osi_record->hash_algorithm, ( buffer + length ),
                  sizeof( osi_record->hash_algorithm ) );

    length += ( strlen( osi_record->hash_algorithm ) + 1 );

    safe_strlcpy( osi_record->filter_description, ( buffer + length ),
                  sizeof( osi_record->filter_description ) );

    length += ( strlen( osi_record->filter_description ) + 1 );
    
    safe_strlcpy( osi_record->block_name, ( buffer + length ),
                  sizeof( osi_record->block_name ) );

    length += ( strlen( osi_record->block_name ) + 1 );

    /* now unpack the reason and attributes. */

    memcpy( &( osi_record->reason ), ( buffer + length ),
            sizeof( osi_record->reason ) );

    memcpy( &( osi_record->attributes ),
            ( buffer + length + sizeof( osi_record->reason ) ),
            sizeof( osi_record->attributes ) );

    /* now unpack the stats. */

    memcpy( &( osi_record->stats ),
            ( buffer + length + sizeof( osi_record->reason ) +
                        sizeof( osi_record->attributes ) ),
            sizeof( osi_record->stats ) );

    unwrap_record( osi_record );

    osi_free( buffer );

    return TRUE;
}

bool module_gdbm_record_exists( database_handler *db, record *osi_record )
{
    datum fetch_key;

    if( db == NULL || osi_record == NULL || db->state != DB_STATE_OPEN )
    {
	    return FALSE;
    }

    fetch_key.dptr  = (char *)osi_record->file_path;
    fetch_key.dsize = strlen( osi_record->file_path );

    if( gdbm_exists( (GDBM_FILE)db->handle, fetch_key ) )
    {
	    return TRUE;
    }

    return FALSE;
}

bool module_gdbm_exists( struct database_handler *db )
{
	if( db != NULL )
    {
	    return file_exists( db->name );
    }

	return FALSE;
}

bool module_gdbm_store_header( database_handler *db, char *header )
{
	datum store_key;
	datum store_data;

    if( db == NULL || header == NULL || db->state != DB_STATE_OPEN )
    {
        return FALSE;
    }

    store_key.dptr   = HEADER_KEY;
    store_key.dsize  = strlen( HEADER_KEY );

    store_data.dptr  = header;
    store_data.dsize = strlen( header );

    if( gdbm_store( (GDBM_FILE)db->handle, store_key, store_data,
                    GDBM_REPLACE ) == 0 )
    {
	    return TRUE;
    }

    return FALSE;
}

char * module_gdbm_retrieve_header( database_handler *db )
{
    datum fetch_key;
    datum data;

    if( db == NULL || db->state != DB_STATE_OPEN )
    {
        return FALSE;
    }

    /* setup key and fetch for record. */

    fetch_key.dptr  = HEADER_KEY;
    fetch_key.dsize = strlen( HEADER_KEY );

    data = gdbm_fetch( (GDBM_FILE)db->handle, fetch_key );

    return (char *)data.dptr;
}

bool module_gdbm_retrieve_first_record( database_handler *db,
                                        record *osi_record )
{
    datum fetch_key;
    datum fetch_data;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) ||
        ( osi_record == NULL ) )
    {
        return FALSE;
    }

    fetch_key = gdbm_firstkey( (GDBM_FILE)db->handle );

    /* if we get the header, we need to skip over it   */
    /* because it won't align to the record structure. */

    if( fetch_key.dptr && strncmp( fetch_key.dptr, HEADER_KEY,
                                   strlen( HEADER_KEY ) ) == 0 )
    {
        char *temp_key = fetch_key.dptr;
        fetch_key = gdbm_nextkey( (GDBM_FILE)db->handle, fetch_key );
        osi_free( temp_key );
    }

    if( fetch_key.dptr == NULL )
    {
        return FALSE;
    }

    safe_strlcpy( osi_record->file_path, fetch_key.dptr,
                  sizeof( osi_record->file_path ) );

    osi_record->file_path[fetch_key.dsize] = '\0';
    osi_free( fetch_key.dptr );

    return module_gdbm_retrieve_record( db, osi_record->file_path, osi_record );
}

bool module_gdbm_retrieve_next_record( database_handler *db,
                                       record *osi_record )
{
	datum fetch_key;
	datum fetch_data;

	if( ( db == NULL ) || ( osi_record == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
	    return FALSE;
    }

    /* setup the fetch key to get the key for the next record. */

    fetch_key.dptr  = osi_record->file_path;
    fetch_key.dsize = strlen( osi_record->file_path );

    /* get the key for the next record, then get the */
    /* next record from the database.                */

    fetch_key = gdbm_nextkey( (GDBM_FILE)db->handle, fetch_key );

    /* if we get the header, we need to skip over it   */
    /* because it won't align to the record structure. */

    if( fetch_key.dptr && strncmp( fetch_key.dptr, HEADER_KEY,
                                   strlen( HEADER_KEY ) ) == 0 )
    {
        char *temp_key = fetch_key.dptr;
        fetch_key = gdbm_nextkey( (GDBM_FILE)db->handle, fetch_key );
        osi_free( temp_key );
    }

    if( fetch_key.dptr == NULL )
    {
        return FALSE;
    }

    safe_strlcpy( osi_record->file_path, fetch_key.dptr,
                  sizeof( osi_record->file_path ) );

    osi_record->file_path[fetch_key.dsize] = '\0';
    osi_free( fetch_key.dptr );

    return module_gdbm_retrieve_record( db, osi_record->file_path, osi_record );
}

bool module_gdbm_empty( database_handler *db )
{
	datum fetch_key;
    datum next_key;

	if( db == NULL || db->state != DB_STATE_OPEN )
    {
	    return FALSE;
    }

	fetch_key = gdbm_firstkey( (GDBM_FILE)db->handle );

	while( fetch_key.dptr )
	{
        gdbm_delete( (GDBM_FILE)db->handle, fetch_key );
	    next_key = gdbm_nextkey( (GDBM_FILE)db->handle, fetch_key );

	    osi_free( fetch_key.dptr );
	    fetch_key = next_key;
	}

    return TRUE;
}

char * module_gdbm_retrieve_last_error( database_handler *db )
{
    return gdbm_strerror( gdbm_errno );
}
