
#include <mysql.h>

#include "debug.h"
#include "database_handler.h"

/* static queries. */

#define CREATE_TABLE_RECORDS_QUERY "CREATE TABLE OSIRIS_RECORDS \
(ID INT NOT NULL AUTO_INCREMENT,\
FILE TEXT NOT NULL,\
HASH_ALGORITHM VARCHAR(20),\
HASH_VALUE VARCHAR(255),\
FILTER_DESCRIPTION TEXT,\
BLOCK_NAME TEXT,\
REASON BIGINT,\
ATTRIBUTES BIGINT,\
STAT MEDIUMBLOB NOT NULL,\
PRIMARY KEY(ID) )"

#define CREATE_TABLE_HEADER_QUERY "CREATE TABLE OSIRIS_HEADER \
(NAME VARCHAR(50) NOT NULL,\
HEADER MEDIUMTEXT NOT NULL)"

#define MAX_ERROR_SIZE 1024

/* locally used to store information related to mysql */
/* connection string.                                 */

typedef struct mysql_connect_info
{
    char host[1024];
    char user[4096];
    char pass[4096];
    char name[4096];

    int port;

} mysql_connect_info;


/* some utility functions we use, and the */
/* static counter used to iterate through */
/* the database rows.                     */

bool make_record_from_result_set( MYSQL_RES *results, record *osi_record );
bool get_mysql_info_from_name( mysql_connect_info *db_info, char *name );
void set_error_message( MYSQL *mysql );

static unsigned int record_count           = 1;
static char error_message[MAX_ERROR_SIZE]  = "";

bool module_mysql_open( database_handler *db, int mode )
{
    bool result = FALSE;
    char query[4096] = "";

    MYSQL *mysql_connection = NULL;
    mysql_connect_info db_info;

    memset( &db_info, 0, sizeof( db_info ) );

    if( db == NULL || db->name == NULL )
    {
        return FALSE;
    }

    if( get_mysql_info_from_name( &db_info, db->name ) == FALSE )
    {
        return FALSE;
    }

    if( ( mysql_connection = mysql_init( NULL ) ) == NULL )
    {
        return FALSE;
    }

    if( mode == DB_READ || mode == DB_WRITE )
    {
        if( mysql_real_connect( mysql_connection, db_info.host,
                            db_info.user, db_info.pass,
                            NULL, db_info.port, NULL, 0 ) == NULL )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
	        return FALSE;
	    }

	/* select the database. */

    	if( mysql_select_db( mysql_connection, db_info.name ) != 0 )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
            return FALSE;
        }
    }

    else if( mode == DB_NEW )
    {
        if( mysql_real_connect( mysql_connection, db_info.host,
   						db_info.user, db_info.pass,
     						NULL, db_info.port, NULL, 0 ) == NULL )
        {
            set_error_message( mysql_connection );
	        mysql_close( mysql_connection );
	        return FALSE;
	    }

        /* in case database already exists, we clobber it. */

        safe_strlcpy( query, "DROP DATABASE IF EXISTS ", sizeof( query ) );
        safe_strlcat( query, db_info.name, sizeof( query ) );

        mysql_query( mysql_connection, query );

	    /* now we create the database. */

        safe_strlcpy( query, "CREATE DATABASE ", sizeof( query ) );
        safe_strlcat( query, db_info.name, sizeof( query ) );

        if( mysql_query( mysql_connection, query ) != 0 )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
            return FALSE;
        }

	    /* select the database. */

    	if( mysql_select_db( mysql_connection, db_info.name ) != 0 )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
            return FALSE;
        }

        /* now we create the osiris table. */

        if( mysql_query( mysql_connection, CREATE_TABLE_RECORDS_QUERY ) != 0 )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
            return FALSE;
        }

        /* now we create the header table. */

        if( mysql_query( mysql_connection, CREATE_TABLE_HEADER_QUERY ) != 0 )
        {
            set_error_message( mysql_connection );
            mysql_close( mysql_connection );
            return FALSE;
        }
    }

    else
    {
        mysql_close( mysql_connection );
        return FALSE;
    }

    db->state  = DB_STATE_OPEN;
    db->handle = mysql_connection;

    return TRUE;
}

bool module_mysql_close( database_handler *db )
{
    if( db != NULL && db->state == DB_STATE_OPEN )
    {
	    db->state = DB_STATE_CLOSED;
	    mysql_close( (MYSQL *)db->handle );

	    return TRUE;
    }

    return FALSE;
}

bool module_mysql_store_record( database_handler *db, record *osi_record )
{
    int length      = 0;
    int struct_size = 0;

    unsigned char query[( (MAX_PATH_LEN*7) + sizeof(struct osi_stat))] = "";

    char *file;
    char *hash_algorithm;
    char *hash_value;
    char *filter_description;
    char *block_name;
    char reason[64];
    char attributes[64];
    char stats[ ( sizeof(struct osi_stat)*2 ) + 1];

    char *format = "INSERT INTO OSIRIS_RECORDS (file,hash_algorithm,hash_value,filter_description,block_name,reason,attributes,stat) VALUES ('%s','%s','%s','%s','%s',%s,%s,'%s')";

    if( ( db == NULL ) || ( osi_record == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    file               = osi_record->file_path;
    hash_algorithm     = osi_record->hash_algorithm;
    hash_value         = osi_record->hash;
    filter_description = osi_record->filter_description;
    block_name         = osi_record->block_name;

    safe_snprintf( reason, sizeof( reason ), "%d",
                   osi_record->reason );

    safe_snprintf( attributes, sizeof( attributes ), "%d",
                   osi_record->attributes );

    /* wrap before we escape. */
     
    wrap_record( osi_record );

    length = mysql_escape_string( stats, (char *)&osi_record->stats,
                                  sizeof( struct osi_stat ) );

    /* check for NULL values on the fields that matter. */

    if( strlen( hash_algorithm ) == 0 )
    {
        hash_algorithm = "";
    }

    if( strlen( hash_value ) == 0 )
    {
        hash_value = "";
    }

    if( strlen( filter_description ) == 0 )
    {
        filter_description = "";
    }
        
    if( strlen( block_name ) == 0 )
    {
        block_name = "";
    }

    /* construct length to allocate for. */

    length += ( strlen( format ) + strlen( file ) + strlen( hash_algorithm ) +
                strlen( hash_value ) + strlen( filter_description ) +
                strlen( block_name ) + strlen( reason ) +
                strlen( attributes ) + 1 );

    safe_snprintf( query, length, format, file, hash_algorithm, hash_value,
                   filter_description, block_name, reason, attributes, stats );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    db->counter++;
	    return TRUE;
    }

    set_error_message( (MYSQL *)db->handle );

    return FALSE;
}

bool module_mysql_retrieve_record( database_handler *db, const char *path,
                                   record *osi_record )
{
    char *format = "SELECT * FROM OSIRIS_RECORDS WHERE BINARY FILE = '%s' LIMIT 1";
    char query[MAX_PATH_LEN*2] = "";

    int length   = 0;
    bool result;

    MYSQL_RES *results;

    if( ( db == NULL ) || ( osi_record == NULL ) || ( path == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    safe_snprintf( query, sizeof( query ), format, path );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    if( ( results = mysql_store_result( (MYSQL *)db->handle ) ) != NULL )
	    {
            result = make_record_from_result_set( results, osi_record );
            mysql_free_result( results );
	    }
    }

    set_error_message( (MYSQL *)db->handle );
    return result;
}

bool module_mysql_record_exists( database_handler *db, record *osi_record )
{
    char *format = "SELECT * FROM OSIRIS_RECORDS WHERE BINARY FILE = '%s' LIMIT 1";
    char query[MAX_PATH_LEN*2] = "";

    bool found         = FALSE;
    MYSQL_RES *results;

    if( ( db == NULL ) || ( osi_record == NULL ) ||
        ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    safe_snprintf( query, sizeof( query ), format, osi_record->file_path );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    if( ( results = mysql_store_result( (MYSQL *)db->handle ) ) != NULL )
	    {
	        if( mysql_num_rows( results ) > 0 )
            {
	            found = TRUE;
            }

            mysql_free_result( results );
	    }
    }

    set_error_message( (MYSQL *)db->handle );
    return found;
}

bool module_mysql_exists( database_handler *db )
{
    MYSQL *mysql_connection = NULL;
    mysql_connect_info db_info;

    if( ( db == NULL ) || ( db->name == NULL ) )
    {
        return FALSE;
    }

    if( get_mysql_info_from_name( &db_info, db->name ) == FALSE )
    {
        return FALSE;
    }

    if( ( mysql_connection = mysql_init( NULL ) ) == NULL )
    {
	    return FALSE;
    }

    if( mysql_real_connect( mysql_connection, db_info.host,
        				    db_info.user, db_info.pass,
        				    NULL, db_info.port, NULL, 0 ) == NULL )
    {
        set_error_message( mysql_connection );
        mysql_close( mysql_connection );

	    return FALSE;
    }

    /* select the database. */

    if( mysql_select_db( mysql_connection, db_info.name ) != 0 )
    {
        set_error_message( mysql_connection );
        mysql_close( mysql_connection );

        return FALSE;
    }

    return TRUE;
}

bool module_mysql_store_header( database_handler *db, char *header )
{
    int length   = 0;

    char *query  = NULL;
    char *format = "INSERT INTO OSIRIS_HEADER (NAME,HEADER) VALUES('%s','%s')";

    if( ( db == NULL ) || ( header == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    length = ( strlen( HEADER_KEY ) + strlen( header ) + strlen( format ) + 1 );

    query = osi_malloc( length );
    safe_snprintf( query, length, format, HEADER_KEY, header );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    osi_free( query );
	    return TRUE;
    }

    set_error_message( (MYSQL *)db->handle );
    osi_free( query );

    return FALSE;
}

char * module_mysql_retrieve_header( database_handler *db )
{
	int length   = 0;
	char *format = "SELECT * FROM OSIRIS_HEADER WHERE NAME = '%s'";

	char *query  = NULL;
	char *header = NULL;

	MYSQL_RES *results;
	MYSQL_ROW row;

    if( db == NULL || db->state != DB_STATE_OPEN )
    {
        return FALSE;
    }

    length = ( strlen( format ) + strlen( HEADER_KEY ) + 1 );

    query  = osi_malloc( length );
    safe_snprintf( query, length, format, HEADER_KEY );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
		if( ( results = mysql_store_result( (MYSQL *)db->handle ) ) != NULL )
		{
			if( ( row = mysql_fetch_row( results ) ) != NULL )
			{
			    length = strlen( row[1] );

			    if( length > 0 )
			    {
			        header = osi_malloc( length + 1 );
			        safe_strlcpy( header, row[1], ( length + 1 ) );
			    }
			}
		}
    }

    else
    {
        set_error_message( (MYSQL *)db->handle );
    }

    mysql_free_result( results );
    osi_free( query );

    return header;
}

bool module_mysql_retrieve_first_record( database_handler *db,
                                         record *osi_record )
{
    char *query        = "SELECT * FROM OSIRIS_RECORDS WHERE ID = '1'";
    bool result;

    MYSQL_RES *results = NULL;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) ||
        ( osi_record == NULL ) )
    {
        return FALSE;
    }

    record_count = 2;

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    if( ( results = mysql_store_result( (MYSQL *)db->handle ) ) != NULL )
	    {
            result = make_record_from_result_set( results, osi_record );
            mysql_free_result( results );
        }
    }

    else
    {
        set_error_message( (MYSQL *)db->handle );
    }

    return result;
}

bool module_mysql_retrieve_next_record( database_handler *db,
                                        record *osi_record )
{
    char query[100]    = "";
    bool result = FALSE;

    MYSQL_RES *results = NULL;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return FALSE;
    }

    safe_snprintf( query, sizeof( query ),
                   "SELECT * FROM OSIRIS_RECORDS WHERE ID = '%d' LIMIT 1",
                   record_count );

    if( mysql_query( (MYSQL *)db->handle, query ) == 0 )
    {
	    if( ( results = mysql_store_result( (MYSQL *)db->handle ) ) != NULL )
	    {
            result = make_record_from_result_set( results, osi_record );
            mysql_free_result( results );
        }
    }

    else
    {
        set_error_message( (MYSQL *)db->handle );
    }

    record_count++;
    return result;
}

bool module_mysql_empty( database_handler *db )
{
    /* we don't empty here, we clobber on creation. */

	return TRUE;
}


char * module_mysql_retrieve_last_error( database_handler *db )
{
    return error_message;
}

void set_error_message( MYSQL *mysql )
{
    if( mysql != NULL && mysql_errno( mysql ) )
    {
        safe_strlcpy( error_message, mysql_error( mysql ), MAX_ERROR_SIZE );
    }
}

/* get all the mysql connection information out of the  */
/* name string.  the db and username are required, but  */
/* the @host:port portion is optional.  if not there    */
/* we set it to localhost:0 to allow mysql to use the   */
/* default mysql port on the local machine.             */
/*                                                      */
/*  format:      name:user:pass@host:port               */

bool get_mysql_info_from_name( mysql_connect_info *db_info, char *name )
{
    char *user = NULL;
    char *pass = NULL;
    char *host = NULL;
    char *port = NULL;

    char buffer[4096] = "";

    if( db_info == NULL || name == NULL )
    {
        return FALSE;
    }

    /* setup default host information. */

    safe_strlcpy( db_info->host, "localhost", sizeof( db_info->host ) );
    db_info->port = 0;

    /* get the database name. */

    if( ( user = get_string_up_to_token( name, buffer, ':' ) ) == NULL )
    {
        return FALSE;
    }

    safe_strlcpy( db_info->name, buffer, sizeof( db_info->name ) );

    /* get the username. it might be last argument, so find out */
    /* if there is a seperator character.                       */

    if( strchr( user, ':' ) )
    {
        if( ( pass = get_string_up_to_token( user, buffer, ':' ) ) == NULL )
        {
            return FALSE;
        }

        safe_strlcpy( db_info->user, buffer, sizeof( db_info->user ) );
    }

    /* no more arguments, get username and bail. */

    else
    {
        safe_strlcpy( db_info->user, user, sizeof( db_info->user ) );
        return TRUE;
    }

    /* get the password. if we don't have a host section, then */
    /* we get the password and stop, otherwise, we continue.   */

    if( strchr( pass, '@' ) )
    {
        if( ( host = get_string_up_to_token( pass, buffer, '@' ) ) == NULL )
        {
            return FALSE;
        }

        safe_strlcpy( db_info->pass, buffer, sizeof( db_info->pass ) );
    }

    /* no more arguments, get password and bail. */

    else
    {
        safe_strlcpy( db_info->pass, pass, sizeof( db_info->pass ) );
        return TRUE;
    }

    /* get the hostname, and optional port number, find out if   */
    /* we have only the hostname, or both the hostname and port. */

    if( strchr( host, ':' ) )
    {
        if( ( port = get_string_up_to_token( host, buffer, ':' ) ) == NULL )
        {
            return FALSE;
        }

        safe_strlcpy( db_info->host, buffer, sizeof( db_info->host ) );
    }

    /* no more arguments, just get hostname and bail. */

    else
    {
        safe_strlcpy( db_info->host, host, sizeof( db_info->host ) );
        return TRUE;
    }

    /* get port argument. */
  
    if( port != NULL )
    {
        db_info->port = atoi( port );
    }

    else
    {
        return FALSE;
    }

    return TRUE;
}

bool make_record_from_result_set( MYSQL_RES *results, record *osi_record )
{
    MYSQL_ROW row;
    unsigned long *lengths;

    if( ( results == NULL ) || ( osi_record == NULL ) )
    {
        return FALSE;
    }

    /* verify we have results. */

    if( ( row = mysql_fetch_row( results ) ) == NULL )
    {
        return FALSE;
    }

    /* verify we have the known number of fields. */

    if( mysql_num_fields( results ) != 9 )
    {
        return FALSE;
    }

    /* create the record structure here. */

    if( row[1] != NULL )
    {
        safe_strlcpy( osi_record->file_path, row[1],
                      sizeof( osi_record->file_path ) );
    }

    if( row[2] != NULL )
    {
        safe_strlcpy( osi_record->hash_algorithm, row[2],
                      sizeof( osi_record->hash_algorithm ) );
    }

    if( row[3] != NULL )
    {
        safe_strlcpy( osi_record->hash, row[3], sizeof( osi_record->hash ) );
    }

    if( row[4] != NULL )
    {
        safe_strlcpy( osi_record->filter_description, row[4],
                      sizeof( osi_record->filter_description ) );
    }
        
    if( row[5] != NULL )
    {
        safe_strlcpy( osi_record->block_name, row[5],
                      sizeof( osi_record->block_name ) );
    }

    if( row[8] != NULL )
    {
        lengths = mysql_fetch_lengths( results );

        if( lengths != NULL && lengths[8] > 0 )
        {
	        memcpy( &(osi_record->stats), row[8], lengths[8] );
            unwrap_record( osi_record );
        }
    }

    if( row[6] != NULL )
    {
        osi_record->reason = atol( row[6] );
    }

    if( row[7] != NULL )
    {
        osi_record->attributes = atol( row[7] );
    }

    return TRUE;
}

