
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      osiris.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#include "osiris.h"
#include "config.h"


/******************************************************************************
**
**    Function: main
**
**    Purpose:  system driver, hopefully, it returns 0.
**
******************************************************************************/

int main( int argument_count, char *argument_list[] )
{
    initialize( argument_count, argument_list );
    process();
    halt();

    return 0;
}

/*****************************************************************************
**
**    Function: initialize
**
**    Purpose:  create a config module, allocate all of our global
**              data.  parse the config data, then open any required
**              resources such as output files, and/or database(s).
**
******************************************************************************/

void initialize( int argument_count, char *argument_list[] )
{
    OSI_DEBUG( __POS__,  DEBUG_OSIRIS, "initializing osiris." );

    /* lower our priority as low as possible, then setup */
    /* our signal handlers.                              */

#ifndef WIN32

#ifdef HAVE_SETPRIORITY
    setpriority( PRIO_PROCESS, 0, LOW_PROCESS_PRIORITY );
#endif
    signal( SIGINT, signal_handler );
    signal( SIGTERM, signal_handler );

#endif

    /* create the config module and setup  */
    /* our global block structure with     */
    /* the default values.                 */

    config_module      = osi_config_new( NULL );

    directory_list     = string_list_new();
    bad_directory_list = string_list_new();
    database_list      = string_list_new();
    bad_database_list  = string_list_new();
    db_handles_list    = list_new();

    initialize_global_block();

    /* if a config file was specified on the command line */
    /* get the value and place into config module.        */

    get_configuration_file_name( argument_count, argument_list );

    /* if we are using a config file, parse it.  we have to */
    /* be carefule of precedence here.  we read the globals */
    /* first, then parse the command line to overide the    */
    /* globals, then we read the other blocks so they take  */
    /* on the defaults from the global block after it has   */
    /* been overidden by the command line.                  */

    if( strlen( config_module->file_name ) > 0 )
    {
        verify_config_file();

        osi_config_parse_globals( config_module );
        verify_config_state();

        parse_arguments( argument_count, argument_list );

        osi_config_parse_non_globals( config_module );
        verify_config_state();
    }

    /* no config file specified on the command line so */
    /* just parse the command line arguments.          */

    else
    {
        parse_arguments( argument_count, argument_list );

        /* if there were no directories specified */
        /* on the command line, add the current   */
        /* directory to the directory list.       */

        if( directory_list->size == 0 )
        {
            string_list_add_item( directory_list, DEFAULT_GLOBAL_DIRECTORY );
        }
    }

    /* if we have any directories specified on the   */
    /* command line we go through the list and       */
    /* expand each directory, then we validate each  */
    /* directory and print warnings for invalid      */
    /* directories and prompt for continuation, then */
    /* we create blocks for each directory.          */

    expand_directory_list();
    verify_directory_list();

    /* create directory blocks and try to open all   */
    /* of the databases.                             */

     create_blocks_for_directory_list();
     open_config_db_files();
}

/******************************************************************************
**
**    Function: halt
**
**    Purpose:  close all resources such as output files or database(s)
**              and free up all of our global data initialized at start.
**
******************************************************************************/

void halt()
{
    /* try to close all db file handles. */

    OSI_DEBUG( __POS__,  DEBUG_OSIRIS, "shutting down osiris." );
    close_config_db_files();

    /* destroy config file modlue and */
    /* the directory list.            */

    osi_config_destroy( config_module );

    string_list_destroy( directory_list );
    string_list_destroy( database_list );
    string_list_destroy( bad_directory_list );
    string_list_destroy( bad_database_list );

    destroy_database_handlers();
    list_destroy( db_handles_list );

    OSI_DEBUG( __POS__,  DEBUG_OSIRIS, "halt complete." );
}

/******************************************************************************
**
**    Function: signal_handler
**
**    Purpose:  intercept common fatal signals and shutdown cleanly.
**
******************************************************************************/

void signal_handler( int signal )
{
    switch( signal )
    {
        case SIGINT:
        case SIGTERM:

            osi_log_message(
    "osiris: received fatal signal, shutting down in a safe manner." );
            osi_log_message(
    "warning: all osiris databases created this session will be invalid." );

            halt();
            exit(0);

            break;

        default:
        break;
    }
}

/******************************************************************************
**
**    Function: get_configuration_file_name
**
**    Purpose:  used as a pre-scan of the CLA to try to get
**              the config file name before any parsing is done.
**
*****************************************************************************/

bool get_configuration_file_name( int argument_count, char *argument_list[] )
{
    int index;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
   "doing a pre-search of command line arguments, looking for config file..." );

    if( argument_count == 0 || argument_list == NULL )
    {
        return FALSE;
    }

    for( index = 0; index < argument_count; index++ )
    {
        if( strcmp( argument_list[index], "-f" ) == 0 &&
          ( index < ( argument_count - 1 ) ) )
        {
            index++;
            safe_strlcpy( config_module->file_name, argument_list[index],
                          sizeof( config_module->file_name ) );

            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                      "found \"%s\"", argument_list[index] );

            return TRUE;
        }
    }

    OSI_DEBUG( __POS__,  DEBUG_OSIRIS, "no config file found." );

    return FALSE;
}

/******************************************************************************
**
**    Function: parse_arguments
**
**    Purpose:  master routine to read in all of the CLAs and
**              take action appropriatly.
**
******************************************************************************/

void parse_arguments( int argument_count, char *argument_list[] )
{
    int count;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
               "parsing command line arguments (%d)", ( argument_count - 1 ) );

    for( count = 1; count < argument_count; count++ )
    {
        /* get directory paths. */

        if( strcmp( argument_list[count], "-i" ) == 0 )
        {
            OSI_DEBUG( __POS__, DEBUG_OSIRIS, "parsing directory list." );

            /* skip to first path. */

            count++;

            while( count < argument_count  )
            {
                /* we've gone too far, move back one argument. */
                /* and break out of the loop.                  */

                if( argument_list[count][0] == '-' )
                {
                    count--;
                    break;
                }

                OSI_DEBUG( __POS__,DEBUG_OSIRIS,
                           "read directory \"%s\"", argument_list[count] );

                string_list_add_item( directory_list, argument_list[count++] );
            }
        }

        else if( ( strcmp( argument_list[count], "-h" ) == 0 ) ||
                   strcmp( argument_list[count], "--help" ) == 0 )
        {
            print_usage();
            halt();
            exit(0);
        }

        else if( ( strcmp( argument_list[count], "-v" ) == 0 ) ||
                   strcmp( argument_list[count], "--version" ) == 0 )
        {
            print_program_info();
            halt();
            exit(0);
        }

        else if( strcmp( argument_list[count], "--follow-links" ) == 0 ||
                 strcmp( argument_list[count], "-l" ) == 0 )
        {
            global_block->follow_links = TRUE;
        }

        else if( strcmp( argument_list[count], "--quiet" ) == 0 ||
                 strcmp( argument_list[count], "-q" ) == 0 )
        {
            global_block->verbose = FALSE;
        }

        else if( strcmp( argument_list[count], "--shallow" ) == 0 ||
                 strcmp( argument_list[count], "-s" ) == 0 )
        {
            global_block->recursive = FALSE;
        }

        else if( strcmp( argument_list[count], "--continue" ) == 0 ||
                 strcmp( argument_list[count], "-c" ) == 0 )
        {
            global_block->prompt = FALSE;
        }

        else if( strcmp( argument_list[count], "--ignore-errors" ) == 0 ||
                 strcmp( argument_list[count], "-n" ) == 0 )
        {
            global_block->show_errors = FALSE;
        }

        else if( ( strcmp( argument_list[count], "--config-file" ) == 0 ||
                   strcmp( argument_list[count], "-f" ) == 0 )
                 && ( count < ( argument_count - 1 ) ) )
        {
            count++;
            safe_strlcpy( config_module->file_name,
                          argument_list[count],
                          sizeof( config_module->file_name ) -1 );
        }

        else if( ( strcmp( argument_list[count], "--test-config" ) == 0 ||
                   strcmp( argument_list[count], "-t" ) == 0 )
                 && ( count < ( argument_count - 1 ) ) )
        {
            count++;
            safe_strlcpy( config_module->file_name,
                          argument_list[count],
                          sizeof( config_module->file_name ) -1 );

            if( argument_count != 3 )
            {
                print_usage();
                halt();
                exit(0);
            }

            verify_config_file();
            osi_config_verify( config_module );
            verify_config_state();

            osi_log_message( "config file: %s is valid.",
                             config_module->file_name );

            halt();
            exit(0);
        }

        else if( strcmp( argument_list[count], "-o" ) == 0 &&
               ( count < ( argument_count - 1 ) ) )
        {
            count++;
            safe_strlcpy( global_block->database,
                          argument_list[count],
                          sizeof( global_block->database ) - 1 );
        }

        else if( ( strcmp( argument_list[count], "--hash" ) == 0 ||
                   strcmp( argument_list[count], "-a" ) == 0 )
              && ( count < ( argument_count - 1 ) ) )
        {
            count++;

            if( strncasecmp( argument_list[count], "sha", 4 ) == 0 )
            {
                global_block->hash = HASH_SHA;
            }

            else if( strncasecmp( argument_list[count], "haval", 5 ) == 0 )
            {
                global_block->hash = HASH_HAVAL;
            }

            else if( strncasecmp( argument_list[count], "md5", 3 ) == 0 )
            {
                global_block->hash = HASH_MD5;
            }

            else if( strncasecmp( argument_list[count], "ripemd", 6 ) == 0 )
            {
                global_block->hash = HASH_RIPEMD;
            }

            else
            {
                print_usage();
                halt();
                exit(0);
            }
        }

        else
        {
            print_usage();
            halt();
            exit(0);
        }

    } /* end for loop. */
}

/******************************************************************************
**
**    Function: void print_usage
**
**    Purpose:  prints usage statement and all command line arguments
**
*******************************************************************************/

void print_usage()
{

    fprintf( stdout,"\
\n\
usage: osiris [options ...] | -t <path>\n\
\n\
  options:\n\
\n\
     -f, --config-file <path>  path to configuration file to use.\n\
     -a, --hash <algorithm>    checksum algorithm to use.\n\
                               ( md5, sha, haval, ripemd ).\n\
\n\
     -q, --quiet               do not print file logging activity.\n\
     -c, --continue            continue without stopping for any warnings.\n\
     -n, --ignore-errors       do not print any errors encountered.\n\
     -s, --shallow             shallow mode, do not enter subdirectories.\n\
     -l, --follow-links        follow all symbolic links.\n\
     -o <db>                   database export path ( e.g. /var/log/scan.osi ).\n\
     -i <path> ...             directory or directories to scan.\n\
     -t, --test <path>         perform a syntax test on config file.\n\
     -v, --version             print program version information.\n\
     -h, --help                prints this usage statement.\n\
\n\
\n\
 all options apply to the global configuration.\n\
 all directories specified on the command line use the global configuration.\n" );

    print_program_info();
}

/******************************************************************************
**
**    Function: print_summary
**
**    Purpose:  at the end of a database run, print out a
**              basic report showing all of the counters and stats.
**
******************************************************************************/

void print_summary()
{
    char *config_name = config_module->file_name ;

    if( config_name == NULL || strlen( config_name ) == 0 )
    {
        config_name = "none, using command line configuration.";
    }

    osi_log_message( "" );
    osi_log_message( "---------------[ osiris ]-----------------" );
    osi_log_message( "" );

    osi_log_message( " start: %s", start_time );
    osi_log_message( " end:   %s", end_time );

    osi_log_message( "" );
    osi_log_message( " config: %s", config_name );
    osi_log_message( "" );

    osi_log_message( " files logged:               %d", files_logged_count );
    osi_log_message( " files encountered:          %d", files_encountered_count );

    osi_log_message( "" );

    osi_log_message( " symbolic links followed:    %d", symlinks_followed_count );
    osi_log_message( " symbolic links encountered: %d", symlinks_encountered_count );

    osi_log_message( "" );

    osi_log_message( " files/links unreadable:     %d", files_unreadable_count );
    osi_log_message( " directories unreadable:     %d", directories_unreadable_count );
    osi_log_message( " database errors:            %d", database_errors_count );

    print_program_info();
}

/******************************************************************************
**
**    Function: void print_program_info
**
**    Purpose:  prints program name, version, authors, etc.
**
******************************************************************************/

void print_program_info()
{
    osi_log_message( "" );
    osi_log_message( "-------------------------------------------" );

    osi_log_message( "   %s version %s [%s]",
                     PROGRAM_NAME, PROGRAM_VERSION, DB_MODULE );

    osi_log_message( "   osiris.shmoo.com              (c) 2003   " );
    osi_log_message( "-------------------------------------------" );

    osi_log_message( "" );
}

/******************************************************************************
**
**    Function: open_config_db_files
**
**    Purpose:  try to open up all of the databases including
**              those specified on the command line.  warn user
**              in case any failed to open.
**
******************************************************************************/

void open_config_db_files()
{
    int index;
    list blocks;
    database_handler *db;

    /* make sure all the blocks have db pointers    */
    /* pointing to read db handles, some may be     */
    /* sharing the same db, this is all setup here. */

    create_databases_for_blocks();

    /* now we loop through the blocks, opening the  */
    /* databases, if the db is already open, we     */
    /* simply igore it.                             */

    blocks = list_get_first_node( config_module->blocks );

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "opening all unique databases." );

    while( blocks != NULL )
    {
        current_block = blocks->data;

        if( current_block != NULL )
        {
            db = current_block->db;
        }

        else
        {
			break;
        }

        if( db && db->state == DB_STATE_CLOSED )
        {
            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                       "thinking about opening database: %s",
                       db->name );

            /* check for db existance, if so, prompt for overwrite */

            if( current_block->prompt && db->exists( db ) )
            {
                if( osi_continue_prompt(
                    "warning: database: %s already exists,\noverwrite?",
                     db->name ) == FALSE )
                {
                    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                               "not clobbering database: %s",
                                db->name );
                    halt();
                    exit(0);
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "clobbering database: %s", db->name );
            }

            /* if we fail, we add to the bad list. */

            if( db->open( db, DB_NEW ) == FALSE )
            {
                if( string_list_contains_item( bad_database_list,
                                               db->name ) == FALSE )
                {
                    string_list_add_item( bad_database_list, db->name );
                }
			}

			else
			{

                OSI_DEBUG( __POS__, DEBUG_OSIRIS, "opened db: %s.", db->name );

		        /* make sure we clobber this database so we have some */
		        /* confidence of integrity.                           */

                db->empty( db );
		    }

            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                       "database: %s open/close state: %d",
                        db->name, db->state );
        }

        blocks = blocks->next;
    }

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "scanning the bad database list." );

    /* now prompt for continuation in the case of any */
    /* unopened databases.                            */

    if( bad_database_list->size > 0 )
    {
        print_database_error( db );

        osi_log_message(
            "warning: the following databases could not be opened:" );

        osi_log_message( "" );

        for( index = 0; index < bad_database_list->size; index++ )
        {
            osi_log_message( bad_database_list->list[index] );
        }

        osi_log_message( "" );

        if( global_block->prompt &&
            osi_continue_prompt( "continue?" ) == FALSE )
        {
            halt();
            exit(0);
        }
    }

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "done opening config db files." );
}


/******************************************************************************
**
**    Function: close_config_db_files
**
**    Purpose:  loop through all of the open databases and
**              close each one.
**
******************************************************************************/

void close_config_db_files()
{
    list blocks;
    database_handler *db;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "closing all database file(s)." );

    blocks = list_get_first_node( config_module->blocks );

    while( blocks != NULL )
    {
        current_block = blocks->data;
        db = current_block->db;

        if( db )
        {
            db->close( db );
        }

        blocks = blocks->next;
    }
}

/******************************************************************************
**
**    Function: process
**
**    Purpose:  master routine that handles the traversal of the
**              two file system processing each block.
**
******************************************************************************/

void process()
{
    list head;
    time_t current_time;

    /* note the starting time. */

    current_time = time( NULL );

    safe_strlcpy( start_time, (char *)ctime( &current_time ),
                  sizeof( start_time ) );

    start_time[strlen( start_time ) - 1] = '\0';

    /* get first node from the block list. */

    head = list_get_first_node( config_module->blocks );
    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
        "\n----------- [ starting osiris traversal process ] -----------\n" );

    /* skip over the global node. */

    head = head->next;

    /* loop through entire block list processing each one. */

    while( head != NULL )
    {
        current_block = head->data;
        process_block( current_block );
        head = head->next;
    }

    /* note the ending timestamp. */

    current_time = time( NULL );

    safe_strlcpy( end_time, (char *)ctime( &current_time ),
                  sizeof( end_time ) );

    end_time[strlen( end_time ) - 1] = '\0';

    /* store database headers, print counters and results. */

    store_database_headers();
    print_summary();
}

/******************************************************************************
**
**    Function: process_block
**
**    Purpose:  not much.  right now it's basically empty but it
**              may serve some extra use in the future so it stays.
**
******************************************************************************/

void process_block( block *config_block )
{
    if( config_block != NULL )
    {
        struct stat file_stats;
       
        if( stat( config_block->directory, &file_stats ) == 0 )
        {
            files_encountered_count++;
            process_file( config_block->directory, &file_stats );
        }

        process_directory( config_block->directory );
    }
}

/******************************************************************************
**
**    Function: process_directory
**
**    Purpose:  recursive function that processes the file
**              system according to the current block.
**
******************************************************************************/

void process_directory( const char *directory_path )
{
    struct dirent *file;
    DIR *directory;

    char file_path[MAXPATHLEN] = "";
    struct stat file_stats;

    int length;
    bool result;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
        "\n----------- directory \"%s\" -----------\n", directory_path );

    /* yes this is nasty, but it's temporary until I can */
    /* find a better solution to this issue on darwin.   */

#ifdef PLATFORM_DARWIN
    if( strcmp( directory_path, "/dev/fd" ) == 0 )
    {
        return;
    }
#endif

    if( ( directory = opendir( directory_path ) ) == NULL )
    {
        OSI_DEBUG( __POS__, DEBUG_OSIRIS, "unable to open directory \"%s\"",
                   directory_path );

        if( current_block->show_errors )
        {
            osi_log_message(
                "==> error: unable to read directory %s, skipping.",
                directory_path );
        }

        directories_unreadable_count++;
        return;
    }

    while( ( file = readdir( directory ) ) )
    {
        /* first, ignore the file if it is either the current directory */
        /* or the previous directory, this will screw up recursion.     */

        if( strcmp( file->d_name, "." ) == 0 )
        {
            continue;
        }

        if( strcmp( file->d_name, ".." ) == 0 )
        {
            continue;
        }

        files_encountered_count++;

        /* create the full path to the file. */

        file_path[0] = '\0';

        /* don't add an extra '/' if one is already there. */

        if( strcmp( directory_path, "/" ) != 0 )
        {
            safe_strlcat( file_path, directory_path, sizeof( file_path ) );
        }

        safe_strlcat( file_path, "/", sizeof( file_path ) );
        safe_strlcat( file_path, file->d_name, sizeof( file_path ) );

        /* lstat the file. */

        OSI_DEBUG( __POS__, DEBUG_OSIRIS, "performing lstat() on file: \"%s\"",
                   file_path );

        if( lstat( file_path, &file_stats ) != 0 )
        {
            OSI_DEBUG( __POS__, DEBUG_OSIRIS, "lstat failed!" );

            if( current_block->show_errors )
            {
                osi_log_message(
            "==> error: unable to obtain information about file: %s, skipping",
            file_path );
            }

            files_unreadable_count++;
            continue;
        }

        /* check for symbolic links first. if it's a link */
        /* we check to see if we are following links and  */
        /* then we stat the file the link points to.      */

        if( ( result = S_ISLNK( file_stats.st_mode ) ) )
        {
            symlinks_encountered_count++;
        }

        if( result && current_block->follow_links )
        {
	        OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                       "found link, logging link file \"%s\"",
                       file_path );

	        process_file( file_path, &file_stats );

            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                      "following link: \"%s\"", file_path );

            if( stat( file_path, &file_stats ) != 0 )
            {
                OSI_DEBUG( __POS__, DEBUG_OSIRIS, "unable to stat link." );

                if( current_block->show_errors )
                {
                    osi_log_message(
            "==> error: unable to obtain information about link: %s, skipping",
                    file_path );
                }

                files_unreadable_count++;
                continue;
            }

            /* get the path the link points to. */

            length = readlink( file_path, file_path, sizeof( file_path ) );

            if( length > 0 )
            {
                file_path[length] = '\0';
            }

            symlinks_followed_count++;
            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                       "link points to \"%s\"", file_path );
        }

        /* process the file, this includes directories. */

        process_file( file_path, &file_stats );

        /* if it's a directory, we only process it if: */
        /*                                             */
        /*  1) current block is recursive.             */
        /*  2) it's not on the no_enter list.          */
        /*  3) there is no block for it.               */

        if( S_ISDIR( file_stats.st_mode ) )
        {
            if( current_block->recursive )
            {
                if( string_list_contains_item( current_block->no_entry_list,
                                               file_path ) )
                {
                    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                            "directory is in no entry list, skipping." );
                    continue;
                }

                if( string_list_contains_item( config_module->block_names,
                                               file_path ) )
                {
                    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                                "directory has its own block, skipping." );
                    continue;
                }

                process_directory( file_path );
            }
        }

    } /* end while */

    closedir( directory );
}


/******************************************************************************
**
**    Function: process_file
**
**    Purpose:  examine the file and loop through the actions for this
**              block until one catches this file.  if caught, we log
**              it, otherwise, we skip over it completely.
**
******************************************************************************/

void process_file( const char *file_path, struct stat *file_stats )
{
    list action_node;
    action *current_action;

    char *filter_argument;
    bool filter_result;

    record file_record;

    record file_record_fork;   /* possible resource fork. */
    struct stat fork_stats;

    /* it's important to initialize these, this way */
    /* we can ignore directories for hashes, and we */
    /* it makes it possible to keep them off of the */
    /* duplicate checksum lists.                    */

    file_record.hash[0]                = '\0';
    file_record.hash_algorithm[0]      = '\0';
    file_record.block_name[0]          = '\0';

#ifdef PLATFORM_DARWIN
    file_record_fork.hash[0]           = '\0';
    file_record_fork.hash_algorithm[0] = '\0';
    file_record_fork.block_name[0]     = '\0';
#endif

    if( file_stats == NULL )
    {
        OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                  "null stats, not processing file: \"%s\"", file_path );
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "processing file: \"%s\"", file_path );
    action_node = current_block->actions->next;

    /* if there are no actions in the list, then */
    /* we try to get the global action.          */

    if( action_node == NULL )
    {
        OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                   "no actions for this block, using global action." );

        action_node = global_block->actions->next;
    }

    safe_strlcpy( file_record.file_path, file_path,
                  sizeof( file_record.file_path ) );

#ifdef PLATFORM_DARWIN
    safe_strlcpy( file_record_fork.file_path, file_path,
                  sizeof( file_record_fork.file_path ) );

    safe_strlcat( file_record_fork.file_path, "/rsrc",
                  sizeof( file_record_fork.file_path ) );
#endif

    while( action_node != NULL )
    {
        /* get an action, store action attributes  */
        /* and the filter type.                    */

        current_action              = action_node->data;

        file_record.attributes      = current_action->attributes;
        file_record.reason          = current_action->action_filter.type;

        /* construct the filter description. */

        filter_argument = get_filter_name_from_type( file_record.reason );

        if( filter_argument != NULL )
        {
            safe_strlcpy( file_record.filter_description,
                          filter_argument,
                          sizeof( file_record.filter_description )  );
        }

        /* there are 4 possibliities for actions:  */
        /*                                         */
        /* 1) Include filter                       */
        /* 2) Exclude filter                       */
        /* 3) IncludeAll                           */
        /* 4) ExcludeAll                           */

        switch( current_action->type )
        {
            case ACTION_TYPE_INCLUDE:

                filter_argument = current_action->action_filter.argument;
                filter_result   = current_action->action_filter.perform(
                                    file_stats, file_path, filter_argument );

                if( filter_result )
                {
                    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                        "file caught by \"Include\" action." );

                    safe_strlcat( file_record.filter_description, "(",
                                  sizeof( file_record.filter_description ) );

                    safe_strlcat( file_record.filter_description,
                                  filter_argument,
                                  sizeof( file_record.filter_description ) );

                    safe_strlcat( file_record.filter_description, ")",
                                  sizeof( file_record.filter_description ) );

#ifdef PLATFORM_DARWIN
                   /* check for resource forks, darwin only. */
                   /* only for non-directory files.          */

                   if( ( !S_ISDIR( file_record.stats.permissions ) ) &&
                       stat( file_record_fork.file_path,
                             &fork_stats ) == 0 )
                    {
                        if( fork_stats.st_size > 0 )
                        {
                            files_encountered_count++;

                          file_record_fork.attributes = file_record.attributes;
                          file_record_fork.reason     = file_record.reason;

                            if( file_record.filter_description != NULL )
                            {
                              safe_strlcpy( file_record_fork.filter_description,
                                file_record.filter_description,
                                sizeof( file_record_fork.filter_description ) );
                            }

                            log_file( &file_record_fork, &fork_stats );
                        }
                    }
#endif /* PLATFORM_DARWIN */

                    log_file( &file_record, file_stats );
                    return;
                }

                break;

            case ACTION_TYPE_EXCLUDE:

                filter_argument = current_action->action_filter.argument;
                filter_result   = current_action->action_filter.perform(
                                    file_stats, file_path, filter_argument );

                if( filter_result )
                {
                    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                               "file caught by \"Exclude\" action." );

                    safe_strlcat( file_record.filter_description, "(",
                                  sizeof( file_record.filter_description ) );

                    safe_strlcat( file_record.filter_description,
                                  filter_argument,
                                  sizeof( file_record.filter_description ) );

                    safe_strlcat( file_record.filter_description, ")",
                                  sizeof( file_record.filter_description ) );

                    return;
                }

                break;

            case ACTION_TYPE_INCLUDE_ALL:

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "file caught by \"IncludeAll\" action." );

                safe_strlcpy( file_record.filter_description, "IncludeAll",
                              sizeof( file_record.filter_description ) );

                log_file( &file_record, file_stats );

                /* check for resource fork. */
#ifdef PLATFORM_DARWIN

                if( stat( file_record_fork.file_path,
                          &fork_stats ) == 0 )
                {
                    if( fork_stats.st_size > 0 )
                    {
                        files_encountered_count++;

                        file_record_fork.attributes = file_record.attributes;
                        file_record_fork.reason     = file_record.reason;

                        if( file_record.filter_description != NULL )
                        {
                            safe_strlcpy( file_record_fork.filter_description,
                                file_record.filter_description,
                                sizeof( file_record_fork.filter_description ) );
                        }

                        log_file( &file_record_fork, &fork_stats );
                    }
                }

#endif /* PLATFORM_DARWIN */

                return;
                break;    /* yes, I know, bad habit from school. */

            case ACTION_TYPE_EXCLUDE_ALL:

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "file caught by \"ExcludeAll\" action." );

                return;
                break;

            case ACTION_TYPE_UNKNOWN:
            default:

                osi_log_message( "unrecognized action type, run away!!" );
                exit(1);

                break;
        }

        /* move to the next action. */

        action_node = action_node->next;
    }
}

/******************************************************************************
**
**    Function: log_file
**
**    Purpose:  setup the record for logging, then call upon the
**              current block's database to store it.
**
******************************************************************************/

void log_file( record *file_record, struct stat *stats )
{
    bool result   = FALSE;
    char *escaped = NULL;

    if( ( file_record == NULL ) || ( stats == NULL ) )
    {
        return;
    }
    
    /* store the block name for scale. */
    
    safe_strlcpy( file_record->block_name, current_block->directory,
                  sizeof( file_record->block_name ) );

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "logging file: \"%s\" with block:%s",
               file_record->file_path, file_record->block_name );

    /* first we hash file according to the current */
    /* block's hash algorithm attribute, don't     */
    /* do hashes if it's a directory, or zero in   */
    /* size, they will all be the same value.      */
    /*                                             */
    /* Also, don't hash if the file is not a       */
    /* regular file!                               */

    if( S_ISREG( stats->st_mode ) || S_ISDIR( stats->st_mode ) ) 
    {
        switch( current_block->hash )
        {
            case HASH_MD5:

                if( stats->st_size == 0 )
                {
                    break;
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS, "hashing file using MD5." );

                safe_strlcpy( file_record->hash_algorithm, "md5",
                              sizeof( file_record->hash_algorithm ) );

                hash_file_md5( file_record->file_path, file_record->hash,
                               sizeof( file_record->hash ) );
                break;

            case HASH_SHA:

                if( stats->st_size == 0 )
                {
                    break;
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS, "hashing file using SHA1." );

                safe_strlcpy( file_record->hash_algorithm, "sha",
                              sizeof( file_record->hash_algorithm ) );

                hash_file_sha( file_record->file_path, file_record->hash,
                               sizeof( file_record->hash ) );
                break;

            case HASH_HAVAL:

                if( stats->st_size == 0  )
                {
                    break;
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "hashing file using HAVAL-256." );

                safe_strlcpy( file_record->hash_algorithm, "haval",
                              sizeof( file_record->hash_algorithm ) );

                hash_file_haval( file_record->file_path, file_record->hash,
                                 sizeof( file_record->hash ) );
                break;

            case HASH_RIPEMD:

                if( stats->st_size == 0 )
                {
                    break;
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "hashing file using RIPEMD-160." );

                safe_strlcpy( file_record->hash_algorithm, "ripemd",
                              sizeof( file_record->hash_algorithm ) );

                hash_file_ripemd( file_record->file_path, file_record->hash,
                                  sizeof( file_record->hash ) );
                break;

            default:

                if( stats->st_size == 0 )
                {
                    break;
                }

                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "hashing file resorting to default: MD5." );

                safe_strlcpy( file_record->hash_algorithm, "md5",
                              sizeof( file_record->hash_algorithm ) );

                hash_file_md5( file_record->file_path, file_record->hash,
                               sizeof( file_record->hash ) );
                break;

        } /* end switch. */ 

    } /* end if regular file. */

    /* now put stats into our record structure. */

    file_record->stats.device = stats->st_dev;
    file_record->stats.inode = stats->st_ino;
    file_record->stats.permissions = stats->st_mode;
    file_record->stats.links = stats->st_nlink;
    file_record->stats.uid = stats->st_uid;
    file_record->stats.gid = stats->st_gid;
    file_record->stats.mtime = stats->st_mtime;
    file_record->stats.atime = stats->st_atime;
    file_record->stats.ctime = stats->st_ctime;
    file_record->stats.device_type = stats->st_rdev;
    file_record->stats.bytes = stats->st_size;
    file_record->stats.blocks = stats->st_blocks;
    file_record->stats.block_size = stats->st_blksize;
#ifdef HAVE_ST_FLAGS
    file_record->stats.flags = stats->st_flags;
#else
    file_record->stats.flags = 0;
#endif

    /* log the record to the appropriate database.  */
    /* unless db is NULL, in which case we are most */
    /* likely logging to stdout.                    */

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "calling db module storage routine" );

    if( current_block->db != NULL )
    {
        result = current_block->db->store_record( current_block->db,
                                                  file_record );

        if( ( result == FALSE ) && current_block->show_errors )
        {
            print_database_error( current_block->db );
            osi_log_message(
            "==> error:  unable to store file: %s into database, skipping",
            file_record->file_path );

            database_errors_count++;
            return;
        }

        OSI_DEBUG( __POS__, DEBUG_OSIRIS, "logged file to the database." );
    }

    /* if we are logging activity to the console, we  */
    /* print the file path and filter that caught us. */

    escaped = escape_filename( file_record->file_path );

    if( current_block->verbose && escaped )
    {
        osi_log_message( "    %s  -  %s", escaped,
                         file_record->filter_description );
    }

    files_logged_count++;
}

/******************************************************************************
**
**    Function: initialize_global_block
**
**    Purpose:  setup the global block from our config module according
**              to the application defaults.
**
******************************************************************************/

void initialize_global_block()
{
    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
        "initializing the global block with the default values." );

    /* setup our pointer to global block. */

    global_block = osi_config_get_global_block( config_module );

    /* apply default values to the global block before  */
    /* we accept data from config file or command line. */

    global_block->verbose       = DEFAULT_GLOBAL_VERBOSE;
    global_block->prompt        = DEFAULT_GLOBAL_PROMPT;
    global_block->show_errors   = DEFAULT_GLOBAL_SHOW_ERRORS;
    global_block->recursive     = DEFAULT_GLOBAL_RECURSIVE;
    global_block->follow_links  = DEFAULT_GLOBAL_FOLLOW_LINKS;
    global_block->hash          = DEFAULT_GLOBAL_HASH;

    safe_strlcpy( global_block->database, DEFAULT_GLOBAL_DATABASE,
                  sizeof( global_block->database ) );
}

/******************************************************************************
**
**    Function: create_default_block
**
**    Purpose:  create a new block with the global block
**              as its parent.  Insert a single action that
**              is defined as the default action type.  Usually
**              this is 'IncludeAll all'
**
******************************************************************************/

block *create_default_block()
{
    block *default_block;
    action *default_action;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "creating default global block." );

    default_block  = new_block_from_block( global_block );
    default_action = new_action();

    default_action->type       = DEFAULT_GLOBAL_ACTION;
    default_action->attributes = DEFAULT_GLOBAL_ATTRIBUTES;

    list_add( default_block->actions, default_action );

    return default_block;
}

/******************************************************************************
**
**    Function: expand_directory_list
**
**    Purpose:  loop through the directories specified on the
**              command line and expand them to their full
**              paths.
**
******************************************************************************/

void expand_directory_list()
{
    int index;
    char *directory;

    char directory_start[MAXPATHLEN]    = "\0";
    char directory_name[MAXPATHLEN]     = "\0";

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
              "converting all directories to full paths." );

    /* stategy here is to change to every directory in the list, then */
    /* make use of getwd() to get the full current directory path.    */

    getcwd( directory_start, MAXPATHLEN );

    for( index = 0; index < directory_list->size; index++ )
    {
        directory = directory_list->list[index];

        if( directory_exists( directory ) == FALSE )
        {
            string_list_add_item( bad_directory_list, directory );
            continue;
        }

        /* in case relative paths were specified on command line */
        /* we need to go back to where we started from.          */

        chdir( directory_start );
        chdir( directory_list->list[index] );

        osi_free( directory_list->list[index] );
        getcwd( directory_name, MAXPATHLEN );

        directory_list->list[index] = (char *)osi_malloc( 
                                            strlen( directory_name ) + 1 );

        safe_strlcpy( directory_list->list[index], directory_name,
                      strlen( directory_name ) + 1 );
    }

    chdir( directory_start );
}


/******************************************************************************
**
**    Function: create_blocks_for_directory_list
**
**    Purpose:  for each directory specified on the command line
**              we setup default blocks.
**
******************************************************************************/

void create_blocks_for_directory_list()
{
    int index;
    block *directory_block;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
              "creating blocks for directories specified on command line (%d).",
              directory_list->size );

    /* go through all the directories in the directory list */
    /* and create a default block for them, then add them   */
    /* to the config_module's block list.                   */

    for( index = 0; index < directory_list->size; index++ )
    {
        directory_block = create_default_block();

        safe_strlcpy( directory_block->directory, directory_list->list[index],
                      sizeof( directory_block->directory ) - 1 );

        list_add( config_module->blocks, directory_block );
    }
}


/******************************************************************************
**
**    Function: destroy_database_handlers
**
**    Purpose:  go through our db handler list and destroy
**              each one.  they should be closed elsewhere!!
**
******************************************************************************/

void destroy_database_handlers()
{
    list handle = list_get_first_node( db_handles_list );

    while( handle )
    {
        database_handler *db = handle->data;
        osi_free( db );
        handle = handle->next;
    }
}

/******************************************************************************
**
**    Function: create_databases_for_blocks
**
**    Purpose:  loop through the blocks and make sure that
**              each one has a database handler.  in most cases
**              many will actually share the same handler.  this
**              is where we make this happen.
**
******************************************************************************/

void create_databases_for_blocks()
{
    char *db_name;
    bool create_new;

    list db_handle;
    list head;

    head = list_get_first_node( config_module->blocks );

    /* loop through all blocks, creating a db */
    /* when necessary.                        */

    while( head != NULL )
    {
        current_block = head->data;
        create_new    = TRUE;

        db_handle     = list_get_first_node( db_handles_list );
        db_name       = current_block->database;

        /* no need to proceed if we are looking for */
        /* an empty database name.                  */

        if( strlen( db_name ) == 0 )
        {
            head = head->next;
            continue;
        }

        OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                   "looking at block: (%s), with db name: (%s)",
                   current_block->directory, db_name );

        /* look for db name in list, if we find it */
        /* then we point this block's db correctly */
        /* otherwise, we create a new db handler   */
        /* and add it to the list.                 */

        while( db_handle != NULL )
        {
            database_handler *db = db_handle->data;

            /* we found it in the list, don't create a new one. */

            if( strcmp( db_name, db->name ) == 0 )
            {
                OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                           "db: %s already has handler, using that one.",
                           db->name );

                current_block->db = db;
                create_new = FALSE;

                break;
            }

            db_handle = db_handle->next;
        }

        if( create_new )
        {
            OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                       "db: %s is new, creating db handler for it.",
                        db_name );

            /* we didn't find it, so we create one.     */
            /* and add to  db handle list, and          */
            /* then add the name to the database list.  */

            current_block->db       = database_handler_new();
            current_block->db->name = db_name;

            list_add( db_handles_list, current_block->db );
            string_list_add_item( database_list, db_name );
        }

        head = head->next;
    }
}


/******************************************************************************
**
**    Function: verify_config_state
**
**    Purpose:  look at the error and warning counts on the config module
**              and exit on errors, prompt on warnings.
**
******************************************************************************/

void verify_config_state()
{
    bool result;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "verifying state of config module..." );

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "config module contains %d errors.",
               config_module->error_count );

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "config module contains %d warnings.",
               config_module->warning_count );

    if( config_module->error_count > 0 )
    {
        halt();
        osi_log_message_exit(
                        "config file: %s contains %d fatal error(s), exiting.",
                         config_module->file_name,
                         config_module->error_count );
    }

    if( ( config_module->warning_count > 0 ) && global_block->prompt )
    {
        result = osi_continue_prompt(
                        "config file: %s contains %d warning(s),\ncontinue?",
                         config_module->file_name,
                         config_module->warning_count );

        if( result == FALSE )
        {
            halt();
            exit(0);
        }
    }
}

/******************************************************************************
**
**    Function: verify_directory_list
**
**    Purpose:  loop through all of the directories in the blocks
**              and verify they exist, and are in fact directories.  we
**              already did this for the ones on the command line.
**
******************************************************************************/

void verify_directory_list()
{
    int index;

    list head;
    block *current_block;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "verifying all directories..." );

    /* we already verify directories on the command line. */
    /* because they have to be verified before they are   */
    /* expanded.  so here we just go through the blocks   */
    /* and look for bad directories.                      */

    /* get first node from the block list. */

    head = list_get_first_node( config_module->blocks );

    /* skip over the global node. */

    head = head->next;

    /* loop through entire block list verifying each directory. */

    while( head != NULL )
    {
        current_block = head->data;

        OSI_DEBUG( __POS__, DEBUG_OSIRIS,
                   "directory \"%s\"", current_block->directory );

        if( current_block != NULL &&
            directory_exists( current_block->directory ) == FALSE )
        {
            OSI_DEBUG( __POS__, DEBUG_OSIRIS, "bad directory --> \"%s\"",
                       current_block->directory );

            string_list_add_item( bad_directory_list,
                                  current_block->directory );
        }

        head = head->next;
    }

    /* now check the list and print warning if any bad   */
    /* directories exist on command line or config file. */

    if( bad_directory_list->size > 0 )
    {
        osi_log_message(
            "warning: the following are not directories or cannot be opened:" );

        osi_log_message( "" );

        for( index = 0; index < bad_directory_list->size; index++ )
        {
            osi_log_message( bad_directory_list->list[index] );
        }

        osi_log_message( "" );

        if( global_block->prompt &&
            osi_continue_prompt( "continue?" ) == FALSE )
        {
            halt();
            exit(0);
        }
    }
}

/******************************************************************************
**
**    Function: verify_config_file
**
**    Purpose:  make sure the config file we are asked to open and parse
**              exists, is a regular file, can be opened, and isn't
**              frickin huge.
**
******************************************************************************/

void verify_config_file()
{
    char *config_file;
    struct stat file_stats;

    config_file = config_module->file_name;

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
               "verifying config file: \"%s\" exists, and is regular...",
                config_file );

    if( stat( config_file, &file_stats ) != 0 )
    {
        osi_log_message_exit( "config file: %s cannot be opened.",
                              config_file );
        halt();
        exit(0);
    }

    if( !S_ISREG( file_stats.st_mode ) )
    {
        osi_log_message_exit( "config file: %s is not a regular file.",
                              config_file );
        halt();
        exit(0);
    }

    if( file_stats.st_size > 100000 )
    {
        osi_log_message_exit( "file: %s seems far too large, sorry.",
                              config_file );
        halt();
        exit(0);
    }

    OSI_DEBUG( __POS__, DEBUG_OSIRIS,
               "config file exists and is a regular file." );
}


/******************************************************************************
**
**    Function: store_database_headers
**
**    Purpose:  create headers for each of the databases we
**              are using.  the first field, purposly, is the
**              version of osiris so that we can handle differences
**              in the future, if we so desire.
**
**              ------------- database header -------------
**
**              database header, contains information about the
**              scan, and the database itself.
**
**              OSIRIS_VERSION:DATE:USERNAME:CONFIG_FILE:COUNT:
**
**
******************************************************************************/

void store_database_headers()
{
    list db_handle;
    database_handler *db;

    char *temp;
    char counter[30]   = "";

    char version[20]   = "";
    char date[100]     = "";
    char username[512] = "";

    int index;
    char buffer[512]   = "";

    time_t current_time;
    struct passwd *current_user;

    char *header       = NULL;
    char *config       = NULL;

    int header_length  = 0;
    int config_length  = 0;

    int total_header_length = 0;

    /* get the most current timestamp, encode it */
    /* and place back into itself.               */

    OSI_DEBUG( __POS__, DEBUG_OSIRIS, "storing database headers..." );

    current_time = time( NULL );
    safe_strlcpy( buffer, (char *)ctime( &current_time ), sizeof( buffer ) );
    buffer[strlen( buffer ) - 1] = '\0';

    base64_encode( date, buffer, strlen( buffer ) );

    /* get the username, encode it and place it  */
    /* back into itself.                         */

    current_user = getpwuid( getuid() );

    if( current_user != NULL )
    {
        safe_strlcpy( buffer, current_user->pw_name, sizeof( username ) );
    }

    else
    {
        safe_strlcpy( buffer, "unknown", sizeof( buffer ) );
    }

    base64_encode( username, buffer, strlen( buffer ) );
    base64_encode( version, PROGRAM_VERSION, strlen( PROGRAM_VERSION ) );

    /* determine the length. this includes the config file */
    /* if we have one to deal with.                        */

    header_length += strlen( version );
    header_length += strlen( date );
    header_length += strlen( username );
    header_length += sizeof( counter );
    header_length += 10;

    /* determine the length of the config file, if present  */
    /* and also create a single string for the config       */

    if( config_module->file_data != NULL )
    {
        for( index = 0; index < config_module->file_data->size; index++ )
        {
            if( config_module->file_data->list[index] != NULL )
            {
                config_length += strlen(
                                     config_module->file_data->list[index] );
            }

             config_length++;    /* for newlines. */
        }

        /* account for base64 encoding size increase by 30%. */

        config = osi_malloc( (int)( config_length * 1.4 ) );
        memset( config, 0, (int)( config_length * 1.4 ) );

        for( index = 0; index < config_module->file_data->size; index++ )
        {
            if( config_module->file_data->list[index] != NULL )
            {
                safe_strlcat( config, config_module->file_data->list[index],
                            ( config_length * 1.4 ) );
            }

            safe_strlcat( config, "\n", ( config_length * 1.4 ) );
        }

        /* encode the config file. */
    }

    total_header_length = ( header_length + ( config_length * 1.4 ) );
    header = osi_malloc( total_header_length );

    safe_strlcpy( header, version, total_header_length );
    safe_strlcat( header, ":", total_header_length );
    safe_strlcat( header, date, total_header_length );
    safe_strlcat( header, ":", total_header_length );
    safe_strlcat( header, username, total_header_length );
    safe_strlcat( header, ":", total_header_length );

    if( config != NULL )
    {
        base64_encode( strrchr( header, '\0' ), config, config_length );
    }

    /* if no config, leave a space so we can   */
    /* easily tell this when examining the db. */

    safe_strlcat( header, " :", total_header_length );

    /* save end position so we can back up and */
    /* tack on the counter each time.          */

    temp = strchr( header, '\0' );

    /* loop through each database, adding the header. since */
    /* the couters may vary, we tack on the record count    */
    /* here and back up each time so we can add the next.   */

    db_handle = list_get_first_node( db_handles_list );

    while( db_handle != NULL )
    {
        db = db_handle->data;

        if( db != NULL )
        {
            safe_snprintf( counter, sizeof( counter ), "%020d:", db->counter );

            safe_strlcpy( temp, counter,
                        ( total_header_length - strlen( header ) ) );

            OSI_DEBUG( __POS__, DEBUG_OSIRIS, "storing header for database: %s",
                       db->name );

            OSI_DEBUG( __POS__, DEBUG_OSIRIS, "header is: %s", header );

            db->store_header( db, header );
            db_handle = db_handle->next;
        }
    }

    if( config != NULL )
    {
        osi_free( config );
    }

    osi_free( header );
}

/******************************************************************************
**
**    Function: print_database_error
**
**    Purpose: retrieve last string error message from the database handler.
**             this should provide some explanation as to why the last db
**             action failed.
**
******************************************************************************/

void print_database_error( database_handler *db )
{
    char *error_message;

    if( db != NULL )
    {
        if( ( error_message = db->retrieve_last_error(db) ) != NULL )
        {
            if( error_message[0] != '\0' )
            {
                osi_log_message( "%s: %s", DB_MODULE, error_message );
            }
        }
    }
}
