
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      scale.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#include "scale.h"
#include "config.h"

/******************************************************************************
**
**    Function: main
**
**    Purpose:  system driver for scale
**
**    Return:   zero, if all goes well.
**
******************************************************************************/

int main( int argument_count, char *argument_list[] )
{
    initialize( argument_count, argument_list );
    
    if( custom_config )
    {
        process_with_config();
    }
        
    else
    {
        process();
    }
        
    halt();
    return 0;
}

/******************************************************************************
**
**    Function: initialize
**
**    Purpose:  allocate for global data, check arguments, and
**              open all output resources.
**
******************************************************************************/

void initialize( int argument_count, char *argument_list[] )
{
    left_database     = database_handler_new();
    right_database    = database_handler_new();

    new_file_list     = string_list_new();
    missing_file_list = string_list_new();

    /* default output is to stdout, otherwise, */
    /* we set it to something else below.      */

    output_file       = stdout;
    check_arguments( argument_count, argument_list );

    open_output_file();
    open_databases();
    
    if( custom_config )
    {
        config_module = osi_config_new( config_file_path );
        verify_config_file();
        osi_config_parse( config_module );
        verify_config_state();
    }
}

/******************************************************************************
**
**    Function: halt
**
**    Purpose:  close output resources, free all global data, and destroy
**              our database handlers.
**
******************************************************************************/

void halt()
{
    close_databases();
    close_output_file();

    string_list_destroy( new_file_list );
    string_list_destroy( missing_file_list );

    database_handler_destroy( left_database );
    database_handler_destroy( right_database );
    
    if( custom_config )
    {
        osi_config_destroy( config_module );
    }
}

/******************************************************************************
**
**    Function: check_arguments
**
**    Purpose:  parse all command line arguments, acting accordingly.
**
******************************************************************************/

void check_arguments( int argument_count, char *argument_list[] )
{
    int count;

    OSI_DEBUG( __POS__,  DEBUG_SCALE, "parsing command line arguments (%d)",
             ( argument_count - 1 ) );

    if( argument_count == 1 )
    {
        print_usage();
        halt();
        exit(0);
    }

    for( count = 1; count < argument_count; count++ )
    {
        if( ( strcmp( argument_list[count], "-p" ) == 0 ||
              strcmp( argument_list[count], "--print" ) == 0 )
          && ( count < ( argument_count - 1 ) ) )
        {
            /* just use one of our open handlers. */

            left_database->name = argument_list[++count];

            if( open_database( left_database ) == TRUE )
            {
                print_database_contents( left_database );
                print_program_info();
                close_database( left_database );
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                 left_database->name );
            }

            halt();
            exit(0);
        }

        if( ( strcmp( argument_list[count], "-x" ) == 0 ||
              strcmp( argument_list[count], "--print-all" ) == 0 )
          && ( count < ( argument_count - 1 ) ) )
        {
            /* just use one of our open handlers. */

            left_database->name = argument_list[++count];

            if( open_database( left_database ) == TRUE )
            {
                print_complete_database_contents( left_database );
                print_program_info();
                close_database( left_database );
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                 left_database->name );
            }

            halt();
            exit(0);
        }

        if( ( strcmp( argument_list[count], "--checksums" ) == 0 ||
              strcmp( argument_list[count], "-a" ) == 0 ) &&
              ( count < ( argument_count - 1 ) ) )
        {
            /* just use one of our open handlers. */

            left_database->name = argument_list[++count];

            if( open_database( left_database ) == TRUE )
            {
                print_database_checksums( left_database );
                print_program_info();
                close_database( left_database );
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                 left_database->name );
            }

            halt();
            exit(0);

        }

        else if( ( strcmp( argument_list[count], "-e" ) == 0 ||
                   strcmp( argument_list[count], "--extract-config" ) == 0 )
                && ( count < ( argument_count - 1 ) ) )
        {
            /* just use one of our open handlers. */

            left_database->name = argument_list[++count];

            if( open_database( left_database ) == TRUE )
            {
                print_database_config( left_database );
                close_database( left_database );
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                 left_database->name );
            }

            halt();
            exit(0);
        }

        else if( ( strcmp( argument_list[count], "-d" ) == 0 ||
                   strcmp( argument_list[count], "--duplicates" ) == 0 ) &&
                   ( argument_count == 3 ) )
        {
            /* just use one of our open handlers. */

            left_database->name = argument_list[++count];

            if( open_database( left_database ) == TRUE )
            {
                print_database_header( left_database );
                print_duplicate_records( left_database );

                close_database( left_database );
                print_program_info();
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                 left_database->name );
            }

            halt();
            exit(0);
        }

        else if( strcmp( argument_list[count], "-d" ) == 0 ||
                 strcmp( argument_list[count], "--show-duplicates" ) == 0 )
        {
            print_duplicates = TRUE;
        }

        else if( strcmp( argument_list[count], "-f" ) == 0 ||
                 strcmp( argument_list[count], "--no-prompt" ) == 0 )
        {
            no_prompts = TRUE;
        }

        else if( strcmp( argument_list[count], "-q" ) == 0 ||
                 strcmp( argument_list[count], "--quiet" ) == 0 )
        {
            quiet_mode = TRUE;
        }

        else if( strcmp( argument_list[count], "-n" ) == 0 ||
                 strcmp( argument_list[count], "--no-new" ) == 0 )
        {
            print_new = FALSE;
        }

        else if( strcmp( argument_list[count], "-m" ) == 0 ||
                 strcmp( argument_list[count], "--no-missing" ) == 0 )
        {
            print_missing = FALSE;
        }
            
        else if( ( strcmp( argument_list[count], "-c" ) == 0 ||
                   strcmp( argument_list[count], "--config" ) == 0 ) &&
                 ( count < ( argument_count - 1 ) ) )
        {
            custom_config = TRUE;
            safe_strlcpy( config_file_path, argument_list[++count],
                          sizeof( config_file_path ) );
        }

        else if( ( strcmp( argument_list[count], "-l" ) == 0 ||
                   strcmp( argument_list[count], "--left" ) == 0 ) &&
                 ( count < ( argument_count - 1 ) ) )
        {
            safe_strlcpy( left_database_path, argument_list[++count],
                          sizeof( left_database_path ) );
        }

        else if( ( strcmp( argument_list[count], "-r" ) == 0 ||
                   strcmp( argument_list[count], "--right" ) == 0 ) &&
                 ( count < ( argument_count - 1 ) ) )
        {
            safe_strlcpy( right_database_path, argument_list[++count],
                          sizeof( right_database_path ) );
        }

        else if( strcmp( argument_list[count], "-o" ) == 0 &&
               ( count < ( argument_count - 1 ) ) )
        {
            safe_strlcpy( output_file_path, argument_list[++count],
                          sizeof( output_file_path ) );
        }

        else if( ( strcmp( argument_list[count], "-h" ) == 0 ) ||
                   strcmp( argument_list[count], "--help" ) == 0 )
        {
            print_usage();
            halt();
            exit(0);
        }

        else if( ( strcmp( argument_list[count], "-v" ) == 0 ) ||
            strcmp( argument_list[count], "--version" ) == 0 )
        {
            print_program_info();
            halt();
            exit(0);
        }

        else if( argument_count == 2 )
        {
            /* just use of our our open handlers. */

            left_database->name = argument_list[1];

            if( open_database( left_database ) == TRUE )
            {
                print_database_header( left_database );
                print_program_info();
                close_database( left_database );
            }

            else
            {
                print_database_error( left_database );
                osi_log_message( "error: unable to open database: %s, exiting.",
                                  argument_list[1] );
            }

            halt();
            exit(0);
        }

        else
        {
            print_usage();
            halt();
            exit(0);
        }

    } /* end for loop. */

    /* simple sanity check, in case of stupid people. */

    if( quiet_mode && strlen( output_file_path ) == 0 )
    {
        osi_log_message( "error: this would be useless, you must specify an output file in quiet mode." );
        halt();
        exit(0);
    }
}

/******************************************************************************
**
**    Function: process
**
**    Purpose:  master routine for scale.  here we process
**              the databases and compare them, print duplicate
**              checksum lists, if requested, then print the results.
**
******************************************************************************/

void process()
{
    OSI_DEBUG( __POS__, DEBUG_SCALE, "scale beginning processing." );

    /* make room for progress counters. */

    if( quiet_mode == FALSE && output_file != stdout )
    {
        osi_log_message( "" );
    }

    print_output_header();
    process_records();

    if( print_new )
    {
        print_new_records();
    }

    if( print_missing )
    {
        print_missing_records();
    }

    if( print_duplicates )
    {
        print_duplicate_records( left_database );
        print_duplicate_records( right_database );
    }

    print_summary();

    if( quiet_mode == FALSE )
    {
        print_program_info();
    }
}


/******************************************************************************
**
**    Function: process_with_config
**
**    Purpose:  loop through the db records comparing them according
**              to the passed in config file.
**
******************************************************************************/

void process_with_config()
{
    OSI_DEBUG( __POS__, DEBUG_SCALE, "scale beginning processing." );

    /* make room for progress counters. */

    if( quiet_mode == FALSE && output_file != stdout )
    {
        osi_log_message( "" );
    }

    print_output_header();
    process_records_with_config();

    if( print_new )
    {
        print_new_records();
    }

    if( print_missing )
    {
        print_missing_records();
    }

    if( print_duplicates )
    {
        print_duplicate_records( left_database );
        print_duplicate_records( right_database );
    }

    print_summary();

    if( quiet_mode == FALSE )
    {
        print_program_info();
    }
}

/******************************************************************************
**
**    Function: process_records
**
**    Purpose:  loop through the db records comparing them.
**
******************************************************************************/

void process_records()
{
    bool result;

    record left_record;
    record right_record;

    memset( &left_record, 0, sizeof( left_record ) );
    memset( &right_record, 0, sizeof( right_record ) );

    osi_log_message_to_file( output_file, "\n[ file differences ]" );
    osi_log_message_to_file( output_file, "" );

    /* loop through each record in the databases, and  */
    /* perform a comparison on each, we can use either */
    /* database to iterate through, it doesnt matter.  */

    result = left_database->retrieve_first_record( left_database,
                                                   &left_record );

    while( result == TRUE )
    {
        /* first, see if the other database has this record. */

        result = right_database->retrieve_record( right_database, 
                                                  left_record.file_path, 
                                                  &right_record );

        /* it does, so we compare the records. */

        if( result != FALSE )
        {
            compare_records( &left_record, &right_record );
            print_progress_counter( "files compared", compare_count );
        }

        /* otherwise, we do nothing, just go to the next record. */

        result = left_database->retrieve_next_record( left_database,
                                                      &left_record );
    }

    reset_progress_counter();
}

/******************************************************************************
**
**    Function: process_records_with_config
**
**    Purpose:  loop through the db records comparing them according
**              to the given config file.
**
******************************************************************************/

void process_records_with_config()
{
    bool result;

    record left_record;
    record right_record;

    memset( &left_record, 0, sizeof( left_record ) );
    memset( &right_record, 0, sizeof( right_record ) );

    osi_log_message_to_file( output_file, "\n[ file differences ]" );
    osi_log_message_to_file( output_file, "" );

    /* loop through each record in the databases, and  */
    /* perform a comparison on each, we can use either */
    /* database to iterate through, it doesnt matter.  */

    result = left_database->retrieve_first_record( left_database,
                                                   &left_record );
    
    while( result != FALSE )
    {
        /* first, see if the other database has this record. */

        result = right_database->retrieve_record( right_database,
                                                  left_record.file_path,
                                                  &right_record );

        /* it does, so we compare the records. */

        if( result != FALSE )
        {
            /* here we modify the attributes mask according to the */
            /* config file as if it were in the database, then we  */
            /* process the compare as normal.                      */
            
            if( found_config_attributes_for_records( &left_record,
                                                     &right_record ) )
            {
                compare_records( &left_record, &right_record );
                print_progress_counter( "files compared", compare_count );
            }
        }

        /* otherwise, we do nothing, just go to the next record. */

        result = left_database->retrieve_next_record( left_database,
                                                      &left_record );
    }

    reset_progress_counter();
}

/******************************************************************************
**
**    Function: found_config_attributes_for_records
**
**    Purpose:  find block ( if any ) in config module, look for rule that
**              caught the file, acquire the attributes mask and place
**              that mask back into both records so we can compare them
**              according to the attributes specified in the config file.
**
******************************************************************************/

bool found_config_attributes_for_records( record *left_record,
                                          record *right_record )
{
    block *matching_block;
    action *matching_action;
    
    bool result = FALSE;
    
    matching_block = get_block_with_name( config_module,
                                          left_record->block_name );
    
    if( matching_block != NULL )
    {
        char filter_argument[MAX_FILTER_ARGUMENT_LENGTH] = "";

        get_filter_value( left_record->filter_description, filter_argument );
        matching_action = get_action_from_block_with_filter( matching_block,
                                                      left_record->reason,
                                                      filter_argument );
        
        /* if we were unable to get the action from this block, it might */
        /*  have used the global block's default rule, so try to get it  */
        /*  from the global block.                                       */
        
        if( matching_action == NULL )
        {
            block *global_block = osi_config_get_global_block( config_module );
            matching_action = get_action_from_block_with_filter( global_block,
                                                left_record->reason,
                                                filter_argument );
        }
        
        if( matching_action != NULL )
        {
            left_record->attributes  = matching_action->attributes;
            right_record->attributes = matching_action->attributes;
        
            result = TRUE;
        }
    }
        
    return result;
}



/******************************************************************************
**
**    Function: compare_records
**
**    Purpose:  if records differ, print to output resource the
**              differences.
**
******************************************************************************/

void compare_records( record *left_record, record *right_record )
{
    bool different        = FALSE;

    char left_buffer[50]  = "";
    char right_buffer[50] = "";

    struct osi_stat left_stats;
    struct osi_stat right_stats;

    osi_uint32 left_attributes;
    osi_uint32 right_attributes;

    if( left_record == NULL || right_record == NULL )
    {
        return;
    }

    compare_count++;

    left_stats  = left_record->stats;
    right_stats = right_record->stats;

    left_attributes  = left_record->attributes;
    right_attributes = right_record->attributes;

    /* find out if any of the attributes in the */
    /* specified list differ from each other.   */

    if( records_differ( left_record, right_record ) == FALSE )
    {
        return;
    }

    difference_count++;

    /* something has changed, find out what it is, and */
    /* print out the details.                          */

    osi_log_message_to_file( output_file, "  %s",
                             escape_filename( left_record->file_path ) );

    /* check hash values, set flag if they differ. */

    if( strcmp( left_record->hash, right_record->hash ) != 0 )
    {
        osi_log_message_to_file( output_file, "    %-10.10s %40s ==> %s",
                                 left_record->hash_algorithm,
                                 left_record->hash, right_record->hash );
    }

    /* check all other attributes. */

    if( ( left_attributes & ATTRIBUTE_PERM ) ||
        ( right_attributes & ATTRIBUTE_PERM ) )
    {
        if( left_stats.permissions != right_stats.permissions )
        {
            get_file_attribute_string( left_buffer, sizeof( left_buffer ),
                                       (mode_t)left_stats.permissions );

            get_file_attribute_string( right_buffer, sizeof( right_buffer ),
                                       (mode_t)right_stats.permissions );

            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40s ==> %s", "permissions",
                                     left_buffer, right_buffer );
        }
    }

    if( ( left_attributes & ATTRIBUTE_MTIME ) ||
        ( right_attributes & ATTRIBUTE_MTIME ) )
    {
        if( left_stats.mtime != right_stats.mtime )
        {
            long left  = left_stats.mtime;
            long right = right_stats.mtime;

            strftime( left_buffer, sizeof( left_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&left ) );

            strftime( right_buffer, sizeof( right_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&right ) );

            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40s ==> %s", "mtime",
                                     left_buffer,
                                     right_buffer );
        }
    }

    if( ( left_attributes & ATTRIBUTE_ATIME ) ||
        ( right_attributes & ATTRIBUTE_ATIME ) )
    {
        if( left_stats.atime != right_stats.atime )
        {
            long left  = left_stats.atime;
            long right = right_stats.atime;

            strftime( left_buffer, sizeof( left_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&left ) );

            strftime( right_buffer, sizeof( right_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&right ) );

            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40s ==> %s","atime",
                                     left_buffer, 
                                     right_buffer );
        }
    }

    if( ( left_attributes & ATTRIBUTE_CTIME ) ||
        ( right_attributes & ATTRIBUTE_CTIME ) )
    {
        if( left_stats.ctime != right_stats.ctime )
        {
            long left  = left_stats.ctime;
            long right = right_stats.ctime;

            strftime( left_buffer, sizeof( left_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&left ) );

            strftime( right_buffer, sizeof( right_buffer ), DATE_FORMAT_STRING,
                      localtime( (time_t *)&right ) );

            osi_log_message_to_file( output_file,
                                    "    %-10.10s %40s ==> %s","ctime",
                                    left_buffer, 
                                    right_buffer );
        }
    }

    if( ( left_attributes & ATTRIBUTE_INODE ) || 
        ( right_attributes & ATTRIBUTE_INODE ) )
    {
        if( left_stats.inode != right_stats.inode )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d","inode",
                                     (int)left_stats.inode,
                                     (int)right_stats.inode );
        }
    }

    if( ( left_attributes & ATTRIBUTE_LINKS ) ||
        ( right_attributes & ATTRIBUTE_LINKS ) )
    {
        if( left_stats.links != right_stats.links )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d","links",
                                     (int)left_stats.links,
                                     (int)right_stats.links );
         }
    }

    if( ( left_attributes & ATTRIBUTE_UID ) ||
        ( right_attributes & ATTRIBUTE_UID ) )
    {
        if( left_stats.uid != right_stats.uid )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d", "uid",
                                    (int)left_stats.uid,
                                    (int)right_stats.uid );
        }
    }

    if( ( left_attributes & ATTRIBUTE_GID ) || 
        ( right_attributes & ATTRIBUTE_GID ) )
    {
        if( left_stats.gid != right_stats.gid )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d", "gid",
                                    (int)left_stats.gid,
                                    (int)right_stats.gid );
        }
    }

    if( ( left_attributes & ATTRIBUTE_BYTES ) ||
        ( right_attributes & ATTRIBUTE_BYTES ) )
    {
        if( left_stats.bytes != right_stats.bytes )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d", "bytes",
                                     (int)left_stats.bytes,
                                     (int)right_stats.bytes );
        }
    }

    if( ( left_attributes & ATTRIBUTE_BLOCKS ) ||
        ( right_attributes & ATTRIBUTE_BLOCKS ) )
    {
        if( left_stats.blocks != right_stats.blocks )
        {
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d", "blocks",
                                     (int)left_stats.blocks,
                                     (int)right_stats.blocks );
        }
    }

#if HAVE_ST_FLAGS

    if( ( left_attributes & ATTRIBUTE_FLAGS ) ||
        ( right_attributes & ATTRIBUTE_FLAGS ) )
    {
        if( left_stats.flags != right_stats.flags )
        {
            int left_flags = left_stats.flags;
            int right_flags = right_stats.flags;

#if HAVE_FFLAGSTOSTR

            safe_strlcpy( left_buffer, fflagstostr( left_flags ),
                          sizeof( left_buffer ) );

            if( strlen( left_buffer ) == 0 )
            {
                safe_strlcpy( left_buffer, "-", sizeof( left_buffer ) );
            }

            safe_strlcpy( right_buffer, fflagstostr( right_flags ),
                         sizeof( right_buffer ) );

            if( strlen( right_buffer ) == 0 )
            {
                safe_strlcpy( right_buffer, "-", sizeof( right_buffer ) );
            }

            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40s ==> %s", "flags",
                                     left_buffer,
                                     right_buffer );
#else
            osi_log_message_to_file( output_file,
                                     "    %-10.10s %40d ==> %d", "flags",
                                     left_flags,
                                     right_flags );
#endif
        }
    }

#endif

    osi_log_message_to_file( output_file, "" );
}

/******************************************************************************
**
**    Function: records_differ
**
**    Purpose:  examine all aspects of each record.  if we find
**              a difference, return TRUE, return FALSE otherwise.
**
**    NOTE: this routine is seperate from the above routine because
**          it's easier to print the differences as we come accross
**          them, but not all attributes are always considered. and
**          we don't know before hand to print the file name or not.
**          we should probably just put it into a buffer and print
**          the buffer if needed.
**
******************************************************************************/

bool records_differ( record *left_record, record *right_record )
{
    struct osi_stat left_stats;
    struct osi_stat right_stats;

    osi_uint32 left_attributes;
    osi_uint32 right_attributes;

    if( left_record == NULL || right_record == NULL )
    {
        return FALSE;
    }

    left_stats  = left_record->stats;
    right_stats = right_record->stats;

    left_attributes  = left_record->attributes;
    right_attributes = right_record->attributes;

    /* check hash values, set flag if they differ. */

    if( left_record->hash[0] != '\0' &&
        strcmp( left_record->hash, right_record->hash ) != 0 )
    {
        return TRUE;
    }

    /* otherwise, we have to go through all of the  */
    /* individual attributes and check.             */

    if( ( left_attributes & ATTRIBUTE_PERM ) ||
        ( right_attributes & ATTRIBUTE_PERM ) )
    {
        if( left_stats.permissions != right_stats.permissions )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_MTIME ) ||
        ( right_attributes & ATTRIBUTE_MTIME ) )
    {
        if( left_stats.mtime != right_stats.mtime )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_ATIME ) ||
        ( right_attributes & ATTRIBUTE_ATIME ) )
    {
        if( left_stats.atime != right_stats.atime )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_CTIME ) ||
        ( right_attributes & ATTRIBUTE_CTIME ) )
    {
        if( left_stats.ctime != right_stats.ctime )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_INODE ) ||
        ( right_attributes & ATTRIBUTE_INODE ) )
    {
        if( left_stats.inode != right_stats.inode )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_LINKS ) ||
        ( right_attributes & ATTRIBUTE_LINKS ) )
    {
        if( left_stats.links != right_stats.links )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_UID ) ||
        ( right_attributes & ATTRIBUTE_UID ) )
    {
        if( left_stats.uid != right_stats.uid )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_GID ) ||
        ( right_attributes & ATTRIBUTE_GID ) )
    {
        if( left_stats.gid != right_stats.gid )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_BYTES ) ||
        ( right_attributes & ATTRIBUTE_BYTES ) )
    {
        if( left_stats.bytes != right_stats.bytes )
        {
            return TRUE;
        }
    }

    if( ( left_attributes & ATTRIBUTE_BLOCKS ) ||
        ( right_attributes & ATTRIBUTE_BLOCKS ) )
    {
        if( left_stats.blocks != right_stats.blocks )
        {
            return TRUE;
        }
    }

#ifdef HAVE_ST_FLAGS
    if( ( left_attributes & ATTRIBUTE_FLAGS ) ||
        ( right_attributes & ATTRIBUTE_FLAGS ) )
    {
        if( left_stats.flags != right_stats.flags )
        {
            return TRUE;
        }
    }
#endif

    return FALSE;
}

/******************************************************************************
**
**    Function: print_new_records
**
**    Purpose:  loop through right database, if it's not in the left, that
**              means it's probably a new file.
**
******************************************************************************/

void print_new_records()
{
    int index;
    int count = 0;

    bool result;

    record right_record;
    record left_record;

    char buffer[MAXPATHLEN + 100] = "\0";
    char temp[50] = "\0";

    memset( &left_record, 0, sizeof( left_record ) );
    memset( &right_record, 0, sizeof( right_record ) );

    OSI_DEBUG( __POS__, DEBUG_SCALE, "checking for new files." );

    result = right_database->retrieve_first_record( right_database,
                                                    &right_record );

    while( result != FALSE )
    {
        /* first, see if the left database knew */
        /* about this file.                     */

        result = left_database->retrieve_record( left_database,
                                                 right_record.file_path,
                                                 &left_record );

        /* it doesn't, so we add the name to    */
        /* the list of new records.             */

        if( result == FALSE )
        {
            safe_snprintf( temp, sizeof( temp ), "%d",
                           right_record.stats.device );

            safe_strlcpy( buffer, temp, sizeof( buffer ) );
            safe_strlcat( buffer, ":", sizeof( buffer ) );

            safe_snprintf( temp, sizeof( temp ), "%d",
                           right_record.stats.inode );

            safe_strlcat( buffer, temp, sizeof( buffer ) );
            safe_strlcat( buffer, "  ", sizeof( buffer ) );

            safe_strlcat( buffer, right_record.file_path, sizeof( buffer ) );
            string_list_add_item( new_file_list, buffer );
        }

        /* update counter. */

        new_progress++;
        print_progress_counter( "checking for new files", new_progress );

        /* move on to the next right db record. */

        result = right_database->retrieve_next_record( right_database,
                                                       &right_record );
    }

    reset_progress_counter();
    new_record_count = new_file_list->size;

    osi_log_message_to_file( output_file, "\n[ new files (%d) ]",
                             new_record_count );

    osi_log_message_to_file( output_file, "" );

    for( index = 0; index < new_file_list->size; index++ )
    {
        osi_log_message_to_file( output_file, "  %s",
                                 new_file_list->list[index] );
    }
}

/******************************************************************************
**
**    Function: print_missing_records
**
**    Purpose:  loop through left database, if it's not in the right, that
**              means it's probably a missing/deleted file.
**
******************************************************************************/

void print_missing_records()
{
    int index;
    bool result;

    record right_record;
    record left_record;

    memset( &left_record, 0, sizeof( left_record ) );
    memset( &right_record, 0, sizeof( right_record ) );

    OSI_DEBUG( __POS__, DEBUG_SCALE, "checking for new files." );

    result = left_database->retrieve_first_record( left_database,
                                                   &left_record );

    while( result != FALSE )
    {
        /* first, see if the right database still */
        /* has this file.                         */

        result = right_database->retrieve_record( right_database,
                                                  left_record.file_path,
                                                  &right_record );

        /* it doesn't, so we add the name to    */
        /* the list of missing records.         */

        if( result == FALSE )
        {
            string_list_add_item( missing_file_list,
                                  escape_filename( left_record.file_path ) );
        }

        /* update counter. */

        missing_progress++;
        print_progress_counter( "checking for missing files",
                                missing_progress );

        /* move on to the next right db record. */

        result = left_database->retrieve_next_record( left_database,
                                                      &left_record );
    }

    reset_progress_counter();
    missing_record_count = missing_file_list->size;

    osi_log_message_to_file( output_file, "\n[ missing files (%d) ]",
                             missing_record_count );

    osi_log_message_to_file( output_file, "" );

    for( index = 0; index < missing_file_list->size; index++ )
    {
        osi_log_message_to_file( output_file, "  %s",
                                 missing_file_list->list[index] );
    }
}


/******************************************************************************
**
**    Function: print_duplicate_records
**
**    Purpose:  loop through the database creating a list of files
**              that contain the same checksum or hash.  we then
**              print out the groups of duplicates.
**
******************************************************************************/

void print_duplicate_records( database_handler *db )
{
    bool result;

    int index  = 0;
    bool found = FALSE;

    record current_record;
    record next_record;

    list duplicates = NULL;
    list temp_list  = NULL;

    string_list *temp_string_list = NULL;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return;
    }

    duplicates = list_new();
    result = db->retrieve_first_record( db, &current_record );

    /* loop through records. */

    osi_log_message_to_file( output_file, "\n[ duplicate checksums for %s ]",
                             db->name );

    osi_log_message_to_file( output_file, "" );

    while( result != FALSE )
    {
        temp_list = list_get_first_node( duplicates );

        /* loop through the duplicates list, if */
        /* this hash is found, append the file  */
        /* path to the string_list, otherwise   */
        /* create a new string_list and add it  */
        /* to the duplicates list.              */

        while( temp_list )
        {
            temp_string_list = temp_list->data;

            /* make sure string list has a hash for a head. */
            /* if the hashes match, add file_path to the    */
            /* string list.                                 */

            if( temp_string_list && temp_string_list->size > 1 )
            {
                if( ( strlen( current_record.hash ) > 0 ) &&
                      strcmp( current_record.hash,
                      temp_string_list->list[1] ) == 0 )
                {
                    string_list_add_item( temp_string_list,
                                escape_filename( current_record.file_path ) );

                    found = TRUE;
                }

                /* move onto the next string_list . */
            }

            temp_list = temp_list->next;
        }

        /* if we didn't find something, add a new string list. */
        /* make the first string the algorithm, then the hash, */
        /* then the file path, then add the new string_list to */
        /* the duplicates.                                     */

        if( found == FALSE )
        {
            temp_string_list = string_list_new();
            list_add( duplicates, temp_string_list );

            string_list_add_item( temp_string_list,
                                  current_record.hash_algorithm );

            string_list_add_item( temp_string_list, current_record.hash );
            string_list_add_item( temp_string_list, current_record.file_path );
        }

        /* update counter. */

        duplicate_progress++;
        print_progress_counter( "checking for duplicate checksums",
                                duplicate_progress );

        result = db->retrieve_next_record( db, &current_record );
    }

    /* loop through the duplicates list, if we find a list  */
    /* with more than 2 items, it has a duplicate, so we    */
    /* print the list in pretty form to the output_file.    */

    temp_list = list_get_first_node( duplicates );

    while( temp_list )
    {
        if( ( temp_string_list = temp_list->data ) != NULL )
        {
            if( temp_string_list->size > 3 )
            {
                /* this list has more than one file_path in it. */
                /* first print out the hash, then the files.    */

                osi_log_message_to_file( output_file, "" );

                osi_log_message_to_file( output_file, "  %s(%s) %d files",
                                         temp_string_list->list[0],
                                         temp_string_list->list[1],
                                         ( temp_string_list->size - 2 ) );

                osi_log_message_to_file( output_file, "" );

                for( index = 2; index < temp_string_list->size; index++ )
                {
                    osi_log_message_to_file( output_file, "    %s",
                          escape_filename( temp_string_list->list[index] ) );
                }
            }
        }

        temp_list = temp_list->next;
    }

    reset_progress_counter();
    duplicate_progress = 0;

    /* go through the list, removing all of the string_list */
    /* structures, then destroy the list structure itself.  */

    temp_list = list_get_first_node( duplicates );

    while( temp_list )
    {
        string_list_destroy( temp_list->data );
        temp_list = temp_list->next;
    }

    list_destroy( duplicates );
}

/******************************************************************************
**
**    Function: print_usage
**
**    Purpose:  print usage statement, then describe options.
**
******************************************************************************/

void print_usage()
{
    fprintf( stdout,
"usage: scale -l <db> -r <db> [-o <path>] [-f] [-d] [-q] [-n] [-m] |\n\
              [-c <config>] | -p <db> | -x <db> | -a <db> | -e <db> |\n\
               -d <db> | <db>\n\
\n\
 options:\n\
\n\
     -l, --left <db>            specifies left database for comparison.\n\
     -r, --right <db>           specifies right database for comparison.\n\
     -o <path>                  path for comparison results file.\n\
\n\
     -e, --extract-config <db>  extract config from database.\n\
     -p, --print <db>           print 'ls' listing of database contents.\n\
     -x, --print-all <db>       print complete listing of database contents.\n\
     -a, --checksums <db>       print db file names and their checksums.\n\
\n\
     -c, --config <config>      use specified config during comparison.\n\
\n\
     -f, --no-prompt            overwrite any existing log file.\n\
     -d, --show-duplicates      report files with same checksums.\n\
     -q, --quiet                no log output or progress meters.\n\
\n\
     -n, --no-new               don't include list of new files in report.\n\
     -m, --no-missing           don't include list of missing files in report.\n\
\n\
     -v, --version              print program version information.\n\
     -h, --help                 prints this usage statement.\n" );

    print_program_info();
}

/******************************************************************************
**
**    Function: print_program_info
**
**    Purpose:  print banner, including name and version.
**
******************************************************************************/

void print_program_info()
{
    osi_log_message( "" );

    osi_log_message( "-------------------------------------------" );

    osi_log_message( "   %s version %s [%s]",
                    PROGRAM_NAME, PROGRAM_VERSION, DB_MODULE );

    osi_log_message( "   osiris.shmoo.com              (c) 2003   " );
    osi_log_message( "-------------------------------------------" );

    osi_log_message( "" );
}

/******************************************************************************
**
**    Function: print_database_header
**
**    Purpose:  print the header found in the database showing the
**              details of that database, things like record count,
**              version of osiris, user, date, config/command-line, etc.
**
******************************************************************************/

void print_database_header( database_handler *db )
{
    char *token;
    char *header;

    char version[20]     = "";
    char username[512]   = "";
    char date[50]        = "";

    bool contains_config = TRUE;
    int record_count     = 0;

    if( db == NULL || db->state != DB_STATE_OPEN )
    {
        return;
    }

    if( ( header = db->retrieve_header( db ) ) == NULL )
    {
        print_database_error( db );
        osi_log_message_to_file( output_file,
                "error: unable to retrieve database header from: %s.",
                db->name );
        return;
    }

    /* begin to take apart the header and print it.     */
    /* we know the header contains the following        */
    /* format:                                          */
    /*                                                  */
    /*  OSIRIS_VERSION:DATE:USERNAME:CONFIG_FILE:COUNT  */

    /* get version. */

    if( ( token = strtok( header, ":" ) ) != NULL )
    {
        base64_decode( version, token );
    }

    /* get date. */

    if( ( token = strtok( NULL, ":" ) ) != NULL )
    {
        base64_decode( date, token );
    }

    /* get user who created database. */

    if( ( token = strtok( NULL, ":" ) ) != NULL )
    {
        base64_decode( username, token );
    }

    /* skip over config file. */

    token = strtok( NULL, ":" );

    if( token == NULL || strcmp( token, " " ) == 0 )
    {
        contains_config = FALSE;
    }

    /* get number of records. */

    if( ( token = strtok( NULL, ":" ) ) != NULL )
    {
        record_count = atoi( token );
    }

    /* now we print out the pretty header information. */

    osi_log_message_to_file( output_file, "" );
    osi_log_message_to_file( output_file, "[ database: %s ]", db->name );
    osi_log_message_to_file( output_file, "" );
    osi_log_message_to_file( output_file, "  records:       %d", record_count );

    /* if the database contains a config file, say so. */

    if( contains_config )
    {
        osi_log_message_to_file( output_file, "  source:        config file" );
    }

    else
    {
        osi_log_message_to_file( output_file, "  source:        command line" );
    }

    osi_log_message_to_file( output_file, "" );
    osi_log_message_to_file( output_file, "  created on:    %s", date );
    osi_log_message_to_file( output_file, "  created by:    %s", username );

    osi_log_message_to_file( output_file, "  created with:  osiris %s",
                             version );

    osi_free( header );
}

/******************************************************************************
**
**    Function: print_database_checksums
**
**    Purpose:  print entire database to stdout, printing only the file
**              names and their checksum value.
**
**              perm|links|user|group|bytes|date|filter|file
**
******************************************************************************/

void print_database_checksums( database_handler *db )
{
    bool result;
    char *token;

    record file_record;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_SCALE, "printing database checksums." );

    print_database_header( db );
    osi_log_message( "" );

    /* print all records here. */

    result = db->retrieve_first_record( db, &file_record );

    while( result != FALSE )
    {
        osi_log_message( " (%s) %s", file_record.hash,
                         escape_filename( file_record.file_path ) );

        result = db->retrieve_next_record( db, &file_record );
    }
}

/******************************************************************************
**
**    Function: print_database_contents
**
**    Purpose:  print entire contents of the database to stdout.
**              we follow a similar format as the 'ls -la' command
**              here is the format:
**
**              perm|links|user|group|bytes|date|filter|file
**
******************************************************************************/

void print_database_contents( database_handler *db )
{
    bool result;
    char *token;

    record file_record;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_SCALE, "printing database contents." );

    print_database_header( db );
    osi_log_message( "" );

    /* print all records here. */

    result = db->retrieve_first_record( db, &file_record );

    while( result != FALSE )
    {
        char perms[15]  = "";
        char user[256]  = "";
        char group[256] = "";
        char date[20]   = "";

        struct osi_stat stats;
        struct tm *m_time;
        long time_value;

        struct passwd *file_owner;
        struct group *file_group;

        stats = file_record.stats;
        time_value = stats.mtime;

        m_time  = (struct tm *)localtime( &time_value );

        file_owner  = getpwuid( stats.uid );
        file_group  = getgrgid( stats.gid );

        /* lookup the user and group from the uid/gid  */

        if( file_owner != NULL )
        {
            safe_strlcpy( user, file_owner->pw_name, sizeof( user ) );
        }

        if( file_group != NULL )
        {
            safe_strlcpy( group, file_group->gr_name, sizeof( group ) );
        }

        /* get the permissions string and the mod time */
        /* as a string.                                */

        get_file_attribute_string( perms, sizeof( perms ), stats.permissions );
        strftime( date, sizeof( date ), DATE_FORMAT_STRING, m_time );

        osi_log_message( "  %s %3d %-6.6s %-6.6s %8d %s %-15.15s %s", perms,
                                   (int)stats.links,
                                   user, group,
                                   (int)stats.bytes,
                                   date,
                                   file_record.filter_description,
                                   escape_filename( file_record.file_path ) );

        result = db->retrieve_next_record( db, &file_record );
    }
}

/******************************************************************************
**
**    Function: print_database_contents
**
**    Purpose:  print entire contents of the database to stdout.
**              we follow a similar format as the 'ls -la' command
**              here is the format:
**
**              perm|links|user|group|bytes|date|filter|file
**
******************************************************************************/

void print_complete_database_contents( database_handler *db )
{
    bool result;
    char *token;

    record file_record;

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return;
    }

    OSI_DEBUG( __POS__, DEBUG_SCALE, "printing exhaustive database contents." );

    print_database_header( db );
    osi_log_message( "" );

    /* print all records here. */

    result = db->retrieve_first_record( db, &file_record );

    while( result != FALSE )
    {
        print_record( &file_record );
        result = db->retrieve_next_record( db, &file_record );
    }
}


/******************************************************************************
**
**    Function: print_database_config
**
**    Purpose:  extract the config file from the database
**              header, if there was one.  this is pretty cool
**              and could be useful.
**
******************************************************************************/

void print_database_config( database_handler *db )
{
    char *token;
    char *header;
    char *config;

    OSI_DEBUG( __POS__, DEBUG_SCALE, "printing database config." );

    if( ( db == NULL ) || ( db->state != DB_STATE_OPEN ) )
    {
        return;
    }

    if( ( header = db->retrieve_header( db ) ) == NULL )
    {
        print_database_error( db );
        osi_log_message( "unable to retrieve database header from: %s.",
                         db->name );
        return;
    }

    /* begin to take apart the header and print it.     */
    /* we know the header contains the following        */
    /* format:                                          */
    /*                                                  */
    /*  OSIRIS_VERSION:DATE:USERNAME:CONFIG_FILE:COUNT  */

    /* skip over version,date,and user. */

    token = strtok( header, ":" );
    token = strtok( NULL, ":" );
    token = strtok( NULL, ":" );

    /* get the config file. */

    token = strtok( NULL, ":" );

    if( token == NULL || strcmp( token, " " ) == 0 )
    {
        osi_log_message(
            "%s: this database was not created from a config file.",
            PROGRAM_NAME );

        return;
    }

    /* allocate for config data. */

    config = osi_malloc( strlen( token ) * 1.4 );
    base64_decode( config, token );

    osi_log_message( config );

    osi_free( header );
    osi_free( config );
}


/******************************************************************************
**
**    Function: open_database
**
**    Purpose:  open the specified database, verify that it can be
**              opened and that it is an osiris database.
**
******************************************************************************/

bool open_database( database_handler *db )
{
    char *header = NULL;

    if( db == NULL || db->name == NULL )
    {
        return FALSE;
    }

    /* verify we actually were given database names. */

    if( strlen( db->name ) == 0 )
    {
        return FALSE;
    }

    /* open the databases, if we are unable to open */
    /* either, we shutdown, no need to continue.    */

    if( db->open( db, DB_READ ) == FALSE )
    {
        return FALSE;
    }

    /* do a minimal sanity check on the  database, i.e. we  */
    /* attempt to fetch the header.                         */

    header  = db->retrieve_header( db );

    if( header == NULL )
    {
        osi_log_message(
            "error: database: %s does not contain a valid osiris header.",
            left_database->name );

        return FALSE;
    }

    return TRUE;
}

/******************************************************************************
**
**    Function: close_database
**
**    Purpose:  close the specified database.
**
******************************************************************************/

void close_database( database_handler *db )
{
    if( db != NULL )
    {
        db->close( db );
    }
}

/******************************************************************************
**
**    Function: open_databases
**
**    Purpose:  open left and right databases, exit if either can't be opened.
**
******************************************************************************/

void open_databases()
{
    /* first put names into the db handlers. */

    left_database->name  = left_database_path;
    right_database->name = right_database_path;

    if( open_database( left_database ) == FALSE )
    {
        print_database_error( left_database );
        osi_log_message( "error: left database cannot be opened, exiting." );

        halt();
        exit(0);
    }

    if( open_database( right_database ) == FALSE )
    {
        print_database_error( right_database );
        osi_log_message( "error: right database cannot be opened, exiting." );

        halt();
        exit(0);
    }
}

/******************************************************************************
**
**    Function: close_databases
**
**    Purpose:  close the left and right databases.
**
******************************************************************************/

void close_databases()
{
    close_database( left_database );
    close_database( right_database );
}

/******************************************************************************
**
**    Function: open_output_file
**
**    Purpose:  if we are logging to a file, verify it exists and
**              can be opened, prompt for overwrite.
**
******************************************************************************/

void open_output_file()
{
    bool result;

    /* if an output file was specified, we analyze */
    /* it, then open it.                           */

    if( strlen( output_file_path ) > 0 )
    {
        if( file_exists( output_file_path ) && ( no_prompts == FALSE ) )
        {
            result = osi_continue_prompt(
                        "warning: the file: %s already exists,\noverwrite ?",
                         output_file_path );

            if( result == FALSE )
            {
                halt();
                exit(0);
            }
        }

        output_file = fopen( output_file_path, "w" );

        if( output_file == NULL )
        {
            osi_log_message(
                "error: unable to create output file: %s, exiting.",
                 output_file_path );

            halt();
            exit(0);
        }

        chmod( output_file_path, 0400 );
    }

    /* no output file specified, so we point our   */
    /* output file to stdout.                      */

    else
    {
        output_file = stdout;
    }
}



/******************************************************************************
**
**    Function: close_output_file
**
**    Purpose:  shutdown the output file if we were using one, unless it's
**              stdout, then we do nothing.
**
******************************************************************************/

void close_output_file()
{
    if( ( output_file != NULL ) && ( output_file != stdout ) )
    {
        fclose( output_file );
    }
}

/******************************************************************************
**
**    Function: print_output_header
**
**    Purpose:  print a banner, then the headers of the two
**              databases.  we can add more stuff here later, i'm sure.
**
******************************************************************************/

void print_output_header()
{
    time_t current_time = time( NULL );

    osi_log_message_to_file( output_file, "\nosiris database comparison" );
    osi_log_message_to_file( output_file, (char *)ctime( &current_time ) );

    print_database_header( left_database );
    print_database_header( right_database );
}

/******************************************************************************
**
**    Function: print_summary
**
**    Purpose:  print counters of the scan results to output device.
**
******************************************************************************/

void print_summary()
{
    osi_log_message_to_file( output_file, "records compared:    %d",
                             compare_count );

    osi_log_message_to_file( output_file, "records that differ: %d",
                             difference_count );

    osi_log_message_to_file( output_file, "new records:         %d",
                             new_record_count );

    osi_log_message_to_file( output_file, "missing records:     %d",
                             missing_record_count );

    osi_log_message_to_file( output_file, "" );

    if( custom_config )
    {
        osi_log_message_to_file( output_file,
                    "config file used for comparison: %s",
                    config_file_path );
    }
        
    /* if we are logging to a file, we print the */
    /* summary to stdout anyway.                 */

    if( quiet_mode == FALSE && output_file != stdout )
    {
        osi_log_message( "" );
        osi_log_message( "" );
        osi_log_message( "records compared:    %d", compare_count );
        osi_log_message( "records that differ: %d", difference_count );

        if( print_new )
        {
            osi_log_message( "new records:         %d", new_record_count );
        }

        osi_log_message( "missing records:     %d", missing_record_count );
        osi_log_message( "" );
        
        if( custom_config )
        {
            osi_log_message( "config file used for comparison: %s",
                             config_file_path );

            osi_log_message( "" );
        }
    }
}

/******************************************************************************
**
**    Function: print_progress_counter
**
**    Purpose:  shows the spinning thingy so user has some feedback
**              on the progress of the scan.
**
******************************************************************************/

void print_progress_counter( const char *message, unsigned long counter )
{
    static int count   = 0;

    if( count < 10 )
    {
        count++;
        return;
    }

    /* reset after being stale. */

    count = 0;

    if( output_file != stdout && quiet_mode == FALSE )
    {
        fprintf( stdout, "\r%s:  %d", message, counter );
        fflush( stdout );
    }
}

/******************************************************************************
**
**    Function: reset_progress_counter
**
**    Purpose:  clobber the last number and show that it's done.
**
******************************************************************************/

void reset_progress_counter()
{
    if( output_file != stdout && quiet_mode == FALSE )
    {
        fprintf( stdout, " (done)                                         \n" );
        fflush( stdout );
    }
}

/******************************************************************************
**
**    Function: print_database_error
**
**    Purpose: retrieve last string error message from the database handler.
**             this should provide some explanation as to why the last db
**             action failed.
**
******************************************************************************/

void print_database_error( database_handler *db )
{
    char *error_message;

    if( db != NULL )
    {
        if( ( error_message = db->retrieve_last_error(db) ) != NULL )
        {
            if( error_message[0] != '\0' )
            {
                osi_log_message( "%s: %s", DB_MODULE, error_message );
            }
        }
    }
}

/******************************************************************************
**
**    Function: verify_config_file
**
**    Purpose:  make sure the config file we are asked to open and parse
**              exists, is a regular file, can be opened, and isn't
**              frickin huge.
**
******************************************************************************/

void verify_config_file()
{
    char *config_file;
    struct stat file_stats;

    config_file = config_module->file_name;

    if( stat( config_file, &file_stats ) != 0 )
    {
        osi_log_message_exit( "config file: %s cannot be opened.",
                              config_file );

        halt();
        exit(0);
    }

    if( !S_ISREG( file_stats.st_mode ) )
    {
        osi_log_message_exit( "config file: %s is not a regular file.",
                              config_file );

        halt();
        exit(0);
    }

    if( file_stats.st_size > 100000 )
    {
        osi_log_message_exit( "file: %s seems far too large, sorry.",
                              config_file );

        halt();
        exit(0);
    }
}

/******************************************************************************
**
**    Function: verify_config_state
**
**    Purpose:  look at the error and warning counts on the config module
**              and exit on errors, prompt on warnings.
**
******************************************************************************/

void verify_config_state()
{
    bool result;

    if( config_module->error_count > 0 )
    {
        halt();
        osi_log_message_exit(
                    "config file: %s contains %d fatal error(s), exiting.",
                     config_module->file_name,
                     config_module->error_count );
    }

    if( config_module->warning_count > 0 )
    {
        result = osi_continue_prompt(
                     "config file: %s contains %d warning(s),\ncontinue?",
                      config_module->file_name,
                      config_module->warning_count );

        if( result == FALSE )
        {
            halt();
            exit(0);
        }
    }
}

