
/******************************************************************************
**
**  Copyright (C) 2001  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it, however, you cannot sell it.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**  You should have received a copy of the license attached to the
**  use of this software.  If not, visit www.shmoo.com/osiris for
**  details.
**
******************************************************************************/

/******************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      utilities.c
**    Author:    Brian Wotring
**
**    Date:      June 22, 2001.
**    Project:   osiris
**
******************************************************************************/

#include "utilities.h"

/******************************************************************************
**
**    Function: osi_malloc
**
**    Purpose:  safe wrapper around malloc
**
******************************************************************************/

void * osi_malloc( size_t size )
{
    void *result = (void *)malloc( size );

    if( result != NULL )
    {
        return result;
    }

    fprintf( stderr, "malloc failed, jump ship.\n" );
    exit(0);
}

/******************************************************************************
**
**    Function: osi_free
**
**    Purpose:  safe wrapper around fre
**
******************************************************************************/

void osi_free( void *memory )
{
    if( memory > 0 )
    {
        free( memory );
    }
}

/******************************************************************************
**
**    Function: byte_swap64
**
**    Purpose:  swap byte order in a 64 bit integer, for use
**              in NTOHLL and HTONLL macros.
**
******************************************************************************/

osi_uint64 byte_swap64( osi_uint64 value )
{
    int index;
    int length = 8;

    osi_uint64 swapped = value;
    unsigned char *q = (unsigned char *)&swapped;

    for( index = 0; ( index < (length/2) ); index++ )
    {
        unsigned char temp = q[index];
        q[index] = q[length - index - 1];
        q[length - index - 1] = temp;
    }

    return swapped;
}


/******************************************************************************
**
**    Function: osi_fopen
**
**    Purpose:  safe wrapper around fopen
**
******************************************************************************/

FILE *osi_fopen( const char *file_name, const char *mode )
{
    FILE *file = NULL;
    int fd     = 0;

    struct stat pre_stats;
    struct stat post_stats;

    if( file_name == NULL || mode == NULL )
    {
        return NULL;
    }

    if( stat( file_name, &pre_stats ) != 0 )
    {
        return NULL;
    }

    /* make sure the file is a regular file */

    if( !S_ISREG( pre_stats.st_mode ) )
    {
        return NULL;
    }

    if( strcmp( mode, "r" ) == 0 )
    {
        fd = open( file_name, O_RDONLY, 0 );
    }

    else if( strcmp( mode, "r+" ) == 0 )
    {
        fd = open( file_name, O_RDWR, 0 );
    }

    else if( strcmp( mode, "w" ) == 0 )
    {
        fd = open( file_name, O_RDWR | O_CREAT | O_EXCL, 0 );
    }

    else if( strcmp( mode, "w+" ) == 0 )
    {
        fd = open( file_name, O_RDWR | O_CREAT | O_EXCL, 0 );
    }

    if( fd <= 0 )
    {
        return NULL;
    }

    if( fstat( fd, &post_stats ) != 0 )
    {
        close(fd);
        return NULL;
    }

    /* here we make sure the inode and device numbers are the     */
    /* same in the file we actually opened, compared to the file  */
    /* we performed the initial lstat() call on.                  */

    if( pre_stats.st_ino != post_stats.st_ino ||
        pre_stats.st_dev != post_stats.st_dev )
    {
        close(fd);
        return NULL;
    }

    return fdopen( fd, mode );
}

/******************************************************************************
**
**    Function: trim_white_space
**
**    Purpose:  remove any non printable characters from beginning and
**              end of string, if it's the NULL string, returns NULL
**              otherwise, return pointer to beginning of string.
**
******************************************************************************/

char *trim_white_space( char *string )
{
    int length;

    if( string == NULL || *string == '\0' )
    {
        return NULL;
    }

    /* remove any leading white space. */

    while( *string && !isgraph(*string) )
    {
        string++;
    }

    if( *string == '\0' )
    {
        return NULL;
    }

    /* remove any trailing white space */

    length = strlen( string );

    while( !isgraph( string[length] ) && ( length > 1 ) )
    {
        string[length] = '\0';
        length--;
    }

    if( *string == '\0' )
    {
        return NULL;
    }

    return string;
}


/******************************************************************************
**
**    Function: get_token
**
**    Purpose:  puts next non-white space sequence into token
**              and returns pointer to next token in string.
**              token must be as large as string.
**
******************************************************************************/

char *get_token( char *string, char *token )
{
    string = trim_white_space( string );

    if( string == NULL || token == NULL )
    {
        return NULL;
    }

    while( *string && isgraph(*string) )
    {
        *token = *string;

        token++;
        string++;
    }

    *token = '\0';
    string = trim_white_space( string );

    return string;
}

/******************************************************************************
**
**    Function: lowercase_string
**
**    Purpose:  turns every character in the string into its lowercase form.
**
******************************************************************************/

void lowercase_string( char *string )
{
    if( string == NULL )
    {
        return;
    }

    while( *string )
    {
        *string = tolower ( (int)*string );
        string++;
    }
}

/******************************************************************************
**
**    Function: osi_log_message
**
**    Purpose:  similar to printf, but will add a new line, and flush
**              the output stream.
**
******************************************************************************/

void osi_log_message( char *message, ... )
{
    va_list ap;

    if( message == NULL )
    {
        return;
    }

    va_start( ap, message );
    vfprintf( stdout, message, ap );
    va_end( ap );

    fprintf( stdout, "\n" );
    fflush( stdout );
}


/******************************************************************************
**
**    Function: osi_log_message_exit
**
**    Purpose:  logs message just like log_message, then exits safely.
**
******************************************************************************/

void osi_log_message_exit( char *message, ... )
{
    va_list ap;

    va_start( ap, message );
    vfprintf( stdout, message, ap );
    va_end( ap );

    fprintf( stdout, "\n" );
    fflush( stdout );

    exit(0);
}

/******************************************************************************
**
**    Function: osi_log_message_to_file
**
**    Purpose:  logs message to the specified file, then flushes the stream.
**
******************************************************************************/

void osi_log_message_to_file( FILE *file, char *message, ... )
{
    va_list ap;

    va_start( ap, message );
    vfprintf( file, message, ap );
    va_end( ap );

    fprintf( file, "\n" );
    fflush( file );
}

/******************************************************************************
**
**    Function: osi_continue_prompt
**
**    Purpose:  prints yes/no question to stdout and returns bool based
**              upon the answer given.
**
******************************************************************************/

bool osi_continue_prompt( char *message, ... )
{
    va_list ap;
    char answer[10] = "";

    if( message == NULL )
    {
        return FALSE;
    }

    va_start( ap, message );
    vfprintf( stdout, message, ap );
    va_end( ap );

    fprintf( stdout, " ( y/n ) " );
    fflush( stdout );

    while( fgets( answer, ( sizeof( answer ) - 1 ), stdin ) == NULL )
    {
        ;
    }

    if( answer[0] == 'y' || answer[0] == 'Y' )
    {
        return TRUE;
    }

    return FALSE;
}

/******************************************************************************
**
**    Function: file_exists
**
**    Purpose:  determine if this file exists and is a regular file.
**
*******************************************************************************/

bool file_exists( const char *file_path )
{
    struct stat file_stats;

    if( file_path != NULL )
    {
        if( stat( file_path, &file_stats ) == 0 )
        {
            return TRUE;
        }
    }

    return FALSE;
}

/******************************************************************************
**
**    Function: directory_exists
**
**    Purpose:  determine if this directory exists and is a regular file.
**
******************************************************************************/

bool directory_exists( const char *directory_path )
{
    struct stat directory_stats;

    if( directory_path != NULL )
    {
        if( ( stat( directory_path, &directory_stats ) == 0 ) &&
            S_ISDIR( directory_stats.st_mode ) )
        {
            return TRUE;
        }
    }

    return FALSE;
}

/******************************************************************************
**
**    Function: get_file_attribute_string
**
**    Purpose:  get the permissions string for the file in the
**              standard 10 char form -rwxrwxrwx
**
******************************************************************************/

void get_file_attribute_string( char *buffer, int buffer_size,
                                mode_t attributes )
{
    buffer[0] = '\0';

    if( buffer_size < 11 )
    {
        return;
    }

    safe_strlcpy( buffer, "----------", buffer_size );

    /* first, determine type. */

    if( ( attributes & S_IFMT ) == S_IFDIR )
    {
        buffer[0] = 'd';
    }

    else if( ( attributes & S_IFMT ) == S_IFBLK )
    {
        buffer[0] = 'b';
    }

    else if( ( attributes & S_IFMT ) == S_IFCHR )
    {
        buffer[0] = 'c';
    }

    else if( ( attributes & S_IFMT ) == S_IFIFO )
    {
        buffer[0] = 'p';
    }

    else if( ( attributes & S_IFMT ) == S_IFLNK )
    {
        buffer[0] = 'l';
    }

    /* now find permissions for user bits. */

    if( attributes & S_IREAD )
    {
        buffer[1] = 'r';
    }

    if( attributes & S_IWRITE )
    {
        buffer[2] = 'w';
    }

    if( ( attributes & S_ISUID ) && ( attributes & S_IEXEC ) )
    {
        buffer[3] = 's';
    }

    else if( attributes & S_IEXEC )
    {
        buffer[3] = 'x';
    }

    else if( attributes & S_ISUID )
    {
        buffer[3] = 'S';
    }

    /* now find permissions for group bits. */

    if( attributes & S_IRGRP )
    {
        buffer[4] = 'r';
    }

    if( attributes & S_IWGRP )
    {
        buffer[5] = 'w';
    }

    if( ( attributes & S_ISUID ) && ( attributes & S_IXGRP ) )
    {
        buffer[6] = 's';
    }

    else if( attributes & S_IXGRP )
    {
        buffer[6] = 'x';
    }

    else if( attributes & S_ISUID )
    {
        buffer[6] = 'S';
    }

    /* now find permissions for others bits. */

    if( attributes & S_IROTH )
    {
        buffer[7] = 'r';
    }

    if( attributes & S_IWOTH )
    {
        buffer[8] = 'w';
    }

    if( ( attributes & S_ISVTX ) && ( attributes & S_IXOTH ) )
    {
        buffer[9] = 't';
    }

    else if( attributes & S_IXOTH )
    {
        buffer[9] = 'x';
    }

    else if( attributes & S_ISVTX )
    {
        buffer[9] = 'T';        
    }
}

/******************************************************************************
**
**    Function: get_string_up_to_token
**
**    Purpose:  get the string up to the token, returns the
**              start of the string after the token, NULL if not found.
**              caller is responsible for knowing if there is sufficient
**              space in the result buffer!
**
******************************************************************************/

char * get_string_up_to_token( char *string, char *result, char token )
{
    char *temp;
    int length;

    if( string == NULL || result == NULL )
    {
        return FALSE;
    }

    temp = strchr( string, token );

    if( temp == NULL )
    {
        return NULL;
    }

    length = ( temp - string );
    safe_strlcpy( result, string, ( length + 1 ) );

    /* terminate the buffer, move past the token in string  */
    /* and return a pointer to that string after the token. */

    result[length] = '\0';
    temp++;

    return temp;
}


/* base64 routines stolen fromp apache, we need to give */
/* the proper credits eventually.                       */


#ifdef CHARSET_EBCDIC

#include <sys/types.h>

extern const unsigned char os_toascii[256];
extern const unsigned char os_toebcdic[256];
void ebcdic2ascii(void *dest, const void *srce, size_t count);
void ascii2ebcdic(void *dest, const void *srce, size_t count);

#endif                          /* CHARSET_EBCDIC */

/* aaaack but it's fast and const should make it shared text page. */
static const unsigned char pr2six[256] =
{
#ifndef CHARSET_EBCDIC
    /* ASCII table */
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64, 64, 64, 63,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64,
    64,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
    15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 64,
    64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64
#else /*CHARSET_EBCDIC*/
    /* EBCDIC table */
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 63, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 64, 64, 64, 64, 64, 64,
    64, 35, 36, 37, 38, 39, 40, 41, 42, 43, 64, 64, 64, 64, 64, 64,
    64, 64, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64,  0,  1,  2,  3,  4,  5,  6,  7,  8, 64, 64, 64, 64, 64, 64,
    64,  9, 10, 11, 12, 13, 14, 15, 16, 17, 64, 64, 64, 64, 64, 64,
    64, 64, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 64, 64,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64
#endif /*CHARSET_EBCDIC*/
};

int ap_base64decode_len(const char *bufcoded)
{
    int nbytesdecoded;
    register const unsigned char *bufin;
    register int nprbytes;

    bufin = (const unsigned char *) bufcoded;
    while (pr2six[*(bufin++)] <= 63);

    nprbytes = (bufin - (const unsigned char *) bufcoded) - 1;
    nbytesdecoded = ((nprbytes + 3) / 4) * 3;

    return nbytesdecoded + 1;
}

int base64_decode(char *bufplain, const char *bufcoded)
{
#ifdef CHARSET_EBCDIC
    int i;
#endif                          /* CHARSET_EBCDIC */
    int len;

    len = base64_decode_binary((unsigned char *) bufplain, bufcoded);
#ifdef CHARSET_EBCDIC
    for (i = 0; i < len; i++)
    {
        bufplain[i] = os_toebcdic[bufplain[i]];
    }
#endif                          /* CHARSET_EBCDIC */
    bufplain[len] = '\0';
    return len;
}

/* This is the same as ap_base64udecode() except on EBCDIC machines, where
 * the conversion of the output to ebcdic is left out.
 */
int base64_decode_binary(unsigned char *bufplain,
                                   const char *bufcoded)
{
    int nbytesdecoded;
    register const unsigned char *bufin;
    register unsigned char *bufout;
    register int nprbytes;

    bufin = (const unsigned char *) bufcoded;
    while (pr2six[*(bufin++)] <= 63);
    nprbytes = (bufin - (const unsigned char *) bufcoded) - 1;
    nbytesdecoded = ((nprbytes + 3) / 4) * 3;

    bufout = (unsigned char *) bufplain;
    bufin = (const unsigned char *) bufcoded;

    while (nprbytes > 4)
    {
        *(bufout++) =
            (unsigned char) (pr2six[*bufin] << 2 | pr2six[bufin[1]] >> 4);
        *(bufout++) =
            (unsigned char) (pr2six[bufin[1]] << 4 | pr2six[bufin[2]] >> 2);
        *(bufout++) =
            (unsigned char) (pr2six[bufin[2]] << 6 | pr2six[bufin[3]]);
        bufin += 4;
        nprbytes -= 4;
    }

    /* Note: (nprbytes == 1) would be an error, so just ingore that case */
    if (nprbytes > 1)
    {
        *(bufout++) =
            (unsigned char) (pr2six[*bufin] << 2 | pr2six[bufin[1]] >> 4);
    }

    if (nprbytes > 2)
    {
        *(bufout++) =
            (unsigned char) (pr2six[bufin[1]] << 4 | pr2six[bufin[2]] >> 2);
    }

    if (nprbytes > 3)
    {
        *(bufout++) =
            (unsigned char) (pr2six[bufin[2]] << 6 | pr2six[bufin[3]]);
    }

    nbytesdecoded -= (4 - nprbytes) & 3;
    return nbytesdecoded;
}

static const char basis_64[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

int ap_base64encode_len(int len)
{
    return ((len + 2) / 3 * 4) + 1;
}

int base64_encode(char *encoded, const char *string, int len)
{
#ifndef CHARSET_EBCDIC
    return base64_encode_binary(encoded, (const unsigned char *) string, len);
#else                           /* CHARSET_EBCDIC */
    int i;
    char *p;

    p = encoded;
    for (i = 0; i < len - 2; i += 3) {
        *p++ = basis_64[(os_toascii[string[i]] >> 2) & 0x3F];
        *p++ = basis_64[((os_toascii[string[i]] & 0x3) << 4) |
                        ((int) (os_toascii[string[i + 1]] & 0xF0) >> 4)];
        *p++ = basis_64[((os_toascii[string[i + 1]] & 0xF) << 2) |
                        ((int) (os_toascii[string[i + 2]] & 0xC0) >> 6)];
        *p++ = basis_64[os_toascii[string[i + 2]] & 0x3F];
    }
    if (i < len) {
        *p++ = basis_64[(os_toascii[string[i]] >> 2) & 0x3F];
        if (i == (len - 1)) {
            *p++ = basis_64[((os_toascii[string[i]] & 0x3) << 4)];
            *p++ = '=';
        }
        else {
            *p++ = basis_64[((os_toascii[string[i]] & 0x3) << 4) |
                            ((int) (os_toascii[string[i + 1]] & 0xF0) >> 4)];
            *p++ = basis_64[((os_toascii[string[i + 1]] & 0xF) << 2)];
        }
        *p++ = '=';
    }

    *p++ = '\0';
    return p - encoded;
#endif                          /* CHARSET_EBCDIC */
}

/* This is the same as ap_base64encode() except on EBCDIC machines, where
 * the conversion of the input to ascii is left out.
 */
int base64_encode_binary(char *encoded,
                                       const unsigned char *string, int len)
{
    int i;
    char *p;

    p = encoded;
    for (i = 0; i < len - 2; i += 3) {
        *p++ = basis_64[(string[i] >> 2) & 0x3F];
        *p++ = basis_64[((string[i] & 0x3) << 4) |
                        ((int) (string[i + 1] & 0xF0) >> 4)];
        *p++ = basis_64[((string[i + 1] & 0xF) << 2) |
                        ((int) (string[i + 2] & 0xC0) >> 6)];
        *p++ = basis_64[string[i + 2] & 0x3F];
    }
    if (i < len) {
        *p++ = basis_64[(string[i] >> 2) & 0x3F];
        if (i == (len - 1)) {
            *p++ = basis_64[((string[i] & 0x3) << 4)];
            *p++ = '=';
        }
        else {
            *p++ = basis_64[((string[i] & 0x3) << 4) |
                            ((int) (string[i + 1] & 0xF0) >> 4)];
            *p++ = basis_64[((string[i + 1] & 0xF) << 2)];
        }
        *p++ = '=';
    }

    *p++ = '\0';
    return p - encoded;
}

/******************************************************************************
**
**    Function: escape_filename
**
**    Purpose:  turn control characters into their escaped form.
**
*****************************************************************************/

char * escape_filename( char *file_path )
{
	static char escaped[MAXPATHLEN * 3] = "";

	char *input  = (char *)file_path;
	char *output = (char *)escaped;

    static char *octals[] =
    {
	"000", "001", "002", "003", "004", "005", "006", "007",
	"010", "011", "012", "013", "014", "015", "016", "017",
	"020", "021", "022", "023", "024", "025", "026", "027",
	"030", "031", "032", "033", "034", "035", "036", "037",
	"040", "041", "042", "043", "044", "045", "046", "047",
	"050", "051", "052", "053", "054", "055", "056", "057",
	"060", "061", "062", "063", "064", "065", "066", "067",
	"070", "071", "072", "073", "074", "075", "076", "077",
	"100", "101", "102", "103", "104", "105", "106", "107",
	"110", "111", "112", "113", "114", "115", "116", "117",
	"120", "121", "122", "123", "124", "125", "126", "127",
	"130", "131", "132", "133", "134", "135", "136", "137",
	"140", "141", "142", "143", "144", "145", "146", "147",
	"150", "151", "152", "153", "154", "155", "156", "157",
	"160", "161", "162", "163", "164", "165", "166", "167",
	"170", "171", "172", "173", "174", "175", "176", "177",
	"200", "201", "202", "203", "204", "205", "206", "207",
	"210", "211", "212", "213", "214", "215", "216", "217",
	"220", "221", "222", "223", "224", "225", "226", "227",
	"230", "231", "232", "233", "234", "235", "236", "237",
	"240", "241", "242", "243", "244", "245", "246", "247",
	"250", "251", "252", "253", "254", "255", "256", "257",
	"260", "261", "262", "263", "264", "265", "266", "267",
	"270", "271", "272", "273", "274", "275", "276", "277",
	"300", "301", "302", "303", "304", "305", "306", "307",
	"310", "311", "312", "313", "314", "315", "316", "317",
	"320", "321", "322", "323", "324", "325", "326", "327",
	"330", "331", "332", "333", "334", "335", "336", "337",
	"340", "341", "342", "343", "344", "345", "346", "347",
	"350", "351", "352", "353", "354", "355", "356", "357",
	"360", "361", "362", "363", "364", "365", "366", "367",
	"370", "371", "372", "373", "374", "375", "376", "377",
    };

    if( file_path == NULL )
    {
        return NULL;
    }

    while( *input )
    {
        if( iscntrl( *input ) )
        {
            *output++ = '\\';
            *output++ = octals[*input][0];
            *output++ = octals[*input][1];
            *output++ = octals[*input][2];

            input++;
	    }

	    else
        {
	        *output++ = *input++;
        }
    }

    *output = '\0';
    return escaped;
}

static size_t safe_strnlen( const char *str, size_t nchars )
{
    size_t i,result = 0;
    
    if( str && nchars > 0 )
    {
        for (i = 0; i < nchars && str[i] != '\0'; i++)
        {
            result++;
        }
    }
    
    return result;
}

size_t safe_strlcpy( char * dst, const char *src, size_t nchars )
{
    size_t cpy_len;
    size_t src_len = 0;

    if( !src )
    {
        goto exit_gracefully;
    }
    
    /* while nchars doesn't strictly apply to src, we're not going to */
    /* be able to copy more than that many characters from it anyway... */
    
    src_len = safe_strnlen( src,nchars );

    if( !dst )
    {
        goto exit_gracefully;
    }
    
    if( nchars > 0 )
    {
        cpy_len = MINIMUM( src_len, ( nchars - 1 ) );
        memcpy( dst, src, cpy_len );
        dst[cpy_len] = '\0';
    }
    
exit_gracefully:

    /* note: this is src because we're returning how much */
    /* we would have written, not how much we did write. */
    
    return src_len;
}


size_t safe_strlcat( char *dst, const char *src, size_t nchars )
{
    size_t orig_dst_len = 0;
    size_t src_len = 0;
    size_t result = nchars;
    size_t space_left, cpy_len;  
    char * cpy_start;

    if( !dst )
    {
        goto exit_gracefully;
    }
 
     /*   
     from the openbsd strlcat manpage:
     Note however, that if strlcat() traverses size characters without finding
     a NUL, the length of the string is considered to be size and the destina-
     tion string will not be NUL-terminated (since there was no space for the
     NUL).  This keeps strlcat() from running off the end of a string.  In
     practice this should not happen (as it means that either size is incor-
     rect or that dst is not a proper ``C'' string).  The check exists to pre-
     vent potential security problems in incorrect code.
     */
  
    orig_dst_len = safe_strnlen( dst, nchars );
    
    if( !src )
    {
        goto exit_gracefully;
    }
    
    /* while nchars doesn't strictly apply to src, we're not going to */
    /* be able to copy more than that many characters from it anyway...*/
    
    src_len = safe_strnlen( src,nchars );
    space_left = nchars - orig_dst_len;

    
    if( space_left > 0 )
    {
        cpy_start = dst + orig_dst_len;
        cpy_len = MINIMUM( src_len, space_left - 1 );
        memcpy( cpy_start, src, cpy_len );
        cpy_start[cpy_len] = '\0';
        
        /* remember, how big the write *would* have been... */
        
        result = orig_dst_len + src_len;
    }
    
exit_gracefully:

    return result;
}

size_t safe_vsnprintf( char *str, size_t nchars, const char * format,
                       va_list ap )
{
    size_t result = nchars;

    if( nchars < 1 || !str || !format )
    {
        goto exit_gracefully;
    }
    
    result = vsnprintf( str, nchars, format, ap ); 

    /* some (v)snprintf implmentations don't write a NUL at the end    */
    /* of the buffer if it is overrun. So, ensure that the last char   */
    /* in the buffer is NUL so that we can check if it was overwritten */
    /* by snprintf.                                                    */
    
    str[nchars - 1] = '\0';
    
    /* some (v)snprintf implementations return -1 on overrun. instead, */
    /* return act like the C99 standard: overrun means */
    /* return val is greater than or equal to the buffersize. However, */
    /* C99 says snprintf should return how much buffer space *would* */
    /* have been needed. We can't easily calculate that if the underlying */
    /* snprintf doesn't do it for us. */
    
    if( result < 0 )
    {
        result = nchars;
    } 
    
exit_gracefully:

    return result;
}

size_t safe_snprintf( char *str, size_t nchars, const char *format, ... )
{
    size_t result;
    va_list args;

    va_start(args,format);
    result = safe_vsnprintf( str, nchars, format, args );
    va_end(args);

    return result;
}



